/* 
 * $RCSfile: gj_connection.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/29 16:28:41 $
 * 
 * $Revision: 1.13 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <gnet.h>

#include "gj_connection.h"
#include "gj_support.h"


struct t_connection
{
  guint id;

  gchar *host;
  guint port;

  GTcpSocket *tcpsocket;
  GTcpSocketConnectAsyncID connect_id; /* used to stop async connects */

  gboolean connected;

  /* callbacks */
  socket_cb_in cb_in;
  socket_cb_hup cb_hup;
  socket_cb_connect cb_connect;
  socket_cb_disconnect cb_disconnect;
  socket_cb_error cb_error;
};

GList *connectionList;
guint connectionID = 69;

guint gj_connection_get_next_id();

/* list functions */
gboolean gj_connection_list_add(connection c);
gboolean gj_connection_list_del(connection c);
GList *gj_connection_list_get();

/* item functions */
connection gj_connection_new(const gchar *host, const guint port);
gboolean gj_connection_free(connection c);
void gj_connection_destroy(connection c);

/*
 * code
 */
guint gj_connection_get_next_id()
{
  return connectionID++;
}

gboolean gj_connection_init()
{
  return TRUE;
}

gboolean gj_connection_term()
{
  gint index = 0;

  /* list */
  if(connectionList == NULL)
    return TRUE;

  /* free all items on the list */
  for(index=0;index<g_list_length(connectionList);index++)
    {
      connection c = NULL;
      
      if((c = (connection) g_list_nth_data(connectionList, index)) != NULL)
	gj_connection_free(c);
    }

  g_list_free(connectionList);
  connectionList = NULL;
  return TRUE;
}

/* list functions */
gboolean gj_connection_list_add(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }

  connectionList = g_list_append(connectionList, c);
  return TRUE;
}

gboolean gj_connection_list_del(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }

  connectionList = g_list_remove(connectionList, c);
  return TRUE;
}

GList *gj_connection_list_get()
{
  return connectionList;
}

connection gj_connection_new(const gchar *host, const guint port)
{
  connection c = NULL;

  if(host == NULL || g_utf8_strlen(host, -1) < 1)
    {
      g_warning("%s: host was NULL or length was < 1", __FUNCTION__);
      return NULL;
    }

  if(port < 1 || port > 65535)
    {
      g_warning("%s: port:%d was invalid, range is (1-65535)", __FUNCTION__, port);
      return NULL;
    }

  if((c = gj_connection_find_by_host_and_port(host, port)) != NULL)
    g_warning("%s: connection already exists with host:'%s' and port:%d", __FUNCTION__, host, port);

  if((c = (connection) g_new(struct t_connection,1)) != NULL)
    {
      c->id = gj_connection_get_next_id();

      c->host = g_strdup(host);
      c->port = port;

      c->tcpsocket = NULL;
      c->connect_id = 0;

      /* 	c->watch_id_in = 0; */
	
      c->connected = FALSE;

      c->cb_in = NULL;
      c->cb_hup = NULL;
      c->cb_connect = NULL;
      c->cb_disconnect = NULL;
      c->cb_error = NULL;

      gj_connection_list_add(c);
    }

  return c;
}

gboolean gj_connection_free(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }

  /* free from list */
  gj_connection_list_del(c);
  
  /* destroy members */
  gj_connection_destroy(c);

  /* free struct */
  g_free(c);
  
  return TRUE;
}

void gj_connection_destroy(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return;
    }

  c->id = 0;

  if(c->host != NULL)
    {
      g_free(c->host);
      c->host = NULL;
    }

  c->port = 0;

  if(c->tcpsocket != NULL)
    {
      if(gj_connection_get_connected(c) == TRUE)
	gj_connection_disconnect(c);

      c->tcpsocket = NULL;
    }

  c->connect_id = 0;

  /*     c->watch_id_in = 0; */
    
  c->connected = FALSE;

  c->cb_in = NULL;
  c->cb_hup = NULL;
  c->cb_connect = NULL;
  c->cb_disconnect = NULL;
  c->cb_error = NULL;
}

connection gj_connection_find_by_host_and_port(const gchar *host, const guint port)
{
  gint index = 0;

  if(host == NULL)
    {
      g_warning("%s: host was NULL", __FUNCTION__);
      return NULL;
    }

  if(port < 1 || port > 65535)
    {
      g_warning("%s: port:%d was invalid, range is (1-65535)", __FUNCTION__, port);
      return NULL;
    }

  for(index=0;index<g_list_length(connectionList);index++)
    {
      connection c = g_list_nth_data(connectionList, index);

      if(c == NULL)
	{
	  g_warning("%s: getting nth item(%d) in list was NULL", __FUNCTION__, index);
	  continue;
	}

      if(g_strcasecmp(gj_connection_get_host(c), host) == 0 && gj_connection_get_port(c) == port)
	return c;
    }

  return NULL;  
}

connection gj_connection_find_by_id(const guint id)
{
  gint index = 0;

  if(id < 1)
    {
      g_warning("%s: id:%d was < 1", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(connectionList);index++)
    {
      connection c = g_list_nth_data(connectionList, index);

      if(c == NULL)
	{
	  g_warning("%s: getting nth item(%d) in list was NULL", __FUNCTION__, index);
	  continue;
	}

      if(gj_connection_get_id(c) == id)
	return c;
    }

  return NULL;  
}

/*
 * member functions 
 */

/* gets */ 
const guint gj_connection_get_id(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return 0;
    }

  return c->id;
}

const gchar *gj_connection_get_host(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return NULL;
    }

  return c->host;
}

const guint gj_connection_get_port(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return 0;
    }

  return c->port;
}

GTcpSocket *gj_connection_get_tcpsocket(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return NULL;
    }

  return c->tcpsocket;
}

GTcpSocketConnectAsyncID gj_connection_get_connect_id(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return 0;
    }

  return c->connect_id;    
}

gboolean gj_connection_get_connected(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }

  return c->connected; 
}

socket_cb_in gj_connection_get_cb_in(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return NULL;
    }

  return c->cb_in;
}

socket_cb_hup gj_connection_get_cb_hup(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return NULL;
    }

  return c->cb_hup;
}

socket_cb_connect gj_connection_get_cb_connect(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return NULL;
    }

  return c->cb_connect;
}

socket_cb_disconnect gj_connection_get_cb_disconnect(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return NULL;
    }

  return c->cb_disconnect;
}

socket_cb_error gj_connection_get_cb_error(connection c)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return NULL;
    }

  return c->cb_error;
}

/* sets */
gboolean gj_connection_set_host(connection c, const gchar *host)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(host == NULL)
    {
      g_warning("%s: host was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(c->host != NULL)
    {
      g_free(c->host);
      c->host = NULL;
    }

  c->host = g_strdup(host);
  return TRUE;
}

gboolean gj_connection_set_port(connection c, const guint port)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  c->port = port;
  return TRUE;
}

gboolean gj_connection_set_tcpsocket(connection c, GTcpSocket *ts)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  
      
  c->tcpsocket = ts;
  return TRUE;
}

gboolean gj_connection_set_connect_id(connection c, GTcpSocketConnectAsyncID connect_id)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  c->connect_id = connect_id;
  return TRUE;
}

gboolean gj_connection_set_connected(connection c, gboolean connected)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  c->connected = connected;
  return TRUE;
}

gboolean gj_connection_set_cb_in(connection c, socket_cb_in cb)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  

  c->cb_in = cb;
  return TRUE;
}

gboolean gj_connection_set_cb_hup(connection c, socket_cb_hup cb)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  

  c->cb_hup = cb;
  return TRUE;
}

gboolean gj_connection_set_cb_connect(connection c, socket_cb_connect cb)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  

  c->cb_connect = cb;
  return TRUE;
}

gboolean gj_connection_set_cb_disconnect(connection c, socket_cb_disconnect cb)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  

  c->cb_disconnect = cb;
  return TRUE;
}

gboolean gj_connection_set_cb_error(connection c, socket_cb_error cb)
{
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }  

  c->cb_error = cb;
  return TRUE;
}

/*
 * connection functions
 */

#ifndef WIN32 

gboolean gj_connection_connect(connection c)
{
  GTcpSocketConnectAsyncID connect_id = 0;

  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }     
    
  g_message("%s: connection:0x%.8x, connecting asynchronously to host:'%s', port:%d...", 
	    __FUNCTION__, (gint)c, gj_connection_get_host(c), gj_connection_get_port(c));

  connect_id = gnet_tcp_socket_connect_async(gj_connection_get_host(c),
					     gj_connection_get_port(c), 
					     (GTcpSocketConnectAsyncFunc)gj_connection_connect_cb, 
					     GINT_TO_POINTER(gj_connection_get_id(c)));
    
  /* remember id so we can stop it if we need to */
  g_message("%s: connection:0x%.8x, connect id is %d", __FUNCTION__, (gint)c, (gint)connect_id);
  gj_connection_set_connect_id(c, connect_id);
    
  return TRUE;
}

#else

/* FIXME: Windows glib does has a bug which does not permit me
   to using the async method of connecting.  For the time being, we will
   use synchronous connecting */
gboolean gj_connection_connect(connection c)
{
  GTcpSocket *ts = NULL;
  GIOChannel *ch = NULL;    

  socket_cb_connect cbc = NULL;

  gint id = 0;

  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }     
    
  g_message("%s: connection:0x%.8x, connecting synchronously to host:'%s', port:%d...", 
	    __FUNCTION__, (gint)c, gj_connection_get_host(c), gj_connection_get_port(c));

  ts = gnet_tcp_socket_connect(gj_connection_get_host(c),
			       gj_connection_get_port(c));
				   
  
  if(ts == NULL)
    {
      g_warning("%s: could not create new socket to host:'%s', port:%d", 
		__FUNCTION__, gj_connection_get_host(c), gj_connection_get_port(c));
      return FALSE;
    }

  id = gj_support_get_unique_id();

  gj_connection_set_connect_id(c, 0);
  gj_connection_set_tcpsocket(c, ts);
  gj_connection_set_connected(c, TRUE);
  
#if 0
  /* set type of service - only works if OS supports it */
  gnet_tcp_socket_set_tos(ts, GNET_TOS_RELIABILITY);
#endif
  
  g_message("%s: connection:0x%.8x, connected.", __FUNCTION__, (gint)c);
  
  if((ch = gnet_tcp_socket_get_io_channel(ts)) == NULL)
    {
      /* FIXME: delete and clear up at this point */
      g_warning("%s: connection:0x%.8x, GIOChannel was NULL", __FUNCTION__, (gint)c);
      return FALSE;
    }
  
  /* callbacks */
  g_io_add_watch(ch, G_IO_IN | G_IO_ERR | G_IO_NVAL | G_IO_HUP, 
		 gj_connection_watch_cb, GINT_TO_POINTER(gj_connection_get_id(c)));   
  /*       g_io_add_watch_full(ch, -1, G_IO_IN | G_IO_ERR | G_IO_NVAL | G_IO_HUP,  */
  /* 			  gj_connection_watch_cb, GINT_TO_POINTER(id), NULL);    */
  
  if((cbc = gj_connection_get_cb_connect(c)) != NULL)
    (cbc)(c);

  return TRUE;
}

#endif

connection gj_connection_create(const gchar *host, const guint port)
{
  connection c = NULL;
  const gchar *localhost = NULL;

  if(host == NULL)
    {
      g_warning("%s: host was NULL", __FUNCTION__);
      return NULL;
    }
  
  if(port < 1 || port > 65535)
    {
      g_warning("%s: port:%d was invalid, range is (1-65535)", __FUNCTION__, port);
      return NULL;
    }

#ifndef WIN32
  if((localhost = gnet_inetaddr_get_host_name()) == NULL)
    {
      g_warning("%s: localhost was NULL", __FUNCTION__);
      return NULL;
    }
#endif

  c = gj_connection_new(host, port);

  if(c == NULL)
    {
      g_warning("%s: could not create new connection (for host:'%s', port:%d)", __FUNCTION__, host, port);
      return NULL;
    }

  g_message("%s: connection:0x%.8x, local host is:'%s'", __FUNCTION__, (gint)c, localhost);
  g_message("%s: connection:0x%.8x, host:'%s', port:%d", __FUNCTION__, (gint)c, host, port);

  return c;
}

void gj_connection_connect_cb(GTcpSocket *ts, GTcpSocketConnectAsyncStatus status, gpointer data)
{
  connection c = NULL;
  socket_cb_connect cbc = NULL;

  GIOChannel *ch = NULL;    
  guint id = 0;

  if(data == NULL)
    {
      g_warning("%s: data is NULL, no way to know which connection this is", __FUNCTION__);
      return;
    }

  id = GPOINTER_TO_INT(data);
    
  if((c = gj_connection_find_by_id(id)) == NULL)
    {
      g_warning("%s: could not find connection id:%d", __FUNCTION__, id);
      return;
    }
    
  if(status != GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK)
    {
      socket_cb_error cbe = NULL;

      g_message("%s: GTcpSocket:0x%.8x, GTcpSocketConnectAsyncStatus:%d->'%s'", 
		__FUNCTION__, (gint)ts, (gint)status, gj_connection_connect_translate_status(status));

      if((cbe = gj_connection_get_cb_error(c)) != NULL)
	(cbe)(c, status);
      
      return;
    }

  gj_connection_set_connect_id(c, 0);
  gj_connection_set_tcpsocket(c, ts);
  gj_connection_set_connected(c, TRUE);
  
#if 0
  /* set type of service - only works if OS supports it */
  gnet_tcp_socket_set_tos(ts, GNET_TOS_RELIABILITY);
#endif
  
  g_message("%s: connection:0x%.8x, connected.", __FUNCTION__, (gint)c);
  
  if((ch = gnet_tcp_socket_get_io_channel(ts)) == NULL)
    {
      g_warning("%s: connection:0x%.8x, GIOChannel was NULL", __FUNCTION__, (gint)c);
      return;
    }
  
  /* callbacks */
  g_io_add_watch(ch, G_IO_IN | G_IO_ERR | G_IO_NVAL | G_IO_HUP, 
		 gj_connection_watch_cb, GINT_TO_POINTER(id));   
  /*       g_io_add_watch_full(ch, -1, G_IO_IN | G_IO_ERR | G_IO_NVAL | G_IO_HUP,  */
  /* 			  gj_connection_watch_cb, GINT_TO_POINTER(id), NULL);    */
  
  if((cbc = gj_connection_get_cb_connect(c)) != NULL)
    (cbc)(c);
}

gboolean gj_connection_disconnect(connection c)
{
  GTcpSocket *ts = NULL;
  socket_cb_disconnect cb = NULL; 
    
  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }      

  if((ts = gj_connection_get_tcpsocket(c)) == NULL)
    {
      g_warning("%s: connection:0x%.8x, GTcpSocket was NULL", __FUNCTION__, (gint)c);
      return FALSE;
    }

  gnet_tcp_socket_delete(ts);

  g_message("%s: connection:0x%.8x, disconnected.", __FUNCTION__, (gint)c);
  gj_connection_set_connected(c, FALSE);

  if((cb = gj_connection_get_cb_disconnect(c)) != NULL)  
    (cb)(c);  

  return TRUE;
}

gboolean gj_connection_disconnect_by_host_and_port(const gchar *host, const guint port)
{
  connection c = NULL;

  if((c = gj_connection_find_by_host_and_port(host, port)) == NULL)
    {
      g_warning("%s: could not find connection", __FUNCTION__);
      return FALSE;
    }
    
  return gj_connection_disconnect(c);
}

gboolean gj_connection_watch_cb(GIOChannel* source, GIOCondition condition, gpointer data)
{
  gint id = 0;
  connection c = NULL;
    
  id = GPOINTER_TO_INT(data);
    
  g_log(NULL, G_LOG_LEVEL_DEBUG,"%s: source:0x%.8x, condition:%d->'%s', data:0x%.8x", 
	__FUNCTION__, (gint)source, condition, gj_connection_io_condition_translate(condition), (gint)data);
    
  /*     g_message("source:0x%.8x, condition:%d->'%s', data:0x%.8x", (gint)cource, condition, gj_connection_io_condition_translate(condition), (gint)data); */

  if((c = gj_connection_find_by_id(id)) == NULL)
    {
      g_warning("%s: could not find connection id:%d", __FUNCTION__, id);
      return FALSE;
    }

  if(condition & (G_IO_ERR | G_IO_NVAL))
    {
      g_warning("%s: connection:0x%.8x, GIOChannel:0x%.8x, GIOCondition:%d, G_IO_ERR | G_IO_NVAL received", 
		__FUNCTION__, (gint)c, (gint)source, condition);
      return FALSE;
    }
  else if(condition & G_IO_IN)
    {
      gchar *text = NULL;
      socket_cb_in cb = NULL;

      guint total_bytes = 0;

      while(TRUE)
	{
	  GError *error = NULL;
	  GIOStatus status = 0;

	  gchar buffer[1024] = ""; 
	  guint bytes = 0;     
	  gchar *buffer_terminated = NULL;

	  status = g_io_channel_read_chars(source, buffer, sizeof(buffer), &bytes, &error);
	  total_bytes += bytes; 

	  if(status == G_IO_STATUS_ERROR)
	    {
	      g_warning("%s: connection:0x%.8x, GIOChannel:0x%.8x, read error:%d->'%s'", 
			__FUNCTION__, 
			(gint)c, 
			(gint)source, 
			status, 
			gj_connection_io_status_translate(status));
		
	      if(error != NULL)
		{
		  GIOChannelError cherr = g_io_channel_error_from_errno(error->code);
		    
		  g_message("%s: connection:0x%.8x, GError:%d->'%s' GIOChannelError:%d->'%s'", 
			    __FUNCTION__, 
			    (gint)c, 
			    error->code, 
			    error->message, 
			    cherr, 
			    gj_connection_io_error_translate(cherr));
    		}

	      break;
	    }
	    
	  if(total_bytes == 0)
	    {
	      g_message("%s: connection:0x%.8x, 0 bytes read, connection closed", __FUNCTION__, (gint)c);
	      gj_connection_disconnect(c);

	      return FALSE;
	    }
	    
	  /* if already read ONCE, and this time, 0 was returned, 
	     no more to bytes to read */
	  if(bytes == 0 && total_bytes > bytes)
	    break;

	  g_log(NULL, G_LOG_LEVEL_DEBUG, "%s: connection:0x%.8x, status:%d->'%s'", 
		__FUNCTION__, (gint)c, status, gj_connection_io_status_translate(status));  

	  /* concat to string */
	  buffer_terminated = g_strndup(buffer, bytes);

	  if(buffer_terminated != NULL)
	    {
	      if(text == NULL)
		text = g_strdup(buffer_terminated);
	      else
		text = g_strconcat(text, buffer_terminated, NULL);
		
	      g_free(buffer_terminated);
	    }

	  /* if one read and buffer is not full OR 
	     bytes is > 0 and size of buffer is not filled */
	  if((bytes < sizeof(buffer) && total_bytes == bytes) || 
	     (bytes < sizeof(buffer) && bytes > 0))
	    break;
	}
	
      if(text != NULL)
	{
	    if(g_utf8_validate(text, -1, NULL) == FALSE)
	      g_warning("%s: text is not valid utf8", __FUNCTION__);

	    g_message("%s: connection:0x%.8x, read total of %d bytes (%d chars): -\n<<<< %s", 
		      __FUNCTION__, (gint)c, (gint)total_bytes, (gint)g_utf8_strlen(text, -1), text);

	    if((cb = gj_connection_get_cb_in(c)) != NULL) 
		(cb)(c, text); 
	}
    }
  else if(condition & G_IO_HUP)
    {
      /* FIXME: this should be socket_cb_hup - just havnt had time yet. */
      socket_cb_disconnect cb = NULL;

      g_message("%s: connection:0x%.8x, connection was disconnected", __FUNCTION__, (gint)c);
      gj_connection_set_connected(c, FALSE);

      if((cb = gj_connection_get_cb_disconnect(c)) != NULL) 
	(cb)(c); 

      return FALSE;
    }
  else
    {
      g_message("%s: connection:0x%.8x, received unhandled GIOCondition:%d->'%s'", 
		__FUNCTION__, (gint)c, condition, gj_connection_io_condition_translate(condition));
    }
	
  /* return FALSE to close watch */
  return TRUE;
}

gboolean gj_connection_write(connection c, const gchar *text)
{
  GTcpSocket *ts = NULL;
  GIOChannel *ch = NULL;
  /*     GIOError error; */
  GIOStatus status = 0;
  GError *error = NULL;

  guint bytes = 0;

  if(c == NULL)
    {
      g_warning("%s: connection was NULL", __FUNCTION__);
      return FALSE;
    }      
    
  if(text == NULL || g_utf8_strlen(text, -1) < 1)
    {
      g_warning("%s: connection:0x%.8x, text was NULL or length was < 1", __FUNCTION__, (gint)c);
      return FALSE;
    }

  if((ts = gj_connection_get_tcpsocket(c)) == NULL)
    {
      g_warning("%s: connection:0x%.8x, GTcpSocket was NULL", __FUNCTION__, (gint)c);
      return FALSE;
    }

  if((ch = gnet_tcp_socket_get_io_channel(ts)) == NULL)
    {
      g_warning("%s: connection:0x%.8x, GIOChannel was NULL", __FUNCTION__, (gint)c);
      return FALSE;
    }

  status = g_io_channel_write_chars(ch, text, g_utf8_strlen(text, -1), &bytes, &error);
  g_log(NULL, G_LOG_LEVEL_DEBUG, "%s: connection:0x%.8x, GIOStatus is %d->'%s'", 
	__FUNCTION__, (gint)c, status, gj_connection_io_status_translate(status));
    
  if(error != NULL)
    {
      g_warning("%s: connection:0x%.8x, error:%d->'%s'", __FUNCTION__, (gint)c, error->code, error->message);
      return FALSE;
    }

  g_message("%s: connection:0x%.8x, wrote %.5d bytes: -\n>>>> %s", __FUNCTION__, (gint)c, bytes, text);
    
  return TRUE;
}

/*
 * translations 
 */
const gchar *gj_connection_connect_translate_status(GTcpSocketConnectAsyncStatus status)
{
  switch(status)
    {
    case GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK: return "connection succeeded";
    case GTCP_SOCKET_CONNECT_ASYNC_STATUS_INETADDR_ERROR: return "Error, address lookup failed";
    case GTCP_SOCKET_CONNECT_ASYNC_STATUS_TCP_ERROR: return "Error, could not connect";
    }

  return "unknown";
}

const gchar *gj_connection_io_status_translate(GIOStatus status)
{
  switch(status)
    {
    case G_IO_STATUS_ERROR: return "An error occurred.";
    case G_IO_STATUS_NORMAL: return "Success.";
    case G_IO_STATUS_EOF: return "End of file.";
    case G_IO_STATUS_AGAIN: return "Resource temporarily unavailable.";
    }

  return "unknown";
}

const gchar *gj_connection_io_error_translate(gint error)
{
  switch(error)
    {
      /* Derived from errno */
    case G_IO_CHANNEL_ERROR_FBIG: return "File too large.";
    case G_IO_CHANNEL_ERROR_INVAL: return "Invalid argument.";
    case G_IO_CHANNEL_ERROR_IO: return "IO error.";
    case G_IO_CHANNEL_ERROR_ISDIR: return "File is a directory. ";
    case G_IO_CHANNEL_ERROR_NOSPC: return "No space left on device. ";
    case G_IO_CHANNEL_ERROR_NXIO: return "No such device or address.";
    case G_IO_CHANNEL_ERROR_OVERFLOW: return "Value too large for defined datatype.";
    case G_IO_CHANNEL_ERROR_PIPE: return "Broken pipe.";
      /* Other */
    case G_IO_CHANNEL_ERROR_FAILED: return "Some other error.";
    }

  return "unknown";
}

const gchar *gj_connection_io_condition_translate(gint condition)
{
  switch(condition)
    {
    case G_IO_IN: return "There is data to read.";
    case G_IO_OUT: return "Data can be written (without blocking).";
    case G_IO_PRI: return "There is urgent data to read.";
    case G_IO_ERR: return "Error condition.";
    case G_IO_HUP: return "Hung up (the connection has been broken, usually for pipes and sockets).";
    case G_IO_NVAL: return "Invalid request. The file descriptor is not open.";
    }

  return "unknown";
}

#ifdef __cplusplus
}
#endif
