/* 
 * $RCSfile: gj_event.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/01 18:15:07 $
 * 
 * $Revision: 1.12 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>

#include <time.h> 

#include <glib.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include "gj_event.h"
#include "gj_roster.h"
#include "gj_parser.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_translate.h"
#include "gj_support.h"


/* structs */
struct t_event
{
  guint id;
  gchar *jid;

  gboolean in_event; /* event has been looked at but not free'd, eg. in chat window already */
    
  eventType type;
  eventProc proc;
    
  gint ri_id;
    
  time_t time;
    
  void *data;
};

gboolean gj_event_list_add(event ev);
gboolean gj_event_list_del(event ev);

void gj_event_destroy(event ev);

void gj_event_cb_presence_from(presence pres);
void gj_event_cb_message_from(message msg);

GList *event_list = NULL;
eventProc event_generic_cb = NULL;

/*
 * code
 */
gboolean gj_event_init()
{
  gj_message_set_cb_from(gj_event_cb_message_from);
  gj_presence_set_cb_from(gj_event_cb_presence_from); 
  
  return TRUE;
}

gboolean gj_event_term()
{
  gint index = 0;
    
  /* list */
  if(event_list == NULL)
    return TRUE;
    
  /* free all items on the list */
  for(index=0;index<g_list_length(event_list);index++)
    {
      event ev = NULL;
	
      if((ev = (event) g_list_nth_data(event_list,index)) != NULL)
	gj_event_free(ev);
    }
    
  g_list_free(event_list);
  event_list = NULL;
    
  return TRUE;
}

gboolean gj_event_list_add(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return FALSE;
    }

  event_list = g_list_append(event_list,ev);

  return TRUE;
}

gboolean gj_event_list_del(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return FALSE;
    }

  event_list = g_list_remove(event_list,ev);
  return TRUE;
}

event gj_event_new(gchar *jid, eventType type, void *data)
{
  event ev = NULL;
  rosterItem ri = NULL;

  if(jid == NULL) 
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  if((ri = gj_roster_item_find_by_jid(jid)) == NULL)
    {
      g_warning("%s: could not find roster item by jid:'%s'", __FUNCTION__, jid);
      return NULL;
    }

  if(type < 1 || type >= eventTypeEnd)
    {
      g_warning("%s: event type was out of range", __FUNCTION__);
      return NULL;
    }

  if((ev = g_new(struct t_event,1)))
    {
      ev->id = gj_support_get_unique_id();

      ev->jid = g_strdup(jid);

      ev->type = type;
      ev->data = data;

      ev->in_event = FALSE;
      ev->proc = NULL;
      ev->ri_id = gj_roster_item_get_id(ri);

      ev->time = time(0);
    }

  event_list = g_list_append(event_list, ev);

  return ev;
}

gboolean gj_event_new_by_post_from_jid(gchar *jid, eventType type, void *data)
{
  event ev = NULL;
  eventProc proc = NULL;
  rosterItem ri = NULL;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if((proc = gj_event_get_cb()) == NULL)
    {
      g_warning("%s: no response callback to post new event to", __FUNCTION__);
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_jid(jid)) == NULL)
    {
      g_message("%s: no roster item found, creating temporary solution...", __FUNCTION__);

      if((ri = gj_roster_item_new_unknown_by_jid(jid)) == NULL)
	{
	  g_warning("%s: couldn't add temporary roster item", __FUNCTION__);
	  return FALSE;
	}
    }

  if((ev = gj_event_new(jid, type, data)) == NULL)
    {
      g_warning("%s: could not create new event", __FUNCTION__);
      return FALSE;
    }

  /* FIXME: we need to add this contact to the roster somewhere.. */

  /* call callback to post new event */
  (proc)(ev->id);
  return TRUE;
}

gboolean gj_event_new_by_post_from_ri_id(gint id, eventType type, void *data)
{
  rosterItem ri = NULL;
  
  if(id < 0)
    {
      g_warning("%s: roster item id:%d was < 0", __FUNCTION__, id);
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_id(id)) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  return gj_event_new_by_post_from_jid(gj_roster_item_get_jid(ri), type, data);
}

gboolean gj_event_free(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return FALSE;
    }

  /* free mem */
  gj_event_list_del(ev);  

  /* destroy members */
  gj_event_destroy(ev);

  /* free structure */
  g_free(ev);

  return TRUE;
}

gboolean gj_event_free_by_id(gint eventID)
{
  gint index = 0;

  if(eventID < 0)
    {
      g_warning("%s: eventID:%d was < 0", __FUNCTION__, eventID);
      return FALSE;
    }

  /* loop through all events on contacts event list to look for unique eventID */
  for(index=0;index<g_list_length(event_list);index++)
    {
      event ev = g_list_nth_data(event_list, index);
      
      if(ev == NULL)
	{
	  g_warning("%s: getting nth item(%d) in roster item event list was NULL", __FUNCTION__, index);
	  continue;
	}
      
      if(ev->id == eventID)
	return gj_event_free(ev);
    }   

  g_warning("%s: could not find event by id:%d", __FUNCTION__, eventID);
  return FALSE;
}

void gj_event_destroy(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return;
    }

  ev->id = 0;

  if(ev->jid != NULL)
    {
      g_free(ev->jid);
      ev->jid = NULL;
    }

  ev->type = eventTypeEnd;
  ev->proc = NULL;

  ev->ri_id = -1;

  /*
   * should we be freeing these types here or should the module clear them
   * up when it is terminated, e.g. infoqueryTerm();
   */
  if(ev->data != NULL)
    {
      switch(ev->type)
	{
	case eventTypeMessage:
	  {
	    message msg = (message) ev->data;
	    gj_message_free(msg);
	  }
	  break;

	case eventTypePresence:
	  {
	    presence pres = (presence) ev->data;
	    gj_presence_free(pres);
	  }
	  break;
	
	default:
	  break;
	}
    }
}

/*
 * response proc for inbound events 
 */
gboolean gj_event_set_cb(eventProc proc)
{
  if(proc == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  event_generic_cb = proc;

  return TRUE;
}

eventProc gj_event_get_cb()
{
  return event_generic_cb;
}


/*
 * utils
 */
event gj_event_nth(gint index)
{
  event ev = NULL;

  if(event_list == NULL)
    {
      g_warning("%s: event list was NULL", __FUNCTION__);
      return NULL;
    }

  if(index < 0 || index > g_list_length(event_list))
    {
      g_warning("%s: index:%d was < 0 || > %d", __FUNCTION__, index, g_list_length(event_list));
      return NULL;
    }

  /* get first item and return eventID */
  ev = (event)g_list_nth_data(event_list, index);
  
  return ev;
}

event gj_event_nth_by_roster_item(rosterItem ri, gint index)
{
  gint i = 0;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  /* do not warning this - this is perfectly normal */
  if(event_list == NULL || g_list_length(event_list) < 1)
    return NULL;

  if(index < 0 || index > g_list_length(event_list))
    {
      g_warning("%s: index:%d was < 0 || > %d", __FUNCTION__, index, g_list_length(event_list));
      return NULL;
    }

  for(i=0;i<g_list_length(event_list);i++)
    {
      rosterItem ri_tmp = NULL;
      event ev = g_list_nth_data(event_list, i);

      if(ev == NULL) continue;
      if((ri_tmp = gj_roster_item_find_by_id(gj_event_get_ri_id(ev))) == NULL) continue;
      if(gj_roster_item_get_id(ri_tmp) == gj_roster_item_get_id(ri) && i == index) return ev;
    }

  return NULL;
}

gint gj_event_next(rosterItem ri)
{
  gint index = 0;
  gint id = -1;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return -1;
    }
  
  id = gj_roster_item_get_id(ri);

  /* do not warning this - this is perfectly normal */
  if(event_list == NULL || g_list_length(event_list) < 1)
    return -1;

  /* loop through all event_list on contacts event list to look for unique eventID */
  for(index=0;index<g_list_length(event_list);index++)
    {
      event ev = g_list_nth_data(event_list, index);
      if(ev == NULL) continue;
      if(ev->ri_id == id) return ev->id;
    }
  
  return -1;
}

gint gj_event_count(rosterItem ri)
{
  gint index = 0;
  gint id = -1;
  guint count = 0;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return -1;
    }

  id = gj_roster_item_get_id(ri);

  /* do not warning this - this is perfectly normal */
  if(event_list == NULL)
    return 0;

  /* loop through all event_list on contacts event list to look for unique eventID */
  for(index=0,count=0;index<g_list_length(event_list);index++)
    {
      event ev = g_list_nth_data(event_list, index);
      if(ev == NULL) continue;
      if(ev->ri_id == id) count++;
    }

  return count;
}

event gj_event_find_by_id(gint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: id:%d was < 0", __FUNCTION__, id);
      return NULL;
    }

  /* loop through all event_list on contacts event list to look for unique eventID */
  for(index=0;index<g_list_length(event_list);index++)
    {
      event ev = g_list_nth_data(event_list, index);
      
      if(ev == NULL) continue;
      if(ev->id == id) return ev;
    }
 
  return NULL;
}

event gj_event_find_next_by_type(rosterItem ri, eventType type)
{
  gint index = 0;
  gint id = -1;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  id = gj_roster_item_get_id(ri);

  if(type < 0)
    {
      g_warning("%s: event type:%d was < 0", __FUNCTION__, type);
      return NULL;
    }

  /* do not warning this - this is perfectly normal */
  if(event_list == NULL || g_list_length(event_list) < 1)
    return NULL;

  /* loop through all events on contacts event list to look for unique eventID */
  for(index=0;index<g_list_length(event_list);index++)
    {
      event ev = g_list_nth_data(event_list, index);

      if(ev == NULL) continue;
      if(ev->ri_id == id && ev->type == type) return ev;
    }

  return NULL;
}

event gj_event_find_next_by_type_and_message_type(rosterItem ri, eventType type, messageType messageType)
{
  gint index = 0;
  gint id = -1;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  id = gj_roster_item_get_id(ri);

  if(type < 0)
    {
      g_warning("%s: event type:%d was < 0", __FUNCTION__, type);
      return NULL;
    }

  /* do not warning this - this is perfectly normal */
  if(event_list == NULL || g_list_length(event_list) < 1)
    return NULL;

  /* loop through all events on contacts event list to look for unique eventID */
  for(index=0;index<g_list_length(event_list);index++)
    {
      event ev = g_list_nth_data(event_list, index);
      
      if(ev == NULL) continue;
     
      if(ev->ri_id == id && ev->type == type)
	{
	  message msg = (message) ev->data;
	    
	  if(msg == NULL)
	    continue;

	  if(gj_message_get_type(msg) == messageType)
	    return ev;
	}
    }

  return NULL;
}

gint gj_event_count_by_type(rosterItem ri, eventType type)
{
  gint index = 0;
  gint count = 0;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return -1;
    }

  if(type < 0)
    {
      g_warning("%s: roster item type:%d was < 0", __FUNCTION__, type);
      return -1;
    }

  /* loop through all events on contacts event list to look for unique eventID */
  for(index=0,count=0;index<g_list_length(event_list);index++)
    {
      event ev = g_list_nth_data(event_list, index);
      
      if(ev == NULL) continue;
      if(ev->type == type) count++;
    }

  return count;
}

/*
 * member data
 */ 

/* sets */
gboolean gj_event_set_ri_id(event ev, gint id)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return FALSE;
    }

  if(id < 0)
    {
      g_warning("%s: roster item id:%d was < 0", __FUNCTION__, id);
      return FALSE;
    }

  ev->ri_id = id;
  return TRUE;
}

/* gets */
gint gj_event_get_id(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return -1;
    }

  return ev->id;
}

gchar *gj_event_get_jid(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return NULL;
    }

  return ev->jid;
}

eventType gj_event_get_type(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return -1;
    }

  return ev->type;
}

void *gj_event_get_data(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return NULL;
    }

  return ev->data;
}

time_t gj_event_get_time(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return -1;
    }

  return ev->time;
}

gint gj_event_get_ri_id(event ev)
{
  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return -1;
    } 

  return ev->ri_id;
}

/* convenience function */
rosterItem gj_event_get_roster_item(event ev)
{
  gint id = -1;
  rosterItem ri = NULL;

  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return NULL;
    } 

  id = gj_event_get_ri_id(ev);
  ri = gj_roster_item_find_by_id(id);
  return ri;
}

void gj_event_cb_presence_from(presence pres)
{
  /* show debugging */
  g_message("++++++++++ received presence id:%3.3d ++++++++++", gj_presence_get_request_id(pres));
  g_message("++++");

  g_message("++++ to:'%s', resource:'%s'", gj_presence_get_to_jid(pres), gj_presence_get_to_resource(pres));
  g_message("++++ from:'%s', resource:'%s'", gj_presence_get_from_jid(pres), gj_presence_get_from_resource(pres));

  if(gj_presence_get_type(pres) != -1)
    g_message("++++ type:'%d'->'%s'", gj_presence_get_type(pres), gj_translate_presence_type(gj_presence_get_type(pres)));

  if(gj_presence_get_show(pres) != -1)
    g_message("++++ show:'%d'->'%s'", gj_presence_get_show(pres), gj_translate_presence_show(gj_presence_get_show(pres)));

  g_message("++++ priority:'%d'", gj_presence_get_priority(pres));  

  if(gj_presence_get_status(pres) != NULL)
    g_message("++++ status:'%s'", gj_presence_get_status(pres));

  if(gj_presence_get_error_code(pres) != NULL)
    g_message("++++ error code:'%s'", gj_presence_get_error_code(pres));
  if(gj_presence_get_error_reason(pres) != NULL)
    g_message("++++ error reason:'%s'", gj_presence_get_error_reason(pres));

  g_message("++++");
  g_message("++++++++++++++++++++++++++++++++++++++++++++++");

  /* create new event */
  if(gj_event_new_by_post_from_jid(gj_presence_get_from_jid(pres), eventTypePresence, (void*)pres) == FALSE)
    {
      g_warning("%s: failed to create new event", __FUNCTION__);
      return;
    } 
}

/* 
 * callbacks
 */
void gj_event_cb_message_from(message msg)
{
  if(msg == NULL)
    return;

  g_message("++++++++++ received message id:%3.3d ++++++++++", gj_message_get_request_id(msg));
  g_message("++++");
  g_message("++++ type:'%d'->'%s'", gj_message_get_type(msg), gj_translate_message_type(gj_message_get_type(msg)));
  g_message("++++");

  if(gj_message_get_id(msg) != NULL) g_message("++++ id:'%s'", gj_message_get_id(msg));
  if(gj_message_get_to_jid(msg) != NULL) g_message("++++ to:'%s'", gj_message_get_to_jid(msg));
  if(gj_message_get_to_resource(msg) != NULL) g_message("++++ resource:'%s'", gj_message_get_to_resource(msg));
  if(gj_message_get_from_jid(msg) != NULL) g_message("++++ from:'%s'", gj_message_get_from_jid(msg));
  if(gj_message_get_from_resource(msg) != NULL) g_message("++++ resource:'%s'", gj_message_get_from_resource(msg));
  if(gj_message_get_subject(msg) != NULL) g_message("++++ subject:'%s'", gj_message_get_subject(msg));
  if(gj_message_get_thread(msg) != NULL) g_message("++++ thread:'%s'", gj_message_get_thread(msg));
  if(gj_message_get_body(msg) != NULL) g_message("++++ body:'%s'", gj_message_get_body(msg));  

  if(gj_message_get_xevent_type(msg) != 0) 
  {
      gchar *translation = NULL;
      translation = gj_translate_message_xevent_type(gj_message_get_xevent_type(msg));
      g_message("++++ x:event type:'%d'->'%s'", gj_message_get_xevent_type(msg), translation);  
      if(translation != NULL) g_free(translation);
  }

  if(gj_message_get_xevent_message_id(msg) != NULL) g_message("++++ x:event message id:'%s'", gj_message_get_xevent_message_id(msg));  
  if(gj_message_get_xoob_url(msg) != NULL) g_message("++++ x:oob url:'%s'", gj_message_get_xoob_url(msg));  
  if(gj_message_get_xoob_description(msg) != NULL) g_message("++++ x:oob description:'%s'", gj_message_get_xoob_description(msg));  

  if(gj_message_get_error_code(msg) != NULL)
    g_message("++++ error code:'%s'", gj_message_get_error_code(msg));  
  if(gj_message_get_error_reason(msg) != NULL)
    g_message("++++ error reason:'%s'", gj_message_get_error_reason(msg));  
 
  g_message("++++");
  g_message("+++++++++++++++++++++++++++++++++++++++++++++");

  if(gj_message_get_type(msg) != messageTypeXEvent)
  {
      /* if XEvents ask to be notified when the message has 
	 been displayed, this is the time to send that notification */
      if(gj_message_get_xevent_type(msg) & messageXEventTypeDelivered)
	  gj_message_send_xevent(messageXEventTypeDelivered, gj_message_get_from_jid(msg), gj_message_get_id(msg));
  }

  /* prefer method of posting by ri_id. */
  if(gj_event_new_by_post_from_jid(gj_message_get_from_jid(msg), eventTypeMessage, (void*)msg) == FALSE)
    {
      g_warning("%s: failed to create new event", __FUNCTION__);
      return;
    } 
}

/*
 * debugging 
 */
void gj_event_print_all(rosterItem ri)
{
  gint index = 0;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return;
    }

  g_message("******* Roster Item Event List Start ********");  

  for(index=0;index<gj_event_count(ri);index++)
    {
      event ev = NULL;

      ev = g_list_nth_data(event_list,index);
      if(ev == NULL) continue;

      gj_event_print(ev);
    }

  g_message("******* Roster Item Event List End **********");  

}

void gj_event_print(event ev)
{
  if(ev == NULL)
    return;

  g_message("******* Roster Item Event ********");
  g_message("*******");
  g_message("******* id:%d",ev->id);
  g_message("******* jid:'%s'",ev->jid);
  g_message("*******"); 
  g_message("******* roster item id:%d",(gint)ev->ri_id);
  g_message("*******"); 
  g_message("******* in_event:%s",ev->in_event?"yes":"no");
  g_message("*******"); 
  g_message("******* type:%d",ev->type);
  g_message("******* data:0x%.8x",(gint)ev->data);
  g_message("******* proc:0x%.8x",(gint)ev->proc);
  g_message("*******"); 
  g_message("******* Roster Item Event ********");
}


#ifdef __cplusplus
}
#endif
