/* 
 * $RCSfile: gj_gtk_connection_progress.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/01 22:21:33 $
 * 
 * $Revision: 1.23 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_GNOME
#include <gnome.h>
#endif

#include "common.h"

#include "gj_main.h"
#include "gj_parser.h"
#include "gj_iq_requests.h"
#include "gj_roster.h"
#include "gj_translate.h"
#include "gj_me.h"
#include "gj_support.h"

#include "gj_gtk_roster.h"
#include "gj_gtk_connection_progress.h"
#include "gj_gtk_connection_settings.h"
#include "gj_gtk_password_required.h"


/* local functions */
gboolean gj_gtk_cp_reload();
gboolean gj_gtk_cp_unload();
gboolean gj_gtk_cp_hide();
gboolean gj_gtk_cp_show();

void gj_gtk_cp_set_header(gchar *message, ...);
void gj_gtk_cp_set_detail(gchar *message, ...);


/* 
 * member functions 
 */

/* gets */
gint gj_gtk_cp_get_state();
gj_gtk_cp_callback gj_gtk_cp_get_cb_connected();
gj_gtk_cp_callback gj_gtk_cp_get_cb_disconnected();

gchar *gj_gtk_cp_get_server();
gint16 gj_gtk_cp_get_port();
gchar *gj_gtk_cp_get_username();
gchar *gj_gtk_cp_get_password();
gchar *gj_gtk_cp_get_resource();

gboolean gj_gtk_cp_get_stop_timeout();

/* sets */
gboolean gj_gtk_cp_set_state(gint state);
gboolean gj_gtk_cp_set_cb_connected(gj_gtk_cp_callback cb);
gboolean gj_gtk_cp_set_cb_disconnected(gj_gtk_cp_callback cb);

gboolean gj_gtk_cp_set_server(gchar *server);
gboolean gj_gtk_cp_set_port(gint16 port);
gboolean gj_gtk_cp_set_username(gchar *username);
gboolean gj_gtk_cp_set_password(gchar *password);
gboolean gj_gtk_cp_set_resource(gchar *resource);

gboolean gj_gtk_cp_set_stop_timeout(gboolean stop_timeout);

/* 
 * states
 */

/* translations */
const gchar *gj_gtk_cp_translate_state(gint state);

/* state machine */
void gj_gtk_cp_state_machine(guint eventID);
void gj_gtk_cp_state_handle_uninitiated(guint eventID);
void gj_gtk_cp_state_handle_idle(guint eventID);
void gj_gtk_cp_state_handle_connecting(guint eventID);
void gj_gtk_cp_state_handle_authentication_method_requested(guint eventID);
void gj_gtk_cp_state_handle_authenticating(guint eventID);
void gj_gtk_cp_state_handle_ready(guint eventID);
void gj_gtk_cp_state_handle_disconnected(guint eventID);

void gj_gtk_cp_connection_start(stream stm, gint eventID);
void gj_gtk_cp_connection_stop(stream stm, gint eventID);
void gj_gtk_cp_connection_error(stream stm, gint eventID);

void gj_gtk_cp_complete(gboolean success);

gboolean gj_gtk_cp_progress_timeout(gpointer data);
void gj_gtk_cs_enter_password_cb(gboolean success, const gchar *password);

/* gui callbacks */
void on_button_cp_action_clicked(GtkButton *button, gpointer user_data);
gboolean on_connection_progress_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);

/* local structs */
typedef enum gj_gtk_cp_state 
  {
    gj_gtk_cp_state_uninitiated = 1,
    gj_gtk_cp_state_idle,
    gj_gtk_cp_state_connecting,
    gj_gtk_cp_state_authentication_method_requested,
    gj_gtk_cp_state_authenticating,
    gj_gtk_cp_state_ready,
    gj_gtk_cp_state_disconnected,

    gj_gtk_cp_state_end
  } gj_gtk_cp_state;

struct 
{
  GtkWidget *window;
  
  GtkWidget *button_action;

  GtkWidget *label_desc1;
  GtkWidget *label_desc2;

  GtkWidget *image;

  GtkWidget *progressbar;

} ctlCP;

struct 
{ 
  gint state;

  gchar *server;
  gint16 port;
  gchar *username;
  gchar *password;
  gchar *resource;

  gj_gtk_cp_callback cb_connected;
  gj_gtk_cp_callback cb_disconnected;

  gboolean stop_timeout;

} optCP; 


/*
 * member functions
 */

/* gets */
gint gj_gtk_cp_get_state()
{
  return optCP.state;
}

gj_gtk_cp_callback gj_gtk_cp_get_cb_connected()
{
  return optCP.cb_connected;
}

gj_gtk_cp_callback gj_gtk_cp_get_cb_disconnected()
{
  return optCP.cb_disconnected;
}

gchar *gj_gtk_cp_get_server()
{
  return optCP.server;
}

gint16 gj_gtk_cp_get_port()
{
  return optCP.port;
}

gchar *gj_gtk_cp_get_username()
{
  return optCP.username;
}

gchar *gj_gtk_cp_get_password()
{
  return optCP.password;
}

gchar *gj_gtk_cp_get_resource()
{
  return optCP.resource;
}

gboolean gj_gtk_cp_get_stop_timeout()
{
  return optCP.stop_timeout;
}


/* sets */
gboolean gj_gtk_cp_set_state(gint state)
{
  if(state < 1 || state >= gj_gtk_cp_state_end)
    {
      g_warning("gj_gtk_cp_set_state: state:%d was out of range", state);
      return FALSE;
    }

  g_message("+++++++++++++++++++++++++++++++++++++++++");
  g_message("++++ set state:%d->'%s'", state, gj_gtk_cp_translate_state(state));
  g_message("+++++++++++++++++++++++++++++++++++++++++");

  optCP.state = state;
  return TRUE;
}

/* set properties */
gboolean gj_gtk_cp_set_server(gchar *server)
{
  if(server == NULL)
    {
      g_warning("gj_gtk_cp_set_server: server was NULL");
      return FALSE;
    }

  if(optCP.server != NULL)
    g_free(optCP.server);

  optCP.server = g_strdup(server);
  return TRUE;
}

gboolean gj_gtk_cp_set_port(gint16 port)
{
  if(port < 1024)
    {
      g_warning("gj_gtk_cp_set_port: port:%d was out of range",port);
      return FALSE;
    }

  optCP.port = port;
  return TRUE;
}

gboolean gj_gtk_cp_set_username(gchar *username)
{
  if(username == NULL)
    {
      g_warning("gj_gtk_cp_set_username: username was NULL");
      return FALSE;
    }

  if(optCP.username != NULL)
    g_free(optCP.username);

  optCP.username = g_strdup(username);
  return TRUE;
}

gboolean gj_gtk_cp_set_password(gchar *password)
{
  if(password == NULL)
    {
      g_warning("gj_gtk_cp_set_password: password was NULL");
      return FALSE;
    }

  if(optCP.password != NULL)
    g_free(optCP.password);

  optCP.password = g_strdup(password);
  return TRUE;
}

gboolean gj_gtk_cp_set_resource(gchar *resource)
{
  if(resource == NULL)
    {
      g_warning("gj_gtk_cp_set_resource: resource was NULL");
      return FALSE;
    }

  if(optCP.resource != NULL)
    g_free(optCP.resource);

  optCP.resource = g_strdup(resource);
  return TRUE;
}

gboolean gj_gtk_cp_set_stop_timeout(gboolean stop_timeout)
{
  optCP.stop_timeout = stop_timeout;
  return TRUE;
}

gboolean gj_gtk_cp_set_cb_connected(gj_gtk_cp_callback cb)
{
  g_message("%s: set callback:0x%.8x", __FUNCTION__, (gint)cb);
  optCP.cb_connected = cb;
  return TRUE;
}

gboolean gj_gtk_cp_set_cb_disconnected(gj_gtk_cp_callback cb)
{
  g_message("%s: set callback:0x%.8x", __FUNCTION__, (gint)cb);
  optCP.cb_disconnected = cb;
  return TRUE;
}

const gchar *gj_gtk_cp_translate_state(gint state)
{
  switch(state)
    {
    case gj_gtk_cp_state_uninitiated: return _("uninitiated");
    case gj_gtk_cp_state_idle: return _("idle");
    case gj_gtk_cp_state_connecting: return _("connecting");
    case gj_gtk_cp_state_authentication_method_requested: return _("authentication method requested");
    case gj_gtk_cp_state_authenticating: return _("authenticating");
    case gj_gtk_cp_state_ready: return _("ready");
    case gj_gtk_cp_state_disconnected: return _("disconnected");
    }  

  return "unknown";
}

void gj_gtk_cp_state_machine(guint eventID)
{
  gint state = gj_gtk_cp_get_state();

  g_message("*****************************************");
  g_message("**** entered state:%d->'%s'", state, gj_gtk_cp_translate_state(state));
  g_message("*****************************************");

  switch(state)
    {
    case gj_gtk_cp_state_uninitiated:
      gj_gtk_cp_state_handle_uninitiated(eventID); break;
    case gj_gtk_cp_state_idle:
      gj_gtk_cp_state_handle_idle(eventID); break;
    case gj_gtk_cp_state_connecting:
      gj_gtk_cp_state_handle_connecting(eventID); break;
    case gj_gtk_cp_state_authentication_method_requested:
      gj_gtk_cp_state_handle_authentication_method_requested(eventID); break;
    case gj_gtk_cp_state_authenticating:
      gj_gtk_cp_state_handle_authenticating(eventID); break;
    case gj_gtk_cp_state_ready:
      gj_gtk_cp_state_handle_ready(eventID); break;
    case gj_gtk_cp_state_disconnected:
      gj_gtk_cp_state_handle_disconnected(eventID); break;
    }  
}

void gj_gtk_cp_state_handle_uninitiated(guint eventID)
{
}

void gj_gtk_cp_state_handle_idle(guint eventID)
{
  stream stm = NULL;

  gj_me_set_is_connected(FALSE);
  gj_me_set_uptime_stop();

  /* set up progress bar */
  gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_INFO,GTK_ICON_SIZE_DIALOG);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)1/4);
  gj_gtk_cp_set_header("<b>%s</b>", _("Connecting..."));
  gj_gtk_cp_set_detail("%s '%s' %s %d", 
		       _("server"),
		       gj_gtk_cp_get_server(), 
		       _("on port"),
		       gj_gtk_cp_get_port());

  /* create stream */
  if((stm = gj_stream_new(gj_gtk_cp_get_server())) == NULL)
    {
      gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)4/4);
      gj_gtk_cp_set_header("<b>%s</b>", _("Error"));
      gj_gtk_cp_set_detail(_("Could not create new stream."));

      g_warning("gj_gtk_cp_state_handle_idle: could not create new stream");

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }
 
  /* set struct info */
  gj_stream_set_port(stm, gj_gtk_cp_get_port());
  gj_stream_set_connected_cb(stm, gj_gtk_cp_connection_start);
  gj_stream_set_disconnected_cb(stm, gj_gtk_cp_connection_stop);
  gj_stream_set_error_cb(stm, gj_gtk_cp_connection_error);

  if(gj_stream_build(stm) == FALSE)
    {
      gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)4/4);
      gj_gtk_cp_set_header("<b>%s</b>", _("Error"));
      gj_gtk_cp_set_detail(_("Could not build stream."));

      g_warning("gj_gtk_cp_state_handle_idle: could not build new stream.");

      /* finish */
      gj_gtk_cp_complete(FALSE);

      /* clean up */
      gj_stream_free(stm);
     
      return;
    }

  if(gj_stream_open(stm) == FALSE)
    {
      gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gj_gtk_cp_set_header("<b>%s..</b>", _("Connection Error"));
      gj_gtk_cp_set_detail(_("Unable to open stream."));
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)4/4);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      /* clean up */
      gj_stream_free(stm);
      
      return;
    }

  /* add timeout for pulsing */
  gtk_timeout_add(25, gj_gtk_cp_progress_timeout, NULL);
  gj_gtk_cp_set_stop_timeout(FALSE);

  /* set state */
  gj_gtk_cp_set_state(gj_gtk_cp_state_connecting);
}

void gj_gtk_cp_state_handle_connecting(guint eventID)
{
  gchar *detail = NULL;
  gchar *jid = NULL;

  /* create jid */
  jid = g_strdup_printf("%s@%s%s%s", 
			gj_gtk_cp_get_username(), 
			gj_gtk_cp_get_server(), 
			gj_gtk_cp_get_resource()?"/":"",
			gj_gtk_cp_get_resource()?gj_gtk_cp_get_resource():"");

  /* set the fact that we are now connected */
  gj_me_set_is_connected(TRUE);
  gj_me_set_uptime_start();

  gj_me_set_connected_host(gj_gtk_cp_get_server());
  gj_me_set_connected_port(gj_gtk_cp_get_port());
  gj_me_set_jid(jid);

  /* show in main window */
  gj_gtk_roster_set_status(_("Connected."));

  gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_INFO,GTK_ICON_SIZE_DIALOG);
  gj_gtk_cp_set_header("<b>%s...</b>", _("Requesting Authentication Method"));
  gj_gtk_cp_set_detail("%s '%s'", _("username"), gj_gtk_cp_get_username());
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)2/4);
  g_message("gj_gtk_cp_state_handle_connecting: response received, eventID:%d",eventID);

  /* clean up */
  if(detail != NULL) g_free(detail);
  if(jid != NULL) g_free(jid);

  /* NOTE: no need to free the iq by event id because there was no iq
     created to generate this response! */

  /* next request */
  gj_iq_request_auth_method(gj_gtk_cp_get_username(), gj_gtk_cp_state_machine);

  /* set state */
  gj_gtk_cp_set_state(gj_gtk_cp_state_authentication_method_requested);
}

void gj_gtk_cp_state_handle_authentication_method_requested(guint eventID)
{
  gchar *detail = NULL;
  
  gchar *password_plain = NULL;
  gchar *password_digest = NULL;

  gchar *session_id = NULL;

  stream stm = NULL;

  if(eventID < 0)
    {
      gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gj_gtk_cp_set_header("<b>%s</b>", _("Request for Authentication Method Failed."));
      gj_gtk_cp_set_detail(_("Invalid Response."));
      g_warning("gj_gtk_cp_state_handle_authentication_method_requested: failed to continue with response to authorisation, eventID:%d was < 0",eventID);

      /* set state */
      gj_gtk_cp_set_state(gj_gtk_cp_state_idle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }

  detail = g_strdup_printf(_("username '%s'"),gj_gtk_cp_get_username());

  gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_INFO,GTK_ICON_SIZE_DIALOG);
  gj_gtk_cp_set_header("<b>%s</b>", _("Requesting Authentication..."));
  gj_gtk_cp_set_detail(detail);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)3/4);
  g_message("gj_gtk_cp_state_handle_authentication_method_requested: received available authentication methods, eventID:%d",eventID);

  /* clean up */
  g_free(detail);

  /* find stream */
  if((stm = gj_stream_find_by_to(gj_gtk_cp_get_server())) == NULL)
    {
      gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gj_gtk_cp_set_header("<b>%s</b>", _("Request for Authentication Method Failed."));
      gj_gtk_cp_set_detail(_("Internal Error."));
      g_warning("gj_gtk_cp_state_handle_authentication_method_requested: failed to continue with response to authorisation, could not find stream for connection to '%s'",gj_gtk_cp_get_server());

      /* clean up */
      gj_iq_free_all_related_by_event_id(eventID);
  
      /* set state */
      gj_gtk_cp_set_state(gj_gtk_cp_state_idle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }

  /* get session id */
  session_id = gj_stream_get_id(stm);

  /* clean up */
  gj_iq_free_all_related_by_event_id(eventID);
 
  /* next request */
  if(session_id == NULL || gj_gtk_cs_get_use_plain_text_password() == TRUE)
    {
      /* use plain text */
      gj_iq_request_auth_with_plain_password(gj_gtk_cp_get_username(),
					     gj_gtk_cp_get_password(),
					     gj_gtk_cp_get_resource(),
					     gj_gtk_cp_state_machine);
    }
  else
    {
      /* use digest (SHA1) */
      password_plain = gj_gtk_cp_get_password();
      password_digest = gj_support_get_sha_by_session_id_and_password(session_id, password_plain);

      gj_iq_request_auth_with_digest_password(gj_gtk_cp_get_username(),
					      password_digest,
					      gj_gtk_cp_get_resource(),
					      gj_gtk_cp_state_machine);
      
      if(password_digest != NULL) g_free(password_digest);
    }

  /* set state */
  gj_gtk_cp_set_state(gj_gtk_cp_state_authenticating);
}

void gj_gtk_cp_state_handle_authenticating(guint eventID)
{
  infoquery iq = NULL;
  infoquery iq_related = NULL;

  guint eventIDRelated = 0;

  xmlChar *type = NULL;
  xmlNodePtr node = NULL;

  if((iq = gj_iq_find_by_event_id(eventID)) == NULL)
    {
      gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)4/4);
      gj_gtk_cp_set_header("<b>%s</b>", _("Failed to authenticate."));
      gj_gtk_cp_set_detail(_("Invalid response."));
    
      g_warning("gj_gtk_cp_state_handle_authenticating: failed to find IQ based on eventID:%d",eventID);

      /* clean up */
      gj_iq_free_all_related_by_event_id(eventID);

      /* set state */
      gj_gtk_cp_set_state(gj_gtk_cp_state_idle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }

  eventIDRelated = gj_iq_get_related_id(iq);
  if((iq_related = gj_iq_find_by_event_id(eventIDRelated)) == NULL)
    {      
      gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)4/4);
      gj_gtk_cp_set_header("<b>%s</b>", _("Failed to authenticate."));
      gj_gtk_cp_set_detail(_("Invalid data in related request."));

      g_warning("gj_gtk_cp_state_handle_authenticating: failed to find IQ based on eventIDRelated:%d",eventIDRelated);

      /* clean up */
      gj_iq_free_all_related_by_event_id(eventID);

      /* set state */
      gj_gtk_cp_set_state(gj_gtk_cp_state_idle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }

  node = gj_iq_get_xml(iq_related);
  type = gj_parser_find_by_attr_and_ret(node, (guchar*)"type");

  if(type != NULL && xmlStrcmp(type, (guchar*)"error") == 0)
    {
      gchar *detail = NULL;

      xmlChar *errorCode = NULL;
      xmlChar *errorMessage = NULL;

      if((node = gj_parser_find_by_node(node, (guchar*)"error")) == NULL)
	{
	  g_warning("gj_gtk_cp_state_handle_authenticating: failed to find node 'error'");

	  /* clean up */
	  gj_iq_free_all_related_by_event_id(eventID);

	  /* set state */
	  gj_gtk_cp_set_state(gj_gtk_cp_state_idle);

	  /* finish */
	  gj_gtk_cp_complete(FALSE);

	  return;
	}

      if((errorCode = gj_parser_find_by_attr_and_ret(node, (guchar*)"code")) != NULL)
	detail = g_strdup_printf("%s:%s", _("Error"), errorCode);

      if((errorMessage = gj_parser_find_by_node_and_ret(node, (guchar*)"error")) != NULL) 
	if(detail != NULL) detail = g_strdup_printf("%s - %s",detail,errorMessage);

      gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)4/4);
      gj_gtk_cp_set_header("<b>%s</b>", _("Failed to authenticate."));
      gj_gtk_cp_set_detail(detail);
      if(detail != NULL) g_free(detail);

      g_warning("gj_gtk_cp_state_handle_authenticating: IQ type is error, authentication failed for eventID:%d",eventID);

      /* clean up */
      gj_iq_free_all_related_by_event_id(eventID);

      /* set state */
      gj_gtk_cp_set_state(gj_gtk_cp_state_idle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }
  
  gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_INFO,GTK_ICON_SIZE_DIALOG);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)4/4);
  gj_gtk_cp_set_header("<b>%s</b>", _("Received Authentication."));
  gj_gtk_cp_set_detail(_("Cleaning up..."));

  g_message("gj_gtk_cp_state_handle_authenticating: received authentication response, eventID:%d",eventID);

  /* clean up */
  gj_iq_free_all_related_by_event_id(eventID);

  /* set state */
  gj_gtk_cp_set_state(gj_gtk_cp_state_ready);

  /* finish */
  gj_gtk_cp_complete(TRUE);
}

void gj_gtk_cp_state_handle_ready(guint eventID)
{
}

void gj_gtk_cp_state_handle_disconnected(guint eventID)
{
  gj_gtk_cp_callback callback = NULL;

  /* set state */
  gj_me_set_is_connected(FALSE);
  gj_me_set_uptime_stop();

  /* call callback */
  if((callback = gj_gtk_cp_get_cb_disconnected()) != NULL)
    (callback)();

  /* should we reconnect ? */
  if(gj_gtk_cs_get_auto_reconnect_on_disconnect() == TRUE && gj_gtk_cs_get_save_password() == TRUE)
    {
#if 0
      /* get started ! */
      gj_gtk_cp_set_state(gj_gtk_cp_state_idle);
      gj_gtk_cp_state_machine(0);
#else
      gj_gtk_cp_reload();
#endif
    }
  else
    gj_gtk_cp_unload();
}

void gj_gtk_cp_connection_start(stream stm, gint eventID)
{
  gj_gtk_cp_state_machine(eventID);
}

void gj_gtk_cp_connection_stop(stream stm, gint eventID)
{
  gj_gtk_cp_set_state(gj_gtk_cp_state_disconnected);
  gj_gtk_cp_state_machine(eventID);

  if(stm != NULL)
    gj_stream_free(stm);
}

void gj_gtk_cp_connection_error(stream stm, gint eventID)
{
  /* display error */
  gtk_image_set_from_stock(GTK_IMAGE(ctlCP.image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctlCP.progressbar),(gdouble)4/4);
  gj_gtk_cp_set_header("<b>%s</b>", _("Failed to connect."));
  gj_gtk_cp_set_detail((gchar*)gj_connection_connect_translate_status(eventID));
  
  /* clean up */
  gj_gtk_cp_connection_stop(stm, eventID);
}

void gj_gtk_cp_complete(gboolean success)
{
  gj_gtk_cp_callback callback = NULL;
  
  /* call callback */
  if(success == TRUE)
    {
      if((callback = gj_gtk_cp_get_cb_connected()) != NULL)
	(callback)();

      /* close the window */
      gj_gtk_cp_unload();
    }
  else
    {
      if((callback = gj_gtk_cp_get_cb_disconnected()) != NULL)
	(callback)();
    }
}

gboolean gj_gtk_cp_progress_timeout(gpointer data)
{
  if(gj_gtk_cp_get_state() == gj_gtk_cp_state_authenticating ||
     gj_gtk_cp_get_stop_timeout() == TRUE)
    return FALSE;

  gtk_progress_bar_pulse(GTK_PROGRESS_BAR(ctlCP.progressbar));
  
  /* As this is a timeout function, return TRUE so that it
   * continues to get called */
  return TRUE;
} 


/*
 * CONNECTION window
 */
gboolean gj_gtk_cp_reload()
{
  return gj_gtk_cp_load(gj_gtk_cp_get_cb_connected(), gj_gtk_cp_get_cb_disconnected());
}

gboolean gj_gtk_cp_load(gj_gtk_cp_callback cb_connected,
			gj_gtk_cp_callback cb_disconnected)
{
  GladeXML *xml = NULL;

  if(ctlCP.window != NULL)
    {
      gtk_window_present(GTK_WINDOW(ctlCP.window));
      return TRUE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "connection_progress",
			  NULL)) == NULL)
    {
      g_warning("gj_gtk_cp_load: couldn't open glade file at root 'connection_progress'");
      return FALSE;
    }
  
  ctlCP.window = glade_xml_get_widget(xml,"connection_progress");

  ctlCP.button_action = glade_xml_get_widget(xml,"button_cp_action");

  ctlCP.label_desc1 = glade_xml_get_widget(xml,"label_cp_progress1");
  ctlCP.label_desc2 = glade_xml_get_widget(xml,"label_cp_progress2");

  ctlCP.image = glade_xml_get_widget(xml,"image_cp_icon");

  ctlCP.progressbar = glade_xml_get_widget(xml,"progressbar_cp_progress");

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  g_signal_connect ((gpointer) ctlCP.window, "delete_event",
		    G_CALLBACK (on_connection_progress_delete_event),
		    NULL);
  g_signal_connect ((gpointer) ctlCP.button_action, "clicked",
		    G_CALLBACK (on_button_cp_action_clicked),
		    NULL);
#endif

  /* get started ! */
  gj_gtk_cp_set_state(gj_gtk_cp_state_idle);
  gj_gtk_cp_set_cb_connected(cb_connected);
  gj_gtk_cp_set_cb_disconnected(cb_disconnected);

  gj_gtk_cp_set_server(gj_gtk_cs_get_server());
  gj_gtk_cp_set_port(gj_gtk_cs_get_port());
  gj_gtk_cp_set_username(gj_gtk_cs_get_username());
  gj_gtk_cp_set_password(gj_gtk_cs_get_password());
  gj_gtk_cp_set_resource(gj_gtk_cs_get_resource());

  if(gj_gtk_cs_get_save_password() == FALSE)
    {
      /* get password */
      gj_gtk_pr_load(FALSE, NULL, gj_gtk_cs_enter_password_cb);
    }
  else
    {
      g_message("gj_gtk_cp_load: auto connecting...");
      gj_gtk_cp_show();
      gj_gtk_cp_state_machine(0);
    }

  
  /* clean up */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
} 

gboolean gj_gtk_cp_unload()
{
  gtk_widget_destroy(GTK_WIDGET(ctlCP.window));
  ctlCP.window = NULL;

  if(optCP.server != NULL)
    {
      g_free(optCP.server);
      optCP.server = NULL;
    }

  optCP.port = 0;

  if(optCP.username != NULL)
    {
      g_free(optCP.username);
      optCP.username = NULL;
    }

  if(optCP.password != NULL)
    {
      g_free(optCP.password);
      optCP.password = NULL;
    }

  gj_gtk_cp_set_stop_timeout(TRUE);

  return TRUE;
}

gboolean gj_gtk_cp_hide()
{
  gtk_widget_hide(GTK_WIDGET(ctlCP.window));
  return TRUE;
}

gboolean gj_gtk_cp_show()
{
  gtk_widget_show(GTK_WIDGET(ctlCP.window));

  /* if we do it on load, the window is hidden, and we get warnings */
  gdk_window_set_decorations (GDK_WINDOW(ctlCP.window->window), 
			      GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU);

  return TRUE;
}

void gj_gtk_cp_set_header(gchar *message, ...)
{
  va_list args;
  gchar *buffer = NULL;

  va_start(args,message);
  buffer = g_strdup_vprintf(message,args);
  va_end(args);
 
  if(buffer == NULL)
    {
      g_warning("%s: buffer to use was NULL", __FUNCTION__);
      return;
    }
  
  gtk_label_set_markup(GTK_LABEL(ctlCP.label_desc1), buffer); 
}

void gj_gtk_cp_set_detail(gchar *message, ...)
{
  va_list args;
  gchar *buffer = NULL;

  va_start(args,message);
  buffer = g_strdup_vprintf(message,args);
  va_end(args);
 
  if(buffer == NULL)
    {
      g_warning("%s: buffer to use was NULL", __FUNCTION__);
      return;
    }
  
  gtk_label_set_markup(GTK_LABEL(ctlCP.label_desc2), buffer); 
}

void gj_gtk_cs_enter_password_cb(gboolean success, const gchar *password)
{
  /* this could be because:
     - user closed the dialog
     - user clicked ok, and the password wrong,
     - user clicked cancel     */
  if(success == FALSE)
    {
      g_message("gj_gtk_cs_enter_password_cb: password dialog cancelled, not progressing further.");
      gj_gtk_cp_unload();
      return;
    }

  /* set password */
  gj_gtk_cp_set_password((gchar*)password);

  /* start the state machine */
  gj_gtk_cp_show();
  gj_gtk_cp_state_machine(0);
}

/*
 * GUI events
 */
void on_button_cp_action_clicked(GtkButton *button, gpointer user_data)
{
#if 1
  /* NOTE: FIX ME */
  if(gnome_jabber_term_modules() == FALSE)
    g_warning("gj_gtk_cp_action: gnome_jabber_term returned FALSE");
  
  if(gnome_jabber_init_modules() == FALSE)
    g_warning("gj_gtk_cp_action: gnome_jabber_init returned FALSE");  
#endif

#if 0
  if(gj_gtk_cp_get_server() != NULL && gj_gtk_cp_get_port() > 0)
    {
      stream stm = NULL;

      if((stm = streamFindByTo(gj_gtk_cp_get_server())) != NULL)
	streamFree(stm);
    }
#endif

  gj_gtk_cp_unload();
}

gboolean on_connection_progress_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
  /* return TRUE so it doesnt close */
  return TRUE;
}


#ifdef __cplusplus
}
#endif
 
