/* 
 * $RCSfile: gj_gtk_message.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/31 16:43:54 $
 * 
 * $Revision: 1.20 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <glade/glade.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_GNOME
#include <gnome.h>
#endif

#include "common.h"

#include "gj_main.h"
#include "gj_roster.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_translate.h"
#include "gj_event.h"
#include "gj_support.h"
#include "gj_config.h"

#include "gj_gtk_roster.h"
#include "gj_gtk_message.h" 
#include "gj_gtk_my_preferences.h"
#include "gj_gtk_support.h"

struct t_controls
{
  GtkWidget *window;

  GtkWidget *entry_contact;
  GtkWidget *entry_time;
  GtkWidget *entry_subject;
  GtkWidget *textview;
  
  GtkWidget *label_subject;
  GtkWidget *label_contact;
  GtkWidget *label_next;

  GtkWidget *button_next;
  GtkWidget *button_send;
  GtkWidget *button_reply;
  GtkWidget *button_close;

};

struct t_msgWindow
{
  struct t_controls controls;

  gint ri_id;

  /* properties */
  msgWindowType type;

  gint last_msg_recv_xevent_type;
  gchar *last_msg_recv_xevent_id;

  gint last_msg_sent_xevent_type;
  gchar *last_msg_sent_xevent_id;
};


GList *msg_window_list = NULL;


GdkPixbuf *gj_gtk_mw_load_emoticon(msgEmote icon);

gboolean gj_gtk_mw_load(msgWindow mw);
gboolean gj_gtk_mw_unload(msgWindow mw);
gboolean gj_gtk_mw_unload_by_widget(GtkWidget *widget);

void gj_gtk_mw_raise();

/* list functions */
gboolean gj_gtk_mw_list_add(msgWindow mw);
gboolean gj_gtk_mw_list_del(msgWindow mw);
GList *gj_gtk_mw_list_get();

/* window functions */
msgWindow gj_gtk_mw_new(rosterItem ri);
gboolean gj_gtk_mw_free(msgWindow mw);
void gj_gtk_mw_destroy(msgWindow mw);

/* utilities */
gint gj_gtk_mw_waiting_count(msgWindow mw);
void gj_gtk_mw_waiting_update(msgWindow mw);

gboolean gj_gtk_mw_send_set(msgWindow mw);
gboolean gj_gtk_mw_receive_set(msgWindow mw, message msg);

gboolean gj_gtk_mw_check_for_url(GtkTextBuffer *buffer, gchar *text, guint from_pos);

/* gui callbacks */
void on_textbuffer_mw_changed(GtkTextBuffer *textbuffer, gpointer user_data);
gboolean on_textview_mw_motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data);
gboolean on_textview_mw_find_space(gunichar ch, gpointer user_data);
gboolean on_textview_mw_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);

void on_button_mw_send_clicked(GtkButton *button, gpointer user_data);
void on_button_mw_reply_clicked(GtkButton *button, gpointer user_data);
void on_button_mw_next_clicked(GtkButton *button, gpointer user_data);
void on_button_mw_close_clicked(GtkButton *button, gpointer user_data);

gboolean on_message_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);

/*
 * code
 */

/* list handling */
gboolean gj_gtk_mw_list_add(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  msg_window_list = g_list_append(msg_window_list,mw);
  return TRUE;
}

gboolean gj_gtk_mw_list_del(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  msg_window_list = g_list_remove(msg_window_list,mw);
  return TRUE;
}

GList *gj_gtk_mw_list_get()
{
  return msg_window_list;
}

msgWindow gj_gtk_mw_new(rosterItem ri)
{
  msgWindow mw = NULL;

  if(ri == NULL)
    {
      g_warning("%s: remote JID was NULL", __FUNCTION__);
      return NULL;
    }

  if((mw = (msgWindow) g_new(struct t_msgWindow,1)) != NULL)
    {
      mw->ri_id = gj_roster_item_get_id(ri);

      mw->type = 0;

      mw->last_msg_recv_xevent_type = 0;
      mw->last_msg_recv_xevent_id = NULL;
     
      mw->last_msg_sent_xevent_type = 0;
      mw->last_msg_sent_xevent_id = NULL;

      gj_gtk_mw_list_add(mw);
    }

  return mw;
}

gboolean gj_gtk_mw_free(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  /* free from list */
  gj_gtk_mw_list_del(mw);
  
  /* destroy members */
  gj_gtk_mw_destroy(mw);

  /* free struct */
  g_free(mw);
  
  return TRUE;
}

void gj_gtk_mw_destroy(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return;
    }

  mw->ri_id = 0;

  mw->type = 0;

  mw->last_msg_recv_xevent_type = 0;

  if(mw->last_msg_recv_xevent_id != NULL)
    {
      g_free(mw->last_msg_recv_xevent_id);
      mw->last_msg_recv_xevent_id = NULL;
    }

  mw->last_msg_sent_xevent_type = 0;

  if(mw->last_msg_sent_xevent_id != NULL)
    {
      g_free(mw->last_msg_sent_xevent_id);
      mw->last_msg_sent_xevent_id = NULL;
    }
}

msgWindow gj_gtk_mw_find_by_ri_id(gint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: id:%d was < 0", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(msg_window_list);index++)
    {
      msgWindow mw = g_list_nth_data(msg_window_list, index);

      if(mw == NULL) continue;
      if(mw->ri_id == id) return mw;
    }

  return NULL;  
}

msgWindow gj_gtk_mw_find_by_widget(GtkWidget *widget)
{
  gint index = 0;

  GtkWidget *toplevel = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return NULL;
    }

  /* get top level widget */
  toplevel = gtk_widget_get_toplevel(widget);

  if(toplevel == NULL)
    {
      g_warning("%s: top level widget was NULL or it has no top level widget", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(msg_window_list);index++)
    {
      msgWindow mw = g_list_nth_data(msg_window_list, index);

      if(mw == NULL) continue;
      if(mw->controls.window == toplevel) return mw;
    }

  return NULL;  
}

msgWindow gj_gtk_mw_find_by_ri_id_and_type(gint id, msgWindowType type)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: id:%d was < 0", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(msg_window_list);index++)
    {
      msgWindow mw = g_list_nth_data(msg_window_list, index);

      if(mw == NULL) continue;
      if(mw->ri_id == id && mw->type == type) return mw;
    }

  return NULL;  
}

msgWindow gj_gtk_mw_find_by_last_msg_recv_xevent_id(gchar *message_id)
{
  gint index = 0;

  if(message_id == NULL)
    {
      g_warning("%s: message id was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(msg_window_list);index++)
    {
      msgWindow mw = g_list_nth_data(msg_window_list, index);

      if(mw == NULL) continue;
      if(gj_gtk_mw_get_last_msg_recv_xevent_id(mw) != NULL && 
	 strcmp(gj_gtk_mw_get_last_msg_recv_xevent_id(mw), message_id) == 0)
	return mw;
    }

  return NULL;  
}

msgWindow gj_gtk_mw_find_by_last_msg_sent_xevent_id(gchar *message_id)
{
  gint index = 0;

  if(message_id == NULL)
    {
      g_warning("%s: message id was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(msg_window_list);index++)
    {
      msgWindow mw = g_list_nth_data(msg_window_list, index);

      if(mw == NULL) continue;
      if(gj_gtk_mw_get_last_msg_sent_xevent_id(mw) != NULL && 
	 strcmp(gj_gtk_mw_get_last_msg_sent_xevent_id(mw), message_id) == 0)
	return mw;
    }

  return NULL;  
}

/*
 * member functions
 */

/* gets */
gint gj_gtk_mw_get_ri_id(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return -1;
    }

  return mw->ri_id;
}

msgWindowType gj_gtk_mw_get_type(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return msgWindowTypeEnd;
    }

  return mw->type;
}

gint gj_gtk_mw_get_last_msg_recv_xevent_type(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return 0;
    }

  return mw->last_msg_recv_xevent_type;
}

gchar *gj_gtk_mw_get_last_msg_recv_xevent_id(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return mw->last_msg_recv_xevent_id;
}

gint gj_gtk_mw_get_last_msg_sent_xevent_type(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return 0;
    }

  return mw->last_msg_sent_xevent_type;
}

gchar *gj_gtk_mw_get_last_msg_sent_xevent_id(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return mw->last_msg_sent_xevent_id;
}

/* sets */
gboolean gj_gtk_mw_set_ri_id(msgWindow mw, gint id)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(id < 0)
    {
      g_warning("%s: id:%d was < 0", __FUNCTION__, id);
      return FALSE;
    }

  mw->ri_id = id;
  return TRUE;
}

gboolean gj_gtk_mw_set_type(msgWindow mw, msgWindowType type)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(type < 1 || type > msgWindowTypeEnd)
    {
      g_warning("%s: type%d was out of bounds", __FUNCTION__, type);
      return FALSE;
    }

  mw->type = type;
  return TRUE;
}

gboolean gj_gtk_mw_set_last_msg_recv_xevent_type(msgWindow mw, gint type)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  mw->last_msg_recv_xevent_type = type;
  return TRUE;
}

gboolean gj_gtk_mw_set_last_msg_recv_xevent_id(msgWindow mw, gchar *id)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(mw->last_msg_recv_xevent_id != NULL)
    {
      g_free(mw->last_msg_recv_xevent_id);
      mw->last_msg_recv_xevent_id = NULL;
    }

  mw->last_msg_recv_xevent_id = g_strdup(id);
  return TRUE;
}

gboolean gj_gtk_mw_set_last_msg_sent_xevent_type(msgWindow mw, gint type)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  mw->last_msg_sent_xevent_type = type;
  return TRUE;
}

gboolean gj_gtk_mw_set_last_msg_sent_xevent_id(msgWindow mw, gchar *id)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(mw->last_msg_sent_xevent_id != NULL)
    {
      g_free(mw->last_msg_sent_xevent_id);
      mw->last_msg_sent_xevent_id = NULL;
    }

  mw->last_msg_sent_xevent_id = g_strdup(id);
  return TRUE;
}

void gj_gtk_mw_print_all()
{
  gint index = 0;

  /* loop through all contacts on roster */
  for(index=0;index<g_list_length(msg_window_list);index++)
    {
      msgWindow mw = g_list_nth_data(msg_window_list, index);
      if(mw == NULL) continue;

      gj_gtk_mw_print(mw);
    }  
}

void gj_gtk_mw_print(msgWindow mw)
{
  if(mw == NULL)
    return;

  g_message("++++++++++++++++++++++++++++++++++++++++++++++++");
  g_message("++++");
  g_message("++++ roster item id:%d", mw->ri_id);
  g_message("++++ type:'%d'", mw->type);
  g_message("++++");
  g_message("++++ last_msg_recv_xevent_type:%d", mw->last_msg_recv_xevent_type);
  g_message("++++ last_msg_recv_xevent_id:%s", mw->last_msg_recv_xevent_id);
  g_message("++++ last_msg_sent_xevent_type:%d", mw->last_msg_sent_xevent_type);
  g_message("++++ last_msg_sent_xevent_id:%s", mw->last_msg_sent_xevent_id);
  g_message("++++");
  g_message("++++++++++++++++++++++++++++++++++++++++++++++++");
}


/* 
 * window functions
 */
gboolean gj_gtk_mw_show(msgWindow mw, gboolean show)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(show == TRUE)
    gtk_widget_show(mw->controls.window);
  else
    gtk_widget_hide(mw->controls.window);

  return TRUE;
}

void gj_gtk_mw_raise(msgWindow mw)
{
  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return;
    }

  if(mw->controls.window == NULL || mw->controls.window->window == NULL)
    {
      g_warning("%s: window was NULL or GDK window was NULL", __FUNCTION__);
      return;
    }

  gdk_window_raise(GDK_WINDOW(mw->controls.window->window));
}

gboolean gj_gtk_mw_load(msgWindow mw)
{
  GladeXML *xml;
  GdkPixbuf *pb = NULL;
  GtkTextBuffer *buffer = NULL;

  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "message",
			  NULL)) == NULL)
    {
      g_warning("%s: couldn't open glade file at root 'message'", __FUNCTION__);
      return FALSE;
    }
  
  mw->controls.window = glade_xml_get_widget(xml,"message");

  mw->controls.entry_contact = glade_xml_get_widget(xml, "entry_mw_contact");
  mw->controls.entry_time = glade_xml_get_widget(xml, "entry_mw_time");
  mw->controls.entry_subject = glade_xml_get_widget(xml, "entry_mw_subject");
  mw->controls.textview = glade_xml_get_widget(xml, "textview_mw");

  mw->controls.label_subject = glade_xml_get_widget(xml, "label_mw_subject");
  mw->controls.label_contact = glade_xml_get_widget(xml, "label_mw_contact");
  mw->controls.label_next = glade_xml_get_widget(xml, "label_mw_next");

  mw->controls.button_next = glade_xml_get_widget(xml, "button_mw_next");
  mw->controls.button_send = glade_xml_get_widget(xml, "button_mw_send");
  mw->controls.button_reply = glade_xml_get_widget(xml, "button_mw_reply");
  mw->controls.button_close = glade_xml_get_widget(xml, "button_mw_close");
  
  gdk_window_set_decorations (GDK_WINDOW(mw->controls.window->window), 
			      GDK_DECOR_ALL);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  gtk_widget_set_events(GTK_WIDGET(mw->controls.textview),
			GDK_EXPOSURE_MASK | 
			GDK_POINTER_MOTION_MASK | 
			GDK_POINTER_MOTION_HINT_MASK | 
			GDK_BUTTON_RELEASE_MASK | 
			GDK_STRUCTURE_MASK | 
			GDK_PROPERTY_CHANGE_MASK);

  g_signal_connect ((gpointer) mw->controls.window, "delete_event",
		    G_CALLBACK (on_message_delete_event),
		    NULL);
  g_signal_connect ((gpointer) mw->controls.textview, "motion_notify_event",
		    G_CALLBACK (on_textview_mw_motion_notify_event),
		    NULL);
  g_signal_connect ((gpointer) mw->controls.textview, "button_press_event",
		    G_CALLBACK (on_textview_mw_button_press_event),
		    NULL);
  g_signal_connect ((gpointer) mw->controls.button_reply, "clicked",
		    G_CALLBACK (on_button_mw_reply_clicked),
		    NULL);
  g_signal_connect ((gpointer) mw->controls.button_send, "clicked",
		    G_CALLBACK (on_button_mw_send_clicked),
		    NULL);
  g_signal_connect ((gpointer) mw->controls.button_next, "clicked",
		    G_CALLBACK (on_button_mw_next_clicked),
		    NULL);
  g_signal_connect ((gpointer) mw->controls.button_close, "clicked",
		    G_CALLBACK (on_button_mw_close_clicked),
		    NULL);
#endif

  /* set icon */
  pb = gj_gtk_support_pixbuf_new_from_file(FILE_MESSAGE_NORMAL);
  gtk_window_set_icon(GTK_WINDOW(mw->controls.window), pb);
  g_object_unref(G_OBJECT(pb));

  /* get buffer */
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(mw->controls.textview));

  /* set function callbacks */
  g_signal_connect(G_OBJECT(buffer), "changed", 
		   G_CALLBACK(on_textbuffer_mw_changed), 
		   G_OBJECT(mw->controls.textview));

  /* set styles */
  gtk_text_buffer_create_tag(buffer, "url", 
			     "foreground", "blue", 
			     NULL); 

  /* 
   * cleanup 
   */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_mw_unload(msgWindow mw)
{
  rosterItem ri = NULL;

  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_mw_get_ri_id(mw))) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  /* if no more msgs reset icon */
  gtk_widget_destroy(mw->controls.window);
  
  gj_gtk_roster_handle_next_event_display(gj_roster_item_get_jid(ri));
  gj_gtk_mw_free(mw);
      
  g_message("%s: unloaded...", __FUNCTION__);
  
  return TRUE;
}

gboolean gj_gtk_mw_unload_by_widget(GtkWidget *widget)
{
  msgWindow mw = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return FALSE;
    }

  if((mw = gj_gtk_mw_find_by_widget(widget)) != NULL)
    {
      gj_gtk_mw_unload(mw);
    }
  else
    g_warning("%s: could not find msgWindow from widget", __FUNCTION__);
  
  return TRUE;
}

msgWindow gj_gtk_mw_send_new(rosterItem ri)
{
  msgWindow mw = NULL;
  gchar *tmp = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  mw = gj_gtk_mw_new(ri);
  gj_gtk_mw_load(mw);
  
  /* set type */
  gj_gtk_mw_set_type(mw, msgWindowTypeOutgoing);

  /* set title */
  tmp = g_strdup_printf(_("Send Message"));
  gtk_window_set_title(GTK_WINDOW(mw->controls.window),tmp);
  g_free(tmp);

  /* set details */
  tmp = g_strdup_printf("%s:", _("To"));
  gtk_label_set_markup(GTK_LABEL(mw->controls.label_contact),tmp);
  g_free(tmp);

  /* set properies */
  gtk_widget_show(GTK_WIDGET(mw->controls.button_send));
  gtk_widget_hide(GTK_WIDGET(mw->controls.button_reply));
  gtk_widget_hide(GTK_WIDGET(mw->controls.button_next));
  gtk_editable_set_editable(GTK_EDITABLE(mw->controls.entry_subject),TRUE);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(mw->controls.textview), TRUE);
  gtk_widget_set_sensitive(GTK_WIDGET(mw->controls.button_send), FALSE);

  /* set specifics */
  gj_gtk_mw_send_set(mw);

  return mw;
}

msgWindow gj_gtk_mw_send_new_by_jid(gchar *jid)
{
  msgWindow mw = NULL;
  rosterItem ri = NULL;  
  gchar *tmp = NULL;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  if((ri = gj_roster_item_find_by_jid(jid)) == NULL)
    {
      g_message("%s: no roster item for jid:'%s', creating temporary solution...", __FUNCTION__, jid);
      
      /* FIXME: this is not freed EVER - when should it be? */
      if((ri = gj_roster_item_new_unknown_by_jid(jid)) == NULL)
	{
	  g_warning("%s: could not create new temporary roster item", __FUNCTION__);
	  return NULL;
	}
    }

  mw = gj_gtk_mw_new(ri);
  gj_gtk_mw_load(mw);
  
  /* set type */
  gj_gtk_mw_set_type(mw, msgWindowTypeOutgoing);

  /* set title */
  tmp = g_strdup_printf("Send Message");
  gtk_window_set_title(GTK_WINDOW(mw->controls.window),tmp);
  g_free(tmp);

  /* set details */
  tmp = g_strdup_printf("To:");
  gtk_label_set_markup(GTK_LABEL(mw->controls.label_contact),tmp);
  g_free(tmp);

  /* set properies */
  gtk_widget_show(GTK_WIDGET(mw->controls.button_send));
  gtk_widget_hide(GTK_WIDGET(mw->controls.button_reply));
  gtk_widget_hide(GTK_WIDGET(mw->controls.button_next));
  gtk_editable_set_editable(GTK_EDITABLE(mw->controls.entry_subject),TRUE);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(mw->controls.textview), TRUE);
  gtk_widget_set_sensitive(GTK_WIDGET(mw->controls.button_send), FALSE);

  /* set specifics */
  gj_gtk_mw_send_set(mw);

  return mw;
}

msgWindow gj_gtk_mw_receive_new(rosterItem ri)
{
  msgWindow mw = NULL;
  gchar *tmp = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  mw = gj_gtk_mw_new(ri);
  gj_gtk_mw_load(mw);
  
  /* set type */
  gj_gtk_mw_set_type(mw, msgWindowTypeIncoming);

  /* set title */
  tmp = g_strdup_printf(_("Message Received"));
  gtk_window_set_title(GTK_WINDOW(mw->controls.window),tmp);
  g_free(tmp);

  /* set details */
  tmp = g_strdup_printf("%s:", _("From"));
  gtk_label_set_markup(GTK_LABEL(mw->controls.label_contact),tmp);
  g_free(tmp);

  /* set properies */
  gtk_widget_show(GTK_WIDGET(mw->controls.button_reply));
  gtk_widget_hide(GTK_WIDGET(mw->controls.button_send));
  /*  gtk_widget_set_sensitive(mw->controls.button_next,FALSE);*/
  gtk_editable_set_editable(GTK_EDITABLE(mw->controls.entry_subject),FALSE);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(mw->controls.textview), FALSE);

  return mw;
}

gboolean gj_gtk_mw_receive_set(msgWindow mw, message msg)
{
  gchar *time = NULL;
  gchar *message = NULL;

  /*!!!!!! check mw and msg */
  time = g_strdup(gj_support_get_timestamp_from_timet(gj_message_get_time(msg)));
  /*  time = timestampByTimeT(messageGetTime(msg));*/

  if(gj_message_get_from_jid(msg) != NULL && g_utf8_strlen(gj_message_get_from_jid(msg), -1) > 0)
      gtk_entry_set_text(GTK_ENTRY(mw->controls.entry_contact), gj_message_get_from_jid(msg));

  if(time != NULL && g_utf8_strlen(time, -1) > 0)
      gtk_entry_set_text(GTK_ENTRY(mw->controls.entry_time), time);

  if(gj_message_get_subject(msg) != NULL && g_utf8_strlen(gj_message_get_subject(msg), -1) > 0)
    {
      gtk_entry_set_text(GTK_ENTRY(mw->controls.entry_subject), gj_message_get_subject(msg));
      gtk_widget_show(mw->controls.label_subject);
      gtk_widget_show(mw->controls.entry_subject);
    }
  else
    {
      gtk_widget_hide(mw->controls.label_subject);
      gtk_widget_hide(mw->controls.entry_subject);
    }
      
  if((message = gj_message_get_body(msg)) != NULL && g_utf8_strlen(message, -1) > 0)
    {
      GtkTextBuffer *buffer = NULL;
      GtkTextIter iter;
      GdkPixbuf *pixbuf = NULL;

      guint i = 1;

      /* get buffer */
      buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(mw->controls.textview));
      gtk_text_buffer_set_text(buffer,"",-1);
      gtk_text_buffer_get_iter_at_offset(buffer, &iter, -1);

      /* search for occurences of emote icons */
      for(i=0;i<g_utf8_strlen(message, -1);i++)
	{
	    gchar *p = NULL;
	    p = g_utf8_offset_to_pointer(message, i);
	    
	  if(strncmp(message+i, EMOTE_BIGGRIN, g_utf8_strlen(EMOTE_BIGGRIN, -1)) == 0)
	    {
	      g_message("use bggrin");
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_biggrin)) != NULL)
		{
		  g_message("insert bggrin");
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_BIGGRIN, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_CONFUSED, g_utf8_strlen(EMOTE_CONFUSED, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_confused)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_CONFUSED, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_COOL, g_utf8_strlen(EMOTE_COOL, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_cool)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_COOL, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_EEK, g_utf8_strlen(EMOTE_EEK, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_eek)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_EEK, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_FROWN, g_utf8_strlen(EMOTE_FROWN, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_frown)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_FROWN, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_MAD, g_utf8_strlen(EMOTE_MAD, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_mad)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_MAD, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_REDFACE, g_utf8_strlen(EMOTE_REDFACE, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_redface)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_REDFACE, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_ROLLEYES, g_utf8_strlen(EMOTE_ROLLEYES, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_rolleyes)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_ROLLEYES, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_SMILE, g_utf8_strlen(EMOTE_SMILE, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_smile)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_SMILE, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_TONGUE, g_utf8_strlen(EMOTE_TONGUE, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_tongue)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_TONGUE, -1)-1);
	    }
	  else if(strncmp(message+i, EMOTE_WINK, g_utf8_strlen(EMOTE_WINK, -1)) == 0)
	    {
	      if((pixbuf = gj_gtk_mw_load_emoticon(emote_wink)) != NULL)
		{
		  gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
		  g_object_unref(G_OBJECT(pixbuf));
		}

	      i += (g_utf8_strlen(EMOTE_WINK, -1)-1);
	    }
	  else
	    {
		gchar *str = NULL;
		
		/* create mem - utf should be no longer than 6 bytes */
		str = g_new0(char, 10);
		
		/* copy char */
		str = g_utf8_strncpy(str, p, 1);
		
		/* insert text */
		gtk_text_buffer_insert(buffer, &iter, str, -1);

		if(str != NULL) g_free(str);
	    }
	}
    }
  
  /* clean up */
  g_free(time);

  return TRUE;
}

gboolean gj_gtk_mw_send_set(msgWindow mw)
{
  gchar *contact = NULL;
  gchar *time = NULL;

  rosterItem ri = NULL;

  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_mw_get_ri_id(mw))) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  contact = gj_roster_item_get_jid(ri);
  time = g_strdup(gj_support_get_timestamp());
 
  if(contact != NULL && g_utf8_strlen(contact, -1) > 0)
    gtk_entry_set_text(GTK_ENTRY(mw->controls.entry_contact), contact);

  if(time != NULL && g_utf8_strlen(time, -1) > 0)
    gtk_entry_set_text(GTK_ENTRY(mw->controls.entry_time), time);

  /* clean up */
  g_free(time);

  return TRUE;
}

void gj_gtk_mw_waiting_update(msgWindow mw)
{
  gint count = 0;
  gchar *label = NULL;

  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return;
    }

  /* get count waiting */
  if((count = gj_gtk_mw_waiting_count(mw)) < 0)
    {
      g_warning("%s: count:%d was < 0, could not get the count", __FUNCTION__, count);
      return;
    }
      
  /* change button to say "next (i)" */
  g_message("%s: updating number waiting to %d", __FUNCTION__, count);
  
  if(count > 1)
    {
      /* set next */
      label = g_strdup_printf("%s (<b>%d</b>)", _("Next"), count);
      gtk_label_set_markup(GTK_LABEL(mw->controls.label_next),label);
      g_free(label);
      
      /* enable it */
      gtk_widget_set_sensitive(mw->controls.button_next,TRUE);
    }
  else if(count > 0)
    {
      /* set next */
      label = g_strdup_printf(_("Next"));
      gtk_label_set_markup(GTK_LABEL(mw->controls.label_next),label);
      g_free(label);
      
      /* enable it */
      gtk_widget_set_sensitive(mw->controls.button_next,TRUE);
    }
  else 
    {
      gtk_widget_set_sensitive(mw->controls.button_next,FALSE);
    }
}

gint gj_gtk_mw_waiting_count(msgWindow mw)
{
  rosterItem ri = NULL;
  event ev = NULL;

  gint count = 0;
  gint index = 0;

  if(mw == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return -1;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_mw_get_ri_id(mw))) == NULL)
    {
	g_message("%s: could not find roster item by id:%d removing both resources and checking...", 
		  __FUNCTION__, gj_gtk_mw_get_ri_id(mw));

	return -1;
    }

  /* look for futher events of type 'chat' in queue */
  while((ev = gj_event_nth_by_roster_item(ri,index++)) != NULL)
    {
      message msg = NULL;
      
      if(gj_event_get_type(ev) != eventTypeMessage)
	continue;

      if((msg = (message) gj_event_get_data(ev)) == NULL)
	continue;
      
      if(gj_message_get_type(msg) == messageTypeNormal)
	count++;
    }
  
  /* should we?? take one off because the current event is also in this queue */
  return count;
}

gboolean gj_gtk_mw_waiting_next(msgWindow mw)
{
  message msg = NULL;
  rosterItem ri = NULL;
  event ev = NULL;
  
  gint eventID = 0;
  gint count = 0;
  gint index = 0;

  if(mw == NULL)
    {
      g_warning("%s: mw was NULL", __FUNCTION__);
      return FALSE;
    }

  if((count = gj_gtk_mw_waiting_count(mw)) < 1)
    {
      g_warning("%s: count:%d was < 1", __FUNCTION__, count);
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_mw_get_ri_id(mw))) == NULL)
    {
      g_message("%s: failed to get roster item by id:%d", __FUNCTION__, gj_gtk_mw_get_ri_id(mw));
      return FALSE;
    }

  /* look for futher events of type 'chat' in queue */
  while((ev = gj_event_nth_by_roster_item(ri,index++)) != NULL)
    {
      message msg = NULL;
      
      if(gj_event_get_type(ev) != eventTypeMessage) continue;
      if((msg = (message) gj_event_get_data(ev)) == NULL) continue;

      if(gj_message_get_type(msg) == messageTypeNormal)
	{
	  count++;
	  break;
	}
    }

  if(count == 0)
    {
      g_warning("%s: could not find next roster item event", __FUNCTION__);
      return FALSE;
    }

  msg = (message) gj_event_get_data(ev);
  eventID = gj_event_get_id(ev);
  
  if(msg == NULL)
    {
      g_warning("%s: msg data was NULL, not handling...", __FUNCTION__);
      
      /* free event - no use to me now */
      gj_event_free(ev);
      return FALSE;
    }
  
  g_message("%s: handling next event of type:%d, id:%d", __FUNCTION__, eventTypeMessage, eventID);
  
  /* setup msg */
  gj_gtk_mw_receive_set(mw,msg);	   
  
  /* free event */
  gj_event_free(ev);

  /* update next button */
  gj_gtk_mw_waiting_update(mw);

  return TRUE;
}

msgWindow gj_gtk_mw_notification_new(gint event_id, message msg)
{
  msgWindow mw = NULL;
  event ev = NULL;
  rosterItem ri = NULL;
  
  if(msg == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return NULL;
    }

  if((ev = gj_event_find_by_id(event_id)) == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return NULL;
    }

  if((ri = gj_event_get_roster_item(ev)) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  mw = gj_gtk_mw_find_by_ri_id_and_type(gj_roster_item_get_id(ri), msgWindowTypeIncoming);

  if(mw == NULL)
    {
      mw = gj_gtk_mw_receive_new(ri);
      gj_gtk_mw_receive_set(mw,msg);

      /* free event */
      gj_event_free_by_id(event_id);
      
      /* check messages still waiting and update display */
      gj_gtk_mw_waiting_update(mw);
    }
  else
    {
      if(gj_gtk_mw_waiting_count(mw) > 0)
	{
	  /* add to queue */
	  g_message("%s: there are %d messages now in the queue", __FUNCTION__, gj_gtk_mw_waiting_count(mw));
	  
	  /* 
	   * check other msg windows that are opened
	   *
	   */
	  gj_gtk_mw_waiting_update(mw);
	}

      /* should we show it? */
      if(gj_config_get_mpf_raise_window_on_new_message() == TRUE)
	gj_gtk_mw_raise(mw);
    }

  return mw;
}

void gj_gtk_mw_default_action(rosterItem ri)
{
  msgWindow mw = NULL;
  gint count = 0;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return;
    }

  if((mw = gj_gtk_mw_find_by_ri_id_and_type(gj_roster_item_get_id(ri), msgWindowTypeIncoming)) != NULL)
    {
      g_message("%s: found incoming msg window for jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri));

      if((count = gj_gtk_mw_waiting_count(mw)) > 0)
	{
	  /* 1 or more msgs waiting... */

	  /* show window mw */
	  gj_gtk_mw_show(mw,TRUE);
	  gj_gtk_mw_waiting_next(mw);
	}

#if 0      
      /* if XEvents ask to be notified when the message has 
	 been displayed, this is the time to send that notification */
      if(messageGetXEventType(msg) & messageXEventTypeDisplayed)
	messageSendXEvent(messageXEventTypeDisplayed, messageGetFromJID(msg), messageGetID(msg));

#endif

      return;
    }
  else if((mw = gj_gtk_mw_find_by_ri_id_and_type(gj_roster_item_get_id(ri), msgWindowTypeOutgoing)) != NULL)
    {
      /* show window */
      g_message("%s: found outgoing msg window for jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri));
      return;
    }
  else
    {
      msgWindow mw = NULL;

      g_message("%s: found NO incoming or outgoing msg window for jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri));

      /* if no messages waiting */
      mw = gj_gtk_mw_send_new(ri);
      gj_gtk_mw_show(mw,TRUE);
    }
}

GdkPixbuf *gj_gtk_mw_load_emoticon(msgEmote icon)
{
  GdkPixbuf *pixbuf = NULL;
  GdkPixbuf *scaled = NULL;
  gchar *filename = NULL; 

  switch(icon)
    {
    case emote_biggrin:
      filename = FILE_EMOTE_BIGGRIN;
      break;
    case emote_confused:
      filename = FILE_EMOTE_CONFUSED;
      break;
    case emote_cool:
      filename = FILE_EMOTE_COOL;
      break;
    case emote_eek:
      filename = FILE_EMOTE_EEK;
      break;
    case emote_frown:
      filename = FILE_EMOTE_FROWN;
      break;
    case emote_mad:
      filename = FILE_EMOTE_MAD;
      break;
    case emote_redface:
      filename = FILE_EMOTE_REDFACE;
      break;
    case emote_rolleyes:
      filename = FILE_EMOTE_ROLLEYES;
      break;
    case emote_smile:
      filename = FILE_EMOTE_SMILE;
      break;
    case emote_tongue:
      filename = FILE_EMOTE_TONGUE;
      break;
    case emote_wink:
      filename = FILE_EMOTE_WINK;
      break;
    default:
      g_warning("%s: failed to determin which icon to use from %d", __FUNCTION__, icon);
      return NULL;
      break;
    }

  if((pixbuf = gdk_pixbuf_new_from_file(filename, NULL)) == NULL)
    {
      g_warning("%s: failed to load image file:'%s'", __FUNCTION__, filename);
      return NULL;
    }

  scaled = gdk_pixbuf_scale_simple(pixbuf, 15, 15, GDK_INTERP_BILINEAR);
  g_object_unref(G_OBJECT(pixbuf));
  pixbuf = scaled;

  return pixbuf;
}

/*
 * GUI events
 */
gboolean gj_gtk_mw_check_for_url(GtkTextBuffer *buffer, gchar *text, guint pos)
{
  GtkTextIter iter_start;
  GtkTextIter iter_end;

  gchar *buf = NULL;

  guint start = 0;
  guint end = 0;
  guint length = 0;

  if(GTK_TEXT_BUFFER(buffer) == NULL) 
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }

  /* we use text passed to us so we dont have to RE-get
     the whole text in the textview buffer */
  if(text == NULL) 
    {
      g_warning("%s: text was NULL", __FUNCTION__);
      return FALSE;
    }
  else
    length = g_utf8_strlen(text, -1);

  /* check bounds */
  for(start = pos; start >= 0; --start) 
    if(gj_support_is_word_separator(text[start])) break;

  start++;
  
  for(end = start; end < length; end++) 
    if(gj_support_is_word_separator(text[end])) break;  

  /* check for quotes */
  for(;end>start; end--)
    if(isalpha(text[end-1]) || isdigit(text[end-1]) || text[end-1] == ':') break;

  /* get iters for start and end */
  gtk_text_buffer_get_iter_at_offset(buffer, &iter_start, start);
  gtk_text_buffer_get_iter_at_offset(buffer, &iter_end, end);

  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }

  if(gj_support_is_uri(buf) != uriTypeUnknown) 
    {
      /* highlight as uri */
      gtk_text_buffer_apply_tag_by_name(buffer, "url", &iter_start, &iter_end);
      return TRUE;
    } 
  else
    {
      /* none highlighted */
      gtk_text_buffer_remove_tag_by_name(buffer, "url", &iter_start, &iter_end);
      return TRUE;
    }

  return FALSE;
}

/*
 * GUI events
 */
void on_textbuffer_mw_changed(GtkTextBuffer *textbuffer, gpointer user_data)
{
  GtkTextMark *mark = NULL;
  GtkTextIter iter_space;
  GtkTextIter iter_word_start;
  GtkTextIter iter_word_end;
  GtkTextIter iter_start;
  GtkTextIter iter_end;

  gint length = 0;

  msgWindow mw = NULL;

  /* check WHOLE buffer for urls */
  gtk_text_buffer_get_bounds(textbuffer, &iter_start, &iter_end);
  length = gtk_text_buffer_get_char_count(textbuffer);

  if((mark = gtk_text_buffer_get_mark(textbuffer, "url")) != NULL)
    {
      gtk_text_buffer_get_iter_at_mark(textbuffer, &iter_start, mark);
      /*       gtk_text_mark_set_visible(mark, TRUE); */
    }

  if(user_data != NULL)
    {
      if((mw = gj_gtk_mw_find_by_widget(GTK_WIDGET(user_data))) == NULL)
	{
	  g_warning("%s: failed to find window from widget", __FUNCTION__);
	  return;
	}

      if(length > 0)
	gtk_widget_set_sensitive(GTK_WIDGET(mw->controls.button_send), TRUE);
      else
	gtk_widget_set_sensitive(GTK_WIDGET(mw->controls.button_send), FALSE);
    }

  /* save last iter */
  iter_word_start = iter_start;
  iter_word_end = iter_start;
  iter_space = iter_start;

  while(TRUE)
    {
      gboolean found = FALSE;
      gchar *buf = NULL;

      /* search for next space */
      found = gtk_text_iter_forward_find_char(&iter_space, on_textview_mw_find_space, NULL, NULL);

      if(found == FALSE)
	break;
      
      /* set word end to the space */
      iter_word_start = iter_word_end;
      iter_word_end = iter_space;

      /* move word end iter forward 1 */
      gtk_text_iter_forward_char(&iter_word_start);

      /* get */
      buf = gtk_text_iter_get_text(&iter_word_start, &iter_word_end);
      
      if(buf != NULL) 
	{
	  if(gj_support_is_uri(buf) != uriTypeUnknown) 
	    {
	      /* highlight as uri */
	      gtk_text_buffer_apply_tag_by_name(textbuffer, "url", &iter_word_start, &iter_word_end);
	    } 
	  else
	    {
	      /* none highlighted */
	      gtk_text_buffer_remove_tag_by_name(textbuffer, "url", &iter_word_start, &iter_word_end);
	    }

	  g_free(buf);
	}

      /* set mark, then we can check next time from 
	 the last mark and it is more efficient */
      if((mark = gtk_text_buffer_get_mark(textbuffer, "url")) == NULL)
	{
	  if((mark = gtk_text_buffer_create_mark(textbuffer, "url", &iter_space, FALSE)) == NULL)
	    {
	      g_warning("%s: could not create mark!", __FUNCTION__);
	      return;
	    }

	  /* no need to move it, just continue */
	  continue;
	}

      /* move mark to latest iter */
      gtk_text_buffer_move_mark_by_name(textbuffer, "url", &iter_space);
    }
}

gboolean on_textview_mw_motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
  GdkModifierType state = 0;
  GdkWindow *window = NULL;

  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter;
  GtkTextIter iter_start;
  GtkTextIter iter_end;
  GtkTextWindowType text_window_type = 0;
  
  gboolean found_start = FALSE;
  gboolean found_end = FALSE;

  gint pointer_x = 0;
  gint pointer_y = 0;
  gint x = 0;
  gint y = 0;

  gchar *buf = NULL;

  msgWindow mw = NULL;

  if(event->is_hint)
    {
      window = gdk_window_get_pointer(event->window, &pointer_x, &pointer_y, &state);
    }
  else
    {
      pointer_x = event->x;
      pointer_y = event->y;
      state = event->state;
    }

  if((mw = gj_gtk_mw_find_by_widget(GTK_WIDGET(widget))) == NULL)
    {
      g_warning("%s: failed to find msgWindow from widget", __FUNCTION__);
      return FALSE;
    }
  
  if((textview = GTK_TEXT_VIEW(widget)) == NULL)
    {
      g_warning("%s: view was NULL", __FUNCTION__);
      return FALSE;
    }

  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(textview))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  text_window_type = gtk_text_view_get_window_type(textview, event->window);
  gtk_text_view_window_to_buffer_coords(textview, text_window_type, pointer_x, pointer_y, &x, &y);
  gtk_text_view_get_iter_at_location(textview, &iter, x, y);
 
  if(g_unichar_isspace(gtk_text_iter_get_char(&iter)) == TRUE)
    {
      gdk_window_set_cursor(event->window, NULL);
      return FALSE;
    }

  iter_start = iter;
  iter_end = iter;

  /* search for next space */
  found_start = gtk_text_iter_backward_find_char(&iter_start, on_textview_mw_find_space, NULL, NULL);
  found_end = gtk_text_iter_forward_find_char(&iter_end, on_textview_mw_find_space, NULL, NULL);

  if(found_start == FALSE && found_end == FALSE)
    {
      gdk_window_set_cursor(event->window, NULL);
      return FALSE;
    }

  if(gtk_text_iter_is_start(&iter_start) == FALSE)
    gtk_text_iter_forward_char(&iter_start);
  
  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(gj_support_is_uri(buf) != uriTypeUnknown)
    {
      /* width wise it works... */
      GdkCursor *cursor = NULL;
      cursor = gdk_cursor_new(GDK_HAND2);
      gdk_window_set_cursor (event->window, cursor);
      
      /* clean up */
      gdk_cursor_unref(cursor);
    }  
  else
    {
      /* return to parent window's cursor */
      gdk_window_set_cursor(event->window, NULL);
    }
      
  /* clean up */
  if(buf != NULL) g_free(buf);

  return FALSE;
}

gboolean on_textview_mw_find_space(gunichar ch, gpointer user_data)
{
  if(g_unichar_isspace(ch) == TRUE)
    return TRUE;
    
  return FALSE;
}

gboolean on_textview_mw_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  GdkModifierType state;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter;
  GtkTextIter iter_start;
  GtkTextIter iter_end;
  GtkTextWindowType text_window_type = 0;

  gboolean found_start = FALSE;
  gboolean found_end = FALSE;

  gint pointer_x = 0;
  gint pointer_y = 0;
  gint x = 0;
  gint y = 0;

  gchar *buf = NULL;
    
  /* if not mouse 1 return */
  if(event->button != 1)
    return FALSE;

  gdk_window_get_pointer(event->window, &pointer_x, &pointer_y, &state);
    
  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget)))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
    
  text_window_type = gtk_text_view_get_window_type(GTK_TEXT_VIEW(widget), event->window);
  gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget), text_window_type, pointer_x, pointer_y, &x, &y);
  gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(widget), &iter, x, y);

  iter_start = iter;
  iter_end = iter;
    
  /* search for next space */
  found_start = gtk_text_iter_backward_find_char(&iter_start, on_textview_mw_find_space, NULL, NULL);
  found_end = gtk_text_iter_forward_find_char(&iter_end, on_textview_mw_find_space, NULL, NULL);
    
  if(found_start == FALSE && found_end == FALSE)
    return FALSE;
    
  if(gtk_text_iter_is_start(&iter_start) == FALSE)
    gtk_text_iter_forward_char(&iter_start);
    
  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
    
  if(gj_support_is_uri(buf) != uriTypeUnknown)
    gj_gtk_support_open_url(buf);
    
  /* clean up */
  if(buf != NULL) g_free(buf);
   
  return FALSE;
}

void on_button_mw_send_clicked(GtkButton *button, gpointer user_data)
{
  gchar *subject = NULL;
  gchar *body = NULL;

  msgWindow mw = NULL;
  rosterItem ri = NULL;

  GtkTextBuffer *buffer = NULL;
  GtkTextIter bufStart;
  GtkTextIter bufStop;

  if((mw = gj_gtk_mw_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("%s: msgWindow was NULL", __FUNCTION__);
      return;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_mw_get_ri_id(mw))) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return;
    }

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(mw->controls.textview));
  gtk_text_buffer_get_iter_at_offset(buffer, &bufStart, 0);
  gtk_text_buffer_get_iter_at_offset(buffer, &bufStop, -1);
  body = gtk_text_buffer_get_text(buffer,&bufStart,&bufStop,TRUE);

  if(body == NULL || g_utf8_strlen(body, -1) < 1)
    {
      g_warning("%s: cant send NULL or 0 length message", __FUNCTION__);
      return;
    }

  subject = gtk_editable_get_chars(GTK_EDITABLE(mw->controls.entry_subject),0,-1); 

  /* send */
  gj_message_send(messageTypeNormal, gj_roster_item_get_jid(ri), subject, body, NULL);

  gj_gtk_mw_unload_by_widget(mw->controls.window);
}

void on_button_mw_reply_clicked(GtkButton *button, gpointer user_data)
{
  msgWindow mw = NULL;
  rosterItem ri = NULL;

  if((mw = gj_gtk_mw_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("%s: failed to find window from button widget", __FUNCTION__);
      return;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_mw_get_ri_id(mw))) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return;
    }

  /* unload old ? */
  if(gj_config_get_mpf_close_previous_message_window_on_reply() == TRUE)
    gj_gtk_mw_unload(mw);

  /* create new */
  gj_gtk_mw_send_new(ri);
}

void on_button_mw_next_clicked(GtkButton *button, gpointer user_data)
{
  msgWindow mw = NULL;
  
  if((mw = gj_gtk_mw_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("%s: failed to find window from button widget", __FUNCTION__);
      return;
    }

  gj_gtk_mw_waiting_next(mw);
}

void on_button_mw_close_clicked(GtkButton *button, gpointer user_data)
{
  gj_gtk_mw_unload_by_widget(GTK_WIDGET(button));
}

gboolean on_message_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
  gj_gtk_mw_unload_by_widget(widget);

  return FALSE;
}

#ifdef __cplusplus
}
#endif
