/* 
 * $RCSfile: gj_gtk_rename_contact.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/31 16:43:55 $
 * 
 * $Revision: 1.8 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "common.h"

#include "gj_main.h"
#include "gj_roster.h"
#include "gj_iq_requests.h"

#include "gj_gtk_rename_contact.h"
#include "gj_gtk_msg.h"

struct 
{
  GtkWidget *window;

  GtkWidget *entry;
  GtkWidget *label_jid;
  GtkWidget *label_name;

  GtkWidget *button_ok;

} ctlRC;


gboolean gj_gtk_rc_save();
     
/*
 * internal callbacks 
 */
void gj_gtk_rc_user_changed_response(guint id);

/*
 * gui functions 
 */
void on_entry_rc_changed(GtkEditable *editable, gpointer user_data);
void on_entry_rc_activate(GtkEntry *entry, gpointer user_data);

void on_rename_contact_response(GtkDialog *dialog, gint response, gpointer user_data);


/*
 * window functions
 */

gboolean gj_gtk_rc_load(const gchar *jid, const gchar *name)
{
  GladeXML *xml = NULL;

  if(ctlRC.window != NULL)
    {
      gtk_window_present(GTK_WINDOW(ctlRC.window));
      return TRUE;
    }

  if(jid == NULL || g_utf8_strlen(jid, -1) < 1)
    {
      g_warning("gj_gtk_rc_load: jid was NULL");
      return FALSE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "rename_contact",
			  NULL)) == NULL)
    {
      g_warning("gj_gtk_rc_load: couldn't open glade file at root 'rename_contact'");
      return FALSE;
    }
  
  ctlRC.window = glade_xml_get_widget(xml,"rename_contact");

  ctlRC.entry = glade_xml_get_widget(xml,"entry_rc");
  ctlRC.label_jid = glade_xml_get_widget(xml,"label_rc_jid");
  ctlRC.label_name = glade_xml_get_widget(xml,"label_rc_name");

  ctlRC.button_ok = glade_xml_get_widget(xml,"button_rc_ok");
  
  gdk_window_set_decorations (GDK_WINDOW(ctlRC.window->window), 
			      GDK_DECOR_BORDER | GDK_DECOR_TITLE);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  g_signal_connect ((gpointer) ctlRC.window, "response",
		    G_CALLBACK (on_rename_contact_response),
		    NULL);
  g_signal_connect ((gpointer) ctlRC.entry, "changed",
		    G_CALLBACK (on_entry_rc_changed),
		    NULL);
  g_signal_connect ((gpointer) ctlRC.entry, "activate",
		    G_CALLBACK (on_entry_rc_activate),
		    NULL);
#endif

  /* set up window */
  gtk_label_set_text(GTK_LABEL(ctlRC.label_jid), jid);
  gtk_label_set_text(GTK_LABEL(ctlRC.label_name), name);

  /* clean up */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_rc_unload()
{
  if(ctlRC.window == NULL)
    return FALSE;

  gtk_widget_destroy(GTK_WIDGET(ctlRC.window));

  ctlRC.window = NULL;
  ctlRC.entry = NULL;
  ctlRC.label_jid = NULL;
  ctlRC.label_name = NULL;
  ctlRC.button_ok = NULL;

  return TRUE;
}

gboolean gj_gtk_rc_save()
{
  G_CONST_RETURN gchar *jid = NULL;
  G_CONST_RETURN gchar *name = NULL;
  rosterItem ri = NULL;

  jid = gtk_label_get_text(GTK_LABEL(ctlRC.label_jid));
  name = gtk_entry_get_text(GTK_ENTRY(ctlRC.entry));

  if((ri = gj_roster_item_find_by_jid((gchar*)jid)) == NULL)
    {
      g_warning("gj_gtk_rc_save: could not find roster item by jid:'%s'", jid);
      return FALSE;
    }

  /* save to jabber server */
  gj_iq_request_user_change(gj_roster_item_get_jid(ri), 
			    (gchar*)name, 
			    gj_roster_item_get_groups(ri),
			    gj_gtk_rc_user_changed_response);
  
  return TRUE;
}

/* callbacks */
void gj_gtk_rc_user_changed_response(guint id)
{
  infoquery iq = NULL;

  g_message("gj_gtk_rc_user_changed_response: response with requestID:%d",id);

  if((iq = gj_iq_find_by_related_id(id)) == NULL)
    {
      g_warning("gj_gtk_rc_user_changed_response: failed to find IQ based on responseID:%d",id);

      return;
    }

  /* clean up */
  gj_iq_free_all_related_by_related_id(id);
}


/*
 * GUI events
 */
void on_entry_rc_changed(GtkEditable *editable, gpointer user_data)
{
  gchar *group = NULL;
  group = gtk_editable_get_chars(editable, 0, -1);

  if(group != NULL && g_utf8_strlen(group, -1) > 0)
    {
      gtk_widget_set_sensitive(GTK_WIDGET(ctlRC.button_ok), TRUE);
      g_free(group);
      return;
    }
      
  gtk_widget_set_sensitive(GTK_WIDGET(ctlRC.button_ok), FALSE);
}

void on_entry_rc_activate(GtkEntry *entry, gpointer user_data)
{
  G_CONST_RETURN gchar *group = NULL;
  group = gtk_entry_get_text(entry);

  if(group != NULL && g_utf8_strlen(group, -1) > 0)
    on_rename_contact_response(GTK_DIALOG(ctlRC.window), GTK_RESPONSE_OK, NULL); 
}

void on_rename_contact_response(GtkDialog *dialog, gint response, gpointer user_data)
{
  switch(response) 
    {
    case GTK_RESPONSE_OK:
      gj_gtk_rc_save();
      break;

    case GTK_RESPONSE_HELP:
    case GTK_RESPONSE_NONE:
    case GTK_RESPONSE_REJECT:
    case GTK_RESPONSE_ACCEPT:
    case GTK_RESPONSE_DELETE_EVENT:
    case GTK_RESPONSE_CANCEL:
    case GTK_RESPONSE_CLOSE:
    case GTK_RESPONSE_YES:
    case GTK_RESPONSE_NO:
    case GTK_RESPONSE_APPLY:
      break;
    }

  gj_gtk_rc_unload();
}

#ifdef __cplusplus
}
#endif
