/* 
 * $RCSfile: gj_gtk_user_information.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/31 16:43:55 $
 * 
 * $Revision: 1.21 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <glade/glade.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_GNOME
#include <gnome.h>
#endif

#include "common.h"

#include "gj_main.h"
#include "gj_iq_requests.h"
#include "gj_parser.h"
#include "gj_roster.h"
#include "gj_translate.h"
#include "gj_support.h"

#include "gj_gtk_user_information.h"
#include "gj_gtk_edit_groups.h"


#define GJ_GTK_USER_INFORMATION_UNKNOWN "-"


struct t_ctlUI
{
  GtkWidget *window;
  
  GtkWidget *label_jid;
  GtkWidget *label_name;
  GtkWidget *label_group;
  GtkWidget *label_resource;
  GtkWidget *label_agent;

  GtkWidget *label_presence_type;
  GtkWidget *label_presence_show;
  GtkWidget *label_presence_status;
  GtkWidget *label_presence_subscription;
  GtkWidget *label_presence_subscription_description;
  GtkWidget *image_presence;

  GtkWidget *label_on_roster;
  GtkWidget *image_on_roster;

  GtkWidget *label_full_name;
  GtkWidget *label_name_given;
  GtkWidget *label_name_family;
  GtkWidget *label_name_nickname;

  GtkWidget *label_email;
  GtkWidget *label_url;
  GtkWidget *label_tel;
  
  GtkWidget *label_street;
  GtkWidget *label_extadd;
  GtkWidget *label_locality;
  GtkWidget *label_region;
  GtkWidget *label_pcode;
  GtkWidget *label_country;
  
  GtkWidget *label_org_orgname;
  GtkWidget *label_org_orgunit;
  GtkWidget *label_title;
  GtkWidget *label_role;

  GtkWidget *label_dob;
  GtkWidget *label_description;

  GtkWidget *vbox_version;
  GtkWidget *label_version_name;
  GtkWidget *label_version_version;
  GtkWidget *label_version_os;
  
  GtkWidget *vbox_time;
  GtkWidget *label_time_utc;
  GtkWidget *label_time_zone;
  GtkWidget *label_time_display;

  GtkWidget *vbox_last_action;
  GtkWidget *label_last_action;
  GtkWidget *label_last_seconds;

  GtkWidget *button_refresh;
  GtkWidget *button_close;
  
};

struct t_ui
{
  gint id;

  gchar *jid;
  gchar *name; 
  
  rosterItem ri;

  gint responses;
  
  struct t_ctlUI controls;
};

GList *userInformationList = NULL;


/* list functions */
gboolean gj_gtk_ui_list_add(userInformation ui);
gboolean gj_gtk_ui_list_del(userInformation ui);

/* item functions */
userInformation gj_gtk_ui_new(gchar *jid);
gboolean gj_gtk_ui_free(userInformation ui);
void gj_gtk_ui_destroy(userInformation ui);

/* find */
userInformation gj_gtk_ui_find_by_jid(gchar *jid);
userInformation gj_gtk_ui_find_by_widget(GtkWidget *widget);
userInformation gj_gtk_ui_find_by_id(gint id);

/* members */
gint gj_gtk_ui_get_id(userInformation ui);
const gchar *gj_gtk_ui_get_jid(userInformation ui);
const gchar *gj_gtk_ui_get_name(userInformation ui);
rosterItem gj_gtk_ui_get_ri(userInformation ui);
gboolean gj_gtk_ui_get_responses(userInformation ui);

gboolean gj_gtk_ui_set_name(userInformation ui, gchar *name);
gboolean gj_gtk_ui_set_jid(userInformation ui, gchar *jid);
gboolean gj_gtk_ui_set_ri(userInformation ui, rosterItem ri);
gboolean gj_gtk_ui_set_responses(userInformation ui, gboolean is_responses);

/* controls */
gboolean gj_gtk_ui_setup(userInformation ui);

gboolean gj_gtk_ui_get(userInformation ui);
gboolean gj_gtk_ui_set(userInformation ui);

void gj_gtk_ui_set_group_list();
void gj_gtk_ui_set_label(GtkWidget *widget, gchar *fmt, ...);

/*
 * gui functions 
 */
void on_button_ui_refresh_clicked(GtkButton *button, gpointer user_data);
void on_button_ui_close_clicked(GtkButton *button, gpointer user_data);

gboolean on_user_information_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);

/*
 * code
 */
gboolean gj_gtk_ui_init()
{
  return TRUE;
}

gboolean gj_gtk_ui_term()
{
  gint index = 0;
    
  /* list */
  if(userInformationList == NULL)
    return TRUE;

  /* free all items on the list */
  for(index=0;index<g_list_length(userInformationList);index++)
    {
      userInformation ui = NULL;
	
      if((ui = (userInformation) g_list_nth_data(userInformationList,index)) != NULL)
	gj_gtk_ui_free(ui);
    }
    
  g_list_free(userInformationList);
  userInformationList = NULL;
    
  return TRUE;
}

gboolean gj_gtk_ui_list_add(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_list_add: failed to add item to list, it was NULL");
      return FALSE;
    }

  userInformationList = g_list_append(userInformationList,ui);

  return TRUE;
}

gboolean gj_gtk_ui_list_del(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_list_del: failed to del item from list, it was NULL");
      return FALSE;
    }

  userInformationList = g_list_remove(userInformationList, ui);
  return TRUE;
}

userInformation gj_gtk_ui_new(gchar *jid)
{
  userInformation ui = NULL;

  if(jid == NULL) 
    {
      g_warning("gj_gtk_ui_new: jid was NULL");
      return NULL;
    }

  if((ui = g_new(struct t_ui,1)))
    {
      ui->id = gj_support_get_unique_id();

      ui->jid = g_strdup(jid);
      ui->name = NULL;

      ui->ri = NULL;
    }

  userInformationList = g_list_append(userInformationList, ui);

  return ui;
}

gboolean gj_gtk_ui_free(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_free: failed to free, userInformation was NULL");
      return FALSE;
    }

  /* free mem */
  gj_gtk_ui_list_del(ui);  

  /* destroy members */
  gj_gtk_ui_destroy(ui);

  /* free structure */
  g_free(ui);

  return TRUE;
}

void gj_gtk_ui_destroy(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_destroy: failed to destroy, userInformation was NULL");
      return;
    }

  ui->id = 0;

  if(ui->jid != NULL)
    {
      g_free(ui->jid);
      ui->jid = NULL;
    }

  if(ui->name != NULL)
    {
      g_free(ui->name);
      ui->name = NULL;
    }

  ui->ri = NULL;

  ui->responses = FALSE;
}

userInformation gj_gtk_ui_find_by_jid(gchar *jid)
{
  gint index = 0;

  if(jid == NULL)
    {
      g_warning("gj_gtk_ui_find_by_jid: jid was NULL");
      return NULL;
    }

  /* loop through all userInformationList on contacts userInformation list to look for unique userInformationID */
  for(index=0;index<g_list_length(userInformationList);index++)
    {
      userInformation ui = g_list_nth_data(userInformationList, index);
      
      if(ui == NULL)
	{
	  g_warning("gj_gtk_ui_find_by_jid: getting nth item(%d) in user information list was NULL",index);
	  continue;
	}
      
      if(g_strcasecmp(gj_gtk_ui_get_jid(ui),jid) == 0)
	return ui;
    }
 
  return NULL;
}

userInformation gj_gtk_ui_find_by_widget(GtkWidget *widget)
{
  gint index = 0;
  GtkWidget *toplevel = NULL;

  if(widget == NULL)
    {
      g_warning("gj_gtk_ui_find_by_widget: widget was NULL");
      return NULL;
    }

  toplevel = gtk_widget_get_toplevel(widget);
  if(!GTK_WIDGET_TOPLEVEL(toplevel))
    {
      g_warning("gj_gtk_ui_find_by_widget: could not find toplevel widget for widget:0x%.8x", (gint)widget);
      return NULL;
    }

  for(index=0;index<g_list_length(userInformationList);index++)
    {
      userInformation ui = g_list_nth_data(userInformationList, index);
      
      if(ui == NULL)
	{
	  g_warning("gj_gtk_ui_find_by_widget: getting nth item(%d) in user information list was NULL",index);
	  continue;
	}
      
      if(ui->controls.window == toplevel)
	return ui;
    }
 
  return NULL;
}

userInformation gj_gtk_ui_find_by_id(gint id)
{
  gint index = 0;

  if(id < 1)
    {
      g_warning("gj_gtk_ui_find_by_id: id < 1");
      return NULL;
    }

  for(index=0;index<g_list_length(userInformationList);index++)
    {
      userInformation ui = g_list_nth_data(userInformationList, index);
      
      if(ui == NULL)
	{
	  g_warning("gj_gtk_ui_find_by_id: getting nth item(%d) in user information list was NULL",index);
	  continue;
	}
      
      if(ui->id == id)
	return ui;
    }
 
  return NULL;
}

/*
 * member functions
 */

/* gets */
gint gj_gtk_ui_get_id(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_get_id: userInformation was NULL");
      return -1;
    }
  
  return ui->id;
}

const gchar *gj_gtk_ui_get_jid(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_get_jid: userInformation was NULL");
      return NULL;
    }
  
  return ui->jid;
}

const gchar *gj_gtk_ui_get_name(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_get_name: userInformation was NULL");
      return NULL;
    }
  
  return ui->name;
}

rosterItem gj_gtk_ui_get_ri(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_get_ri: userInformation was NULL");
      return NULL;
    }
  
  return ui->ri;
}

gint gj_gtk_ui_get_responses(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_get_responses: userInformation was NULL");
      return 0;
    }
  
  return ui->responses;
}

/* sets */
gboolean gj_gtk_ui_set_name(userInformation ui, gchar *name)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_set_name: userInformation was NULL");
      return FALSE;
    }

  if(name == NULL)
    {
      g_warning("gj_gtk_ui_set_name: name was NULL");
      return FALSE;
    }

  if(ui->name != NULL)
    {
      g_free(ui->name);
      ui->name = NULL;
    }

  ui->name = g_strdup(name);
  return TRUE;
}

gboolean gj_gtk_ui_set_jid(userInformation ui, gchar *jid)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_set_jid: userInformation was NULL");
      return FALSE;
    }

  if(jid == NULL)
    {
      g_warning("gj_gtk_ui_set_jid: jid was NULL");
      return FALSE;
    }

  if(ui->jid != NULL)
    {
      g_free(ui->jid);
      ui->jid = NULL;
    }

  ui->jid = g_strdup(jid);
  return TRUE;
}

gboolean gj_gtk_ui_set_ri(userInformation ui, rosterItem ri)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_set_ri: userInformation was NULL");
      return FALSE;
    }

  if(ri == NULL)
    {
      g_warning("gj_gtk_ui_set_ri: ri was NULL");
      return FALSE;
    }

  ui->ri = ri;
  return TRUE;
}

gboolean gj_gtk_ui_set_responses(userInformation ui, gint responses)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_set_responses: userInformation was NULL");
      return FALSE;
    }

  ui->responses = responses;

  if(responses < 1)
    gtk_widget_set_sensitive(GTK_WIDGET(ui->controls.button_refresh), TRUE);
  else
    gtk_widget_set_sensitive(GTK_WIDGET(ui->controls.button_refresh), FALSE);

  return TRUE;
}

/*
 * control functions 
 */
gboolean gj_gtk_ui_load(gchar *jid)
{
  GladeXML *xml;
  
  userInformation ui = NULL;
  rosterItem ri = NULL;
  
  if(jid == NULL)
    {
      g_warning("gj_gtk_ui_load: JID was NULL");
      return FALSE;
    }

  /* find previous */
  if((ui = gj_gtk_ui_find_by_jid(jid)) == NULL)
    {
      if((ui = gj_gtk_ui_new(jid)) == NULL)
	{
	  g_warning("gj_gtk_ui_load: userInformation was NULL");
	  return FALSE;
	}

      if((ri = gj_roster_item_find_by_jid(jid)) != NULL)
	  gj_gtk_ui_set_ri(ui, ri);
    } 
  else
    {
      gtk_window_present(GTK_WINDOW(ui->controls.window));
      return TRUE;
    }


  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "user_information",
			  NULL)) == NULL)
    {
      g_warning("gj_gtk_ui_load: couldn't open glade file at root 'user_information'");
      return FALSE;
    }
  
  ui->controls.window = glade_xml_get_widget(xml,"user_information");

  ui->controls.label_jid = glade_xml_get_widget(xml,"label_ui_jid");

  ui->controls.label_name = glade_xml_get_widget(xml,"label_ui_name");

  ui->controls.label_group = glade_xml_get_widget(xml,"label_ui_group");
  ui->controls.label_resource = glade_xml_get_widget(xml,"label_ui_resource");
  ui->controls.label_agent = glade_xml_get_widget(xml,"label_ui_agent");

  ui->controls.label_presence_type = glade_xml_get_widget(xml,"label_ui_presence_type");
  ui->controls.label_presence_show = glade_xml_get_widget(xml,"label_ui_presence_show");
  ui->controls.label_presence_status = glade_xml_get_widget(xml,"label_ui_presence_status");
  ui->controls.label_presence_subscription = glade_xml_get_widget(xml,"label_ui_presence_subscription");
  ui->controls.label_presence_subscription_description = glade_xml_get_widget(xml,"label_ui_presence_subscription_description");
  ui->controls.image_presence = glade_xml_get_widget(xml, "image_ui_presence");

  ui->controls.label_on_roster = glade_xml_get_widget(xml,"label_ui_on_roster");
  ui->controls.image_on_roster = glade_xml_get_widget(xml,"image_ui_on_roster");

  ui->controls.label_full_name = glade_xml_get_widget(xml,"label_ui_full_name");
  ui->controls.label_name_given = glade_xml_get_widget(xml,"label_ui_name_given");
  ui->controls.label_name_family = glade_xml_get_widget(xml,"label_ui_name_family");
  ui->controls.label_name_nickname = glade_xml_get_widget(xml,"label_ui_name_nickname");

  ui->controls.label_email = glade_xml_get_widget(xml,"label_ui_email");
  ui->controls.label_url = glade_xml_get_widget(xml,"label_ui_url");
  ui->controls.label_tel = glade_xml_get_widget(xml,"label_ui_tel");
  
  ui->controls.label_street = glade_xml_get_widget(xml,"label_ui_street");
  ui->controls.label_extadd = glade_xml_get_widget(xml,"label_ui_extadd");
  ui->controls.label_locality = glade_xml_get_widget(xml,"label_ui_locality");
  ui->controls.label_region = glade_xml_get_widget(xml,"label_ui_region");
  ui->controls.label_pcode = glade_xml_get_widget(xml,"label_ui_pcode");
  ui->controls.label_country = glade_xml_get_widget(xml,"label_ui_country");

  ui->controls.label_org_orgname = glade_xml_get_widget(xml,"label_ui_org_orgname");
  ui->controls.label_org_orgunit = glade_xml_get_widget(xml,"label_ui_org_orgunit");
  ui->controls.label_title = glade_xml_get_widget(xml,"label_ui_title");
  ui->controls.label_role = glade_xml_get_widget(xml,"label_ui_role");
  
  ui->controls.label_dob = glade_xml_get_widget(xml,"label_ui_dob");
  ui->controls.label_description = glade_xml_get_widget(xml,"label_ui_description");

  ui->controls.vbox_version = glade_xml_get_widget(xml,"vbox_ui_version");
  ui->controls.label_version_name = glade_xml_get_widget(xml,"label_ui_version_name");
  ui->controls.label_version_version = glade_xml_get_widget(xml,"label_ui_version_version");
  ui->controls.label_version_os = glade_xml_get_widget(xml,"label_ui_version_os");

  ui->controls.vbox_time = glade_xml_get_widget(xml,"vbox_ui_time");
  ui->controls.label_time_utc = glade_xml_get_widget(xml,"label_ui_time_utc");
  ui->controls.label_time_zone = glade_xml_get_widget(xml,"label_ui_time_zone");
  ui->controls.label_time_display = glade_xml_get_widget(xml,"label_ui_time_display");

  ui->controls.vbox_last_action = glade_xml_get_widget(xml,"vbox_ui_last_action");
  ui->controls.label_last_action = glade_xml_get_widget(xml,"label_ui_last_action");
  ui->controls.label_last_seconds = glade_xml_get_widget(xml,"label_ui_last_seconds");
  
  ui->controls.button_refresh = glade_xml_get_widget(xml,"button_ui_refresh");
  ui->controls.button_close = glade_xml_get_widget(xml,"button_ui_close");
  
  gdk_window_set_decorations (GDK_WINDOW(ui->controls.window->window), 
			      GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  g_signal_connect ((gpointer) ui->controls.window, "delete_event",
		    G_CALLBACK (on_user_information_delete_event),
		    NULL);
  g_signal_connect ((gpointer) ui->controls.button_close, "clicked",
		    G_CALLBACK (on_button_ui_close_clicked),
		    NULL);
  g_signal_connect ((gpointer) ui->controls.button_refresh, "clicked",
		    G_CALLBACK (on_button_ui_refresh_clicked),
		    NULL);
#endif

  /* set up labels with local information */
  gj_gtk_ui_setup(ui);

  /* get information from config */
  gj_gtk_ui_get(ui);

  /* clean up */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_ui_unload(userInformation ui)
{
  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_unload: userInformation was NULL");
      return FALSE;
    }

  gtk_widget_destroy(ui->controls.window);

  ui->controls.window = NULL;

  ui->controls.label_jid = NULL; 
  ui->controls.label_name = NULL; 
  ui->controls.label_group = NULL;  
  ui->controls.label_resource = NULL;  
  ui->controls.label_agent = NULL;  

  ui->controls.label_presence_type = NULL; 
  ui->controls.label_presence_show = NULL;  
  ui->controls.label_presence_status = NULL;  
  ui->controls.label_presence_subscription = NULL; 
  ui->controls.label_presence_subscription_description = NULL;  
  ui->controls.image_presence = NULL;

  ui->controls.label_on_roster = NULL;
  ui->controls.image_on_roster = NULL;

  ui->controls.label_full_name = NULL; 
  ui->controls.label_name_given = NULL;  
  ui->controls.label_name_family = NULL; 
  ui->controls.label_name_nickname = NULL;

  ui->controls.label_email = NULL;  
  ui->controls.label_url = NULL; 
  ui->controls.label_tel = NULL; 
  
  ui->controls.label_street = NULL; 
  ui->controls.label_extadd = NULL;  
  ui->controls.label_locality = NULL;  
  ui->controls.label_region = NULL;  
  ui->controls.label_pcode = NULL;  
  ui->controls.label_country = NULL; 

  ui->controls.label_org_orgname = NULL;  
  ui->controls.label_org_orgunit = NULL;   
  ui->controls.label_title = NULL;   
  ui->controls.label_role = NULL;   
  
  ui->controls.label_dob = NULL;   
  ui->controls.label_description = NULL;   

  ui->controls.vbox_version = NULL;  
  ui->controls.label_version_name = NULL;   
  ui->controls.label_version_version = NULL;  
  ui->controls.label_version_os = NULL;   

  ui->controls.vbox_time = NULL;   
  ui->controls.label_time_utc = NULL;   
  ui->controls.label_time_zone = NULL;  
  ui->controls.label_time_display = NULL;   

  ui->controls.vbox_last_action = NULL;   
  ui->controls.label_last_action = NULL;   
  ui->controls.label_last_seconds = NULL;  
  
  ui->controls.button_refresh = NULL;   
  ui->controls.button_close = NULL;   

  gj_gtk_ui_free(ui);
  ui = NULL;

  return TRUE;
}

gboolean gj_gtk_ui_setup(userInformation ui)
{
  rosterItem ri = NULL;

  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_setup: userInformation was NULL");
      return FALSE;
    }

  /* load information into window */
  gj_gtk_ui_set_label(ui->controls.label_jid,"%s",gj_gtk_ui_get_jid(ui));

  if((ri = gj_gtk_ui_get_ri(ui)) != NULL)
    {
      gj_gtk_ui_set_label(ui->controls.label_on_roster,"%s", gj_roster_item_get_is_permanent(ri)?_("Yes"):_("No"));
      gtk_image_set_from_stock(GTK_IMAGE(ui->controls.image_on_roster), gj_roster_item_get_is_permanent(ri)?GTK_STOCK_YES:GTK_STOCK_NO, GTK_ICON_SIZE_BUTTON);
    }
  else
    gj_gtk_ui_set_label(ui->controls.label_on_roster,_("Unknown"));

  /* fill groups combo */
  gj_gtk_ui_set_group_list();
  
  return TRUE;
}

gboolean gj_gtk_ui_get(userInformation ui)
{
  rosterItem ri = NULL;
  presence pres = NULL;

  gchar *groups = NULL;
  gchar *file = NULL;
  gint id = 0;

  if(ui == NULL)
    {
      g_warning("gj_gtk_ui_get: userInformation was NULL");
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_jid((gchar*)gj_gtk_ui_get_jid(ui))) == NULL)
    {
      g_warning("gj_gtk_ui_get: could not find contact with JID:'%s'",gj_gtk_ui_get_jid(ui));
      return FALSE;
    }

  if((pres = gj_roster_item_get_presence(ri)) == NULL)
    {
      g_warning("gj_gtk_ui_get: roster item's presence was NULL for jid:'%s'",gj_gtk_ui_get_jid(ui));
      return FALSE;
    }

  groups = gj_roster_item_get_groups_as_string(ri);
  id = gj_gtk_ui_get_id(ui);

  /* get remote information */
  gj_gtk_ui_set_responses(ui, 4); 

  gj_iq_request_vcard((gchar*)gj_gtk_ui_get_jid(ui), gj_gtk_ui_vcard_response, GINT_TO_POINTER(id));
  gj_iq_request_version((gchar*)gj_gtk_ui_get_jid(ui), gj_gtk_ui_version_response, GINT_TO_POINTER(id));
  gj_iq_request_time((gchar*)gj_gtk_ui_get_jid(ui), gj_gtk_ui_time_response, GINT_TO_POINTER(id));
  gj_iq_request_last((gchar*)gj_gtk_ui_get_jid(ui), gj_gtk_ui_last_response, GINT_TO_POINTER(id));

  /*
   * locally obtainable information
   *
   */

  gj_gtk_ui_set_name(ui, gj_roster_item_get_name(ri));
  gj_gtk_ui_set_label(ui->controls.label_name, gj_roster_item_get_name(ri));
  gj_gtk_ui_set_label(ui->controls.label_group,"%s", groups?groups:GJ_GTK_USER_INFORMATION_UNKNOWN);
  gj_gtk_ui_set_label(ui->controls.label_resource,"%s", gj_roster_item_get_resource(ri)?gj_roster_item_get_resource(ri):GJ_GTK_USER_INFORMATION_UNKNOWN);
  gj_gtk_ui_set_label(ui->controls.label_agent,"%s", gj_roster_item_get_agent(ri)?gj_roster_item_get_agent(ri):GJ_GTK_USER_INFORMATION_UNKNOWN);

  gj_gtk_ui_set_label(ui->controls.label_presence_type,"%s", gj_translate_presence_type(gj_presence_get_type(pres)));
  gj_gtk_ui_set_label(ui->controls.label_presence_show,"%s", gj_translate_presence_show(gj_presence_get_show(pres)));
  gj_gtk_ui_set_label(ui->controls.label_presence_status,"%s", gj_presence_get_status(pres)?gj_presence_get_status(pres):GJ_GTK_USER_INFORMATION_UNKNOWN);
  gj_gtk_ui_set_label(ui->controls.label_presence_subscription,"%s", gj_translate_roster_item_subscription(gj_roster_item_get_subscription(ri)));
  gj_gtk_ui_set_label(ui->controls.label_presence_subscription_description,"%s", gj_translate_roster_item_subscription_as_description(gj_roster_item_get_subscription(ri)));

  if(gj_presence_get_type(pres) != presenceTypeAvailable)
    {
      switch(gj_presence_get_type(pres))
	{
	case presenceTypeUnAvailable: file = FILE_PRESENCE_UNAVAILABLE; break;
	case presenceTypeSubscribe: file = FILE_PRESENCE_SUBSCRIBE; break;
	case presenceTypeUnSubscribe: file = FILE_PRESENCE_UNSUBSCRIBE; break;
	case presenceTypeSubscribed: file = FILE_PRESENCE_SUBSCRIBED; break;
	case presenceTypeUnSubscribed: file = FILE_PRESENCE_UNSUBSCRIBED; break;
	case presenceTypeError: file = FILE_PRESENCE_ERROR; break;
	default: break;
	}
    }
  else
    {
      switch(gj_presence_get_show(pres))
	{
	case presenceShowAway: file = FILE_PRESENCE_AWAY; break;
	case presenceShowChat: file = FILE_PRESENCE_CHAT; break;
	case presenceShowDND: file = FILE_PRESENCE_DND; break;
	case presenceShowNormal: file = FILE_PRESENCE_NORMAL; break;
	case presenceShowXA: file = FILE_PRESENCE_XA; break;
	default: break;
	}
    }
  
  if(file != NULL)
    gtk_image_set_from_file(GTK_IMAGE(ui->controls.image_presence), file);

  /* clean up */
  if(groups != NULL) g_free(groups);

  return TRUE;
}

void gj_gtk_ui_set_label(GtkWidget *widget, gchar *fmt, ...)
{
  va_list args;
  gchar *buffer = NULL;

  if(widget == NULL)
    {
      g_warning("gj_gtk_ui_set_label: widget was NULL");
      return;
    }
  
  va_start(args,fmt);
  buffer = g_strdup_vprintf(fmt,args);
  va_end(args);

  if(buffer == NULL)
    {
      g_warning("gj_gtk_ui_set_label: buffer to use was NULL");
      return;
    }

  gtk_label_set_markup(GTK_LABEL(widget),buffer);
}

void gj_gtk_ui_set_group_list()
{

}

void gj_gtk_ui_vcard_response(guint id)
{
  xmlChar *tmp = NULL;

  infoquery iq = NULL;
  infoquery iq_related = NULL;

  userInformation ui = NULL;
  gpointer user_data = NULL;

  g_message("gj_gtk_ui_vcard_response: response with eventID:%d",id);

  if((iq = gj_iq_find_by_event_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_vcard_response: failed to find IQ based on eventID:%d",id);
      return;
    }
  
  if((iq_related = gj_iq_find_by_related_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_vcard_response: failed to find IQ based on eventIDRelated:%d",id);
      return;
    }

  if((user_data = gj_iq_get_userdata(iq)) == NULL)
    {
      g_warning("gj_gtk_ui_vcard_response: could not find userInformation userdata");
      return;
    }

  if((ui = gj_gtk_ui_find_by_id(GPOINTER_TO_INT(user_data))) == NULL)
  {
    g_warning("gj_gtk_ui_vcard_response: could not find userInformation from id:%d", GPOINTER_TO_INT(user_data));
    return;
  }

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"fn");
  gj_gtk_ui_set_label(ui->controls.label_full_name, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"given");
  gj_gtk_ui_set_label(ui->controls.label_name_given, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"family");
  gj_gtk_ui_set_label(ui->controls.label_name_family, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"nickname");
  gj_gtk_ui_set_label(ui->controls.label_name_nickname, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"url");
  gj_gtk_ui_set_label(ui->controls.label_url, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);
  
  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"tel");
  gj_gtk_ui_set_label(ui->controls.label_tel, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"email");
  gj_gtk_ui_set_label(ui->controls.label_email, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"desc");
  gj_gtk_ui_set_label(ui->controls.label_description, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"bday");
  gj_gtk_ui_set_label(ui->controls.label_dob, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"street");
  gj_gtk_ui_set_label(ui->controls.label_street, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"extadd");
  gj_gtk_ui_set_label(ui->controls.label_extadd, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"locality");
  gj_gtk_ui_set_label(ui->controls.label_locality, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"region");
  gj_gtk_ui_set_label(ui->controls.label_region, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"pcode");
  gj_gtk_ui_set_label(ui->controls.label_pcode, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"country");
  gj_gtk_ui_set_label(ui->controls.label_country, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"orgname");
  gj_gtk_ui_set_label(ui->controls.label_org_orgname, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"orgunit");
  gj_gtk_ui_set_label(ui->controls.label_org_orgunit, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"title");
  gj_gtk_ui_set_label(ui->controls.label_title, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"role");
  gj_gtk_ui_set_label(ui->controls.label_role, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

  gj_gtk_ui_set_responses(ui, gj_gtk_ui_get_responses(ui) -1); 

  /* clean up */
  gj_iq_free_all_related_by_related_id(id);
}

void gj_gtk_ui_version_response(guint id)
{
  xmlChar *tmp = NULL;

  infoquery iq = NULL;
  infoquery iq_related = NULL;

  userInformation ui = NULL;
  gpointer user_data = NULL;

  g_message("gj_gtk_ui_version_response: response with eventID:%d",id);

  if((iq = gj_iq_find_by_event_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_version_response: failed to find IQ based on eventID:%d",id);
      return;
    }
  
  if((iq_related = gj_iq_find_by_related_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_version_response: failed to find IQ based on eventIDRelated:%d",id);
      return;
    }

  if((user_data = gj_iq_get_userdata(iq)) == NULL)
    {
      g_warning("gj_gtk_ui_version_response: could not find userInformation userdata");
      return;
    }

  if((ui = gj_gtk_ui_find_by_id(GPOINTER_TO_INT(user_data))) == NULL)
  {
    g_warning("gj_gtk_ui_version_response: could not find userInformation from id:%d", GPOINTER_TO_INT(user_data));
    return;
  }

  tmp = gj_parser_find_by_attr_and_ret(gj_iq_get_xml(iq_related), (guchar*)"type");
  if(tmp != NULL && xmlStrcmp(tmp, (guchar*)"error") != 0)
    {
      gtk_widget_set_sensitive(ui->controls.vbox_version, TRUE);

      tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"name");
      gj_gtk_ui_set_label(ui->controls.label_version_name, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);
      
      tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"os");
      gj_gtk_ui_set_label(ui->controls.label_version_os, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);
      
      tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"version");
      gj_gtk_ui_set_label(ui->controls.label_version_version, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);      
    }
  else
    {
      gtk_widget_set_sensitive(ui->controls.vbox_version, FALSE);
    }

  gj_gtk_ui_set_responses(ui, gj_gtk_ui_get_responses(ui) -1); 

  /* clean up */
  gj_iq_free_all_related_by_related_id(id);
}

void gj_gtk_ui_time_response(guint id)
{
  xmlChar *tmp = NULL;

  infoquery iq = NULL;
  infoquery iq_related = NULL;

  userInformation ui = NULL;
  gpointer user_data = NULL;

  g_message("gj_gtk_ui_time_response: response with requestID:%d",id);

  if((iq = gj_iq_find_by_event_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_time_response: failed to find IQ based on eventID:%d",id);
      return;
    }
  
  if((iq_related = gj_iq_find_by_related_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_time_response: failed to find IQ based on eventIDRelated:%d",id);
      return;
    }

  if((user_data = gj_iq_get_userdata(iq)) == NULL)
    {
      g_warning("gj_gtk_ui_time_response: could not find userInformation userdata");
      return;
    }

  if((ui = gj_gtk_ui_find_by_id(GPOINTER_TO_INT(user_data))) == NULL)
  {
    g_warning("gj_gtk_ui_time_response: could not find userInformation from id:%d", GPOINTER_TO_INT(user_data));
    return;
  }

  tmp = gj_parser_find_by_attr_and_ret(gj_iq_get_xml(iq_related), (guchar*)"type");
  if(tmp != NULL && xmlStrcmp(tmp, (guchar*)"error") != 0)
    {
      gchar **time = NULL;
      gchar *date = NULL;

      gtk_widget_set_sensitive(ui->controls.vbox_time, TRUE);

      if((tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"utc")) != NULL)
	{
	  time = g_strsplit((gchar*)tmp,"T",-1);

	  if(time[0] != NULL)
	    date = g_strdup_printf("%2.2s-%2.2s-%4.4s",time[0]+6, time[0]+4, time[0]);
	}

      gj_gtk_ui_set_label(ui->controls.label_time_utc, "%s (%s)",time[1]?time[1]:GJ_GTK_USER_INFORMATION_UNKNOWN, date?date:GJ_GTK_USER_INFORMATION_UNKNOWN);
      
      tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"tz");
      gj_gtk_ui_set_label(ui->controls.label_time_zone, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);
      
      tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"display");
      gj_gtk_ui_set_label(ui->controls.label_time_display, "%s",tmp?(gchar*)tmp:GJ_GTK_USER_INFORMATION_UNKNOWN);

      /* clean up */
      if(time != NULL) g_strfreev(time);
      if(date != NULL) g_free(date);
    }
  else
    {
      gtk_widget_set_sensitive(ui->controls.vbox_time, FALSE);
    }

  gj_gtk_ui_set_responses(ui, gj_gtk_ui_get_responses(ui) -1); 

  /* clean up */
  gj_iq_free_all_related_by_related_id(id);
}

void gj_gtk_ui_last_response(guint id)
{
  xmlChar *tmp = NULL;

  infoquery iq = NULL;
  infoquery iq_related = NULL;

  userInformation ui = NULL;
  gpointer user_data = NULL;

  g_message("gj_gtk_ui_last_response: response with requestID:%d",id);

  if((iq = gj_iq_find_by_event_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_last_response: failed to find IQ based on eventID:%d",id);
      return;
    }
  
  if((iq_related = gj_iq_find_by_related_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_last_response: failed to find IQ based on eventIDRelated:%d",id);
      return;
    }

  if((user_data = gj_iq_get_userdata(iq)) == NULL)
    {
      g_warning("gj_gtk_ui_last_response: could not find userInformation userdata");
      return;
    }

  if((ui = gj_gtk_ui_find_by_id(GPOINTER_TO_INT(user_data))) == NULL)
  {
    g_warning("gj_gtk_ui_last_response: could not find userInformation from id:%d", GPOINTER_TO_INT(user_data));
    return;
  }

  tmp = gj_parser_find_by_attr_and_ret(gj_iq_get_xml(iq_related), (guchar*)"type");
  if(tmp != NULL && xmlStrcmp(tmp, (guchar*)"error") != 0)
    {
      xmlNodePtr node = NULL;
      gchar *action = NULL;
      gchar *time = NULL;

      if((node = gj_parser_find_by_node(gj_iq_get_xml(iq_related), (guchar*)"query")) != NULL)
	{
	  /* get seconds */
	  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"seconds")) != NULL)
	    {
	      guint x = atoi((gchar*)tmp);
	      guint seconds = 0;
	      guint minutes = 0;
	      guint hours = 0;
	      guint days = 0;
	      
	      if(x >= 0)
		{
		  /* calculate from there */
		  seconds = x % 60;
		  x /= 60;
		  minutes = x % 60;
		  x /= 60;
		  hours = x % 24;
		  x /= 24;
		  days = x;
		  
		  time = g_strdup_printf("%d %s, %d %s, %d %s, %d %s ago", 
					 days, days==1?_("day"):_("days"), 
					 hours, hours==1?_("hour"):_("hours"),
					 minutes, minutes==1?_("minute"):_("minutes"),
					 seconds, seconds==1?_("second"):_("seconds"));
		}
	    }

	  if((tmp = gj_parser_find_by_node_and_ret(gj_iq_get_xml(iq_related), (guchar*)"query")) != NULL)
	    action = g_strdup_printf("%s",tmp?(gchar*)tmp:_("unknown action"));

	  gj_gtk_ui_set_label(ui->controls.label_last_action, "%s",action?action:GJ_GTK_USER_INFORMATION_UNKNOWN);
	  gj_gtk_ui_set_label(ui->controls.label_last_seconds, "%s",time?time:GJ_GTK_USER_INFORMATION_UNKNOWN);

	  /* clean up */
	  if(action != NULL) g_free(action);
	  if(time != NULL) g_free(time);
	}	  

      gtk_widget_set_sensitive(ui->controls.vbox_last_action, TRUE);
    }
  else
    {
      gtk_widget_set_sensitive(ui->controls.vbox_last_action, FALSE);
    }

  gj_gtk_ui_set_responses(ui, gj_gtk_ui_get_responses(ui) -1); 

  /* clean up */
  gj_iq_free_all_related_by_related_id(id);
}

void gj_gtk_ui_user_changed_response(guint id)
{
  infoquery iq = NULL;

  g_message("gj_gtk_ui_user_changed_response: response with requestID:%d",id);

  if((iq = gj_iq_find_by_related_id(id)) == NULL)
    {
      g_warning("gj_gtk_ui_user_changed_response: failed to find IQ based on responseID:%d",id);

      return;
    }

  /* clean up */
  gj_iq_free_all_related_by_related_id(id);
}

/*
 * GUI events
 */
void on_button_ui_refresh_clicked(GtkButton *button, gpointer user_data)
{
  userInformation ui = NULL;

  if((ui = gj_gtk_ui_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("on_button_ui_refresh_clicked: userInformation was NULL");
      return;
    }

  if(gj_gtk_ui_get_jid(ui) != NULL)
    gj_gtk_ui_get(ui);
}

void on_button_ui_close_clicked(GtkButton *button, gpointer user_data)
{
  userInformation ui = NULL;

  if((ui = gj_gtk_ui_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("on_button_ui_close_clicked: userInformation was NULL");
      return;
    }

  gj_gtk_ui_unload(ui);
}

gboolean on_user_information_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
  userInformation ui = NULL;

  if((ui = gj_gtk_ui_find_by_widget(widget)) == NULL)
    {
      g_warning("on_user_information_delete_event: userInformation was NULL");
      return FALSE;
    }

  gj_gtk_ui_unload(ui);

  return TRUE;
}

#ifdef __cplusplus
}
#endif
