/* 
 * $RCSfile: gj_iq_responses.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/13 17:50:15 $
 * 
 * $Revision: 1.8 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>

#include "gj_iq.h"
#include "gj_iq_responses.h"
#include "gj_roster.h"
#include "gj_me.h"
#include "gj_parser.h"

#include "gj_support.h"

/*
 * ACTUAL REQUESTS from the SERVER!!!, eg:
 *
 *  - jabber:iq:last
 *  - jabber:iq:version
 *  - jabber:iq:time
 *
 *
 *
 *
 *
 */

gboolean gj_iq_request_new(infoquery iq)
{
  const xmlChar *ns = NULL;
  
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }
 
  if((ns = gj_parser_find_by_ns_and_ret(gj_iq_get_xml(iq))) == NULL || xmlStrlen(ns) < 1)
    {
      g_warning("%s: could not find expected xml namespace!", __FUNCTION__);
      return FALSE;
    }
 
  if(gj_iq_get_type(iq) == infoqueryTypeGet)
    {
      if(xmlStrcmp(ns, (guchar*)"jabber:iq:last") == 0)
	gj_iq_respond_to_jabber_iq_last(iq);
      else if(xmlStrcmp(ns, (guchar*)"jabber:iq:time") == 0)
	gj_iq_respond_to_jabber_iq_time(iq);
      else if(xmlStrcmp(ns, (guchar*)"jabber:iq:version") == 0)
	gj_iq_respond_to_jabber_iq_version(iq);
      else
	gj_iq_respond_with_service_unavailable(iq);
    }
  else if(gj_iq_get_type(iq) == infoqueryTypeSet)
    {
      if(xmlStrcmp(ns, (guchar*)"jabber:iq:roster") == 0)
	gj_roster_parse_set_request(iq);
      else
	gj_iq_respond_with_service_unavailable(iq);
    }
  else if(gj_iq_get_type(iq) == infoqueryTypeResult)
    {
      if(xmlStrcmp(ns, (guchar*)"jabber:iq:register") == 0)
	{
	  infoqueryCallback cb = NULL;

	  if((cb = gj_iq_get_callback(iq)) != NULL)
	    (cb)(gj_iq_get_event_id(iq));
	}
    }
  else
    {
      /* error? why? OR result? why? */
      g_warning("%s: unexpected infoquery type:%d, not supported yet!", __FUNCTION__, gj_iq_get_type(iq));
      return FALSE;
    }
  
  /* free iq ?? */
  gj_iq_free(iq);
  
  return TRUE;
}


gboolean gj_iq_respond_to_jabber_iq_last(infoquery iq)
{
  infoquery iq_new = NULL;
  
  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;
  
  xmlChar *tmp = NULL;

  gchar *from = NULL;
  gchar *seconds = NULL;

  if((iq_new = gj_iq_new(infoqueryTypeResult)) == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if((tmp = gj_parser_find_by_attr_and_ret(gj_iq_get_xml(iq), (guchar*)"from")) == NULL || xmlStrlen(tmp) < 1)
    {
      g_warning("%s: failed to obtain 'from' information", __FUNCTION__);
      return FALSE;
    }
  else
    from = g_strdup_printf("%s",tmp);
  
  seconds = g_strdup_printf("%f",gj_me_get_uptime());
  
  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:last", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"seconds", (guchar*)seconds);
 
  /* set struct info */
  gj_iq_set_to(iq_new, (gchar*)from);

  if(gj_iq_build(iq_new, node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  if(gj_iq_send(iq_new) == FALSE)
    {
      g_warning("%s: failed to send query", __FUNCTION__);
      return FALSE;
    }
  
  /* clean up */
  gj_iq_free(iq_new);
  if(from != NULL) g_free(from);
  if(seconds != NULL) g_free(seconds);

  return TRUE;
}

gboolean gj_iq_respond_to_jabber_iq_version(infoquery iq)
{
  infoquery iq_new = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  xmlChar *tmp = NULL;

  gchar *from = NULL;

  if((iq_new = gj_iq_new(infoqueryTypeResult)) == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if((tmp = gj_parser_find_by_attr_and_ret(gj_iq_get_xml(iq), (guchar*)"from")) == NULL || xmlStrlen(tmp) < 1)
    {
      g_warning("%s: failed to obtain 'from' information", __FUNCTION__);
      return FALSE;
    }
  else
    from = g_strdup_printf("%s",tmp);
  
  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:version", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"name", (guchar*)"gnome-jabber");
  sub = xmlNewChild(node, ns, (guchar*)"os", (guchar*)"linux");
  sub = xmlNewChild(node, ns, (guchar*)"version", (guchar*)"0.1");
 
  /* set struct info */
  gj_iq_set_to(iq_new,from);

  if(gj_iq_build(iq_new,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  if(gj_iq_send(iq_new) == FALSE)
    {
      g_warning("%s: failed to send query", __FUNCTION__);
      return FALSE;
    }
  
  /* clean up */
  gj_iq_free(iq_new);
  if(from != NULL) g_free(from);

  return TRUE;
}

gboolean gj_iq_respond_to_jabber_iq_time(infoquery iq)
{
  infoquery iq_new = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  xmlChar *tmp = NULL;

  gchar *from = NULL;
  gchar *utc = NULL;

  if((iq_new = gj_iq_new(infoqueryTypeResult)) == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if((tmp = gj_parser_find_by_attr_and_ret(gj_iq_get_xml(iq), (guchar*)"from")) == NULL || xmlStrlen(tmp) < 1)
    {
      g_warning("%s: failed to obtain 'from' information", __FUNCTION__);
      return FALSE;
    }
  else
    from = g_strdup_printf("%s",tmp);

  utc = g_strdup_printf("%sT%s", gj_support_get_datestamp_long(), gj_support_get_timestamp());
  
  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:time", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"utc", (guchar*)utc);
  sub = xmlNewChild(node, ns, (guchar*)"tz", (guchar*)gj_support_get_time_zone());  /* need to get TZ */
  sub = xmlNewChild(node, ns, (guchar*)"display", (guchar*)gj_support_get_datestamp());
 
  /* set struct info */
  gj_iq_set_to(iq_new,from);

  if(gj_iq_build(iq_new,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  if(gj_iq_send(iq_new) == FALSE)
    {
      g_warning("%s: failed to send query", __FUNCTION__);
      return FALSE;
    }
  
  /* clean up */
  gj_iq_free(iq_new);
  if(from != NULL) g_free(from);
  if(utc != NULL) g_free(utc);

  return TRUE;
}

gboolean gj_iq_respond_with_service_unavailable(infoquery iq)
{
  infoquery iq_new = NULL;

  xmlNodePtr node1 = NULL;
  xmlNodePtr node2 = NULL;
  xmlNsPtr ns = NULL;

  GList *nodes = NULL;
  
  const xmlChar *tmp = NULL;
  
  gchar *namespace = NULL;
  gchar *from = NULL;

  if((tmp = gj_parser_find_by_ns_and_ret(gj_iq_get_xml(iq))) == NULL || xmlStrlen(tmp) < 1)
    {
      g_warning("%s: could not determin namespace from infoquery", __FUNCTION__);
      return FALSE;
    }
  else
    namespace = g_strdup_printf("%s",tmp);

  if((tmp = gj_parser_find_by_attr_and_ret(gj_iq_get_xml(iq), (guchar*)"from")) == NULL || xmlStrlen(tmp) < 1)
    {
      g_warning("%s: failed to obtain 'from' information", __FUNCTION__);
      return FALSE;
    }
  else
    from = g_strdup_printf("%s",tmp);

  if((iq_new = gj_iq_new(infoqueryTypeError)) == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  /* set xml */
  node1 = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node1, (guchar*)namespace, NULL);

  node2 = xmlNewNode(NULL, (guchar*)"error");
  xmlSetProp(node2, (guchar*)"code", (guchar*)"503");  
  xmlNodeSetContent(node2, (guchar*)"Service Unavailable");

  nodes = g_list_append(nodes,node1);
  nodes = g_list_append(nodes,node2);
 
  /* set struct info */
  gj_iq_set_to(iq_new, (gchar*)from);

  if(gj_iq_build_from_list(iq_new,nodes) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  if(gj_iq_send(iq_new) == FALSE)
    {
      g_warning("%s: failed to send query", __FUNCTION__);
      return FALSE;
    }
  
  /* clean up */
  gj_iq_free(iq_new);
  if(namespace != NULL) g_free(namespace);
  if(from != NULL) g_free(from);
  if(nodes != NULL) g_list_free(nodes);

  return TRUE;
}


#ifdef __cplusplus
}
#endif
