/* 
 * $RCSfile: gj_main.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/28 21:57:03 $
 * 
 * $Revision: 1.30 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <stdio.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gnet.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_GNOME
#include <gnome.h>
#endif

/* required for mkdir */
#ifndef WIN32
#  include <locale.h>
#  include <sys/stat.h>
#else
#  include <direct.h>
#endif

#include "common.h"

#include "gj_main.h"
#include "gj_connection.h"
#include "gj_stream.h"
#include "gj_parser.h"
#include "gj_iq.h"
#include "gj_roster.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_translate.h"
#include "gj_log.h"
#include "gj_me.h"
#include "gj_config.h"
#include "gj_history.h"
#include "gj_gatherdata.h"
#include "gj_registration.h"
#include "gj_search.h"
#include "gj_autoaway.h"

#include "gj_gtk_roster.h"
#include "gj_gtk_my_preferences.h"
#include "gj_gtk_user_information.h"
#include "gj_gtk_first_time_run.h"

gboolean gnome_jabber_init();
gboolean gnome_jabber_term();

gboolean gnome_jabber_first_time_check();
gboolean gnome_jabber_files_check();

struct
{
  gchar *file_config;
  gchar *file_glade;

  gchar *dir_config;  

  gboolean first_time_run;

  gboolean debugging;

} gnome_jabber;


/***********************************************************************
 * JABBER SERVERS        
 ***********************************************************************/



const gchar *gnome_jabber_servers[] = {
  "12jabber.com",
  "aedsolucoes.com.br",
  "akl.lt",
  "amessage.at",
  "amessage.be",
  "amessage.ch",
  "amessage.de",
  "amessage.info",
  "amessage.li",
  "bloodyxml.com",
  "bluoptic.com",
  "borderlinenormal.com",
  "dhbit.ca",
  "eifelansinn.net",
  "es.tipic.com",
  "geodude.timmins.net",
  "ibericachat.com",
  "illx.org",
  "im.pizza.net",
  "jabber-fr.net",
  "jabber.2000-plus.pl",
  "jabber.anywise.com",
  "jabber.at",
  "jabber.atari-source.com",
  "jabber.bettercom.de",
  "jabber.cn",
  "jabber.com",
  "jabber.cyber-junk.de",
  "jabber.cz",
  "jabber.daemon.sh",
  "jabber.dk",
  "jabber.earth.li",
  "jabber.freenet.de",
  "jabber.gda.pl",
  "jabber.hu",
  "jabber.kelkoo.net",
  "jabber.kicks-ass.org",
  "jabber.kolbsoft.com",
  "jabber.linux.it",
  "jabber.lumina.ro",
  "jabber.mailclean.net",
  "jabber.mandog.com",
  "jabber.netsample.com",
  "jabber.ngnet.it",
  "jabber.noze.it",
  "jabber.or.id",
  "jabber.or.kr",
  "jabber.org",
  "jabber.palomine.net",
  "jabber.projecto-oasis.cx",
  "jabber.ro",
  "jabber.ru",
  "jabber.sk",
  "jabber.toble.com",
  "jabber.unesco.kz",
  "jabber.wiretrip.org",
  "jabber.wp.pl",
  "jabberlawine.de",
  "jabberpl.org",
  "kanga.nu",
  "myjabber.net",
  "nabla.net",
  "netmindz.net",
  "netnewmedia.de",
  "njs.netlab.cz",
  "nureality.ca",
  "openaether.org",
  "oracle.sweeney.ath.cx",
  "planet-linux.org",
  "rhymbox.com",
  "sml.dyndns.org",
  "sourcecode.de",
  "sucs.org",
  "swissjabber.ch",
  "syndicon.de",
  "theoretic.com",
  "tipic.com",
  "tipic.it",
  "xasa.com",
  NULL 
};

/****************************************************************************/

int main(int argc, char **argv)
{
  /*   g_mem_set_vtable(glib_mem_profiler_table); */
  return gnome_jabber_main(argc, argv);
}

gboolean gnome_jabber_main(int argc, char **argv)
{
  gint i = 0;

  /* init variables */
  gnome_jabber_init();

  for(i=0; i<argc; i++)
    {
      if(strcmp(argv[i], "--debugging") == 0)
	gnome_jabber_set_debugging(TRUE);
    }

  gnet_init();
  g_message("main: initiated gnet");

  /* gtk / gdk / gnome initiations */
  if(gtk_init_check(&argc,&argv) == FALSE)
    {
      g_error("failed to initiate gtk");
      exit(0);
    }
  else
    g_message("main: initiated gtk");

  if(gdk_init_check(&argc,&argv) == FALSE)
    {
      g_error("failed to initiate gdk");
      exit(0);
    }
  else
    g_message("main: initiated gdk");

#ifndef WIN32
  /* internationalisation */
  setlocale(LC_ALL, "");
  bindtextdomain(GETTEXT_PACKAGE, GNOMELOCALEDIR);
  textdomain(GETTEXT_PACKAGE);
  bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif

  /* check to see if we have run this app before? */
  gnome_jabber_files_check();

  if(gnome_jabber_get_debugging() == TRUE)
    {
      gnome_jabber_set_glade_file("gnome-jabber.glade");
      g_message("main: using local gnome-jabber.glade");
    }

  /* config */
  gj_config_init(gnome_jabber_get_config_file());
  g_message("main: initiated config");

  /* logging */
  gj_log_init();
  g_message("main: initiated logging");

#ifdef HAVE_GNOME
  /* sound */
  gnome_sound_init("localhost");
  g_message("main: initiated gnome-sound");
#endif

  /* module initialisation */
  gnome_jabber_init_modules();
  g_message("main: initiated jabber modules");

  /* get config */
  gj_config_read();

  /* check for first time run */
  if(gnome_jabber_get_first_time_run() == TRUE)
    {
      g_message("main: running for the first time...");
      
      /* so we load the connection settings too */
      gj_gtk_roster_load();
      gj_gtk_ftr_load();
    }
  else
    {
      /* roster */
      gj_gtk_roster_load();
      g_message("main: initiated roster");
    }
 
  /* start event loop */
  g_message("main: ready!");
  gtk_main();

  /* module termination */
  gnome_jabber_term_modules();
  g_message("main: terminated jabber modules");

#ifdef HAVE_GNOME
  /* sound */
  gnome_sound_shutdown();
  g_message("main: terminated gnome-sound");
#endif

  /* clean up variables */
  gnome_jabber_term();

  /* end application */
  g_message("exiting...");

  return 0;  
}

gboolean gnome_jabber_first_time_check()
{
  
    
  return FALSE;
}

#ifdef WIN32

gboolean gnome_jabber_files_check()
{
  G_CONST_RETURN gchar *home_dir = g_get_home_dir();
  gchar *config_dir = NULL;
  gchar *history_dir = NULL;
  gchar *file_config = NULL;
/*   gchar *file_css = NULL; */

  g_message("checks: home directory is:'%s'", home_dir);

  config_dir = g_build_path(G_DIR_SEPARATOR_S, home_dir, PATH_USERDIR, NULL);
  history_dir = g_build_path(G_DIR_SEPARATOR_S, home_dir, PATH_FULL_HISTORY, NULL);
  file_config = g_build_filename(home_dir, FILE_FULL_CONFIG, NULL);
/*   file_css = g_build_filename(home_dir, FILE_FULL_CSS, NULL); */

  /* if this directory does not exist, it has not been run by this user before */
  if(g_file_test(config_dir, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) == FALSE)
    {
      g_message("checks: directory '%s' does NOT exist, creating...", config_dir);

      if(mkdir(config_dir) != 0)
	{
	  g_warning("checks: could not create directory, an error occured");
	  return FALSE;
	}

      gnome_jabber_set_first_time_run(TRUE);
    }
  else
    g_message("checks: directory '%s' exists", config_dir);

  /* we create this because it will be the default history directory on first time runs */
  if(g_file_test(history_dir,  (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) == FALSE)
    {
      g_message("checks: directory '%s' does NOT exist, creating...", history_dir);
      
      if(mkdir(history_dir) != 0)
	{
	  g_warning("checks: could not create directory, an error occured");
	  return FALSE;
	}
    }
  else
    g_message("checks: directory '%s' exists", history_dir);

  /* if this file does not exist, it is possible this is a first time run */
  if(g_file_test(file_config,  (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) == FALSE)
    {
      g_message("checks: file '%s' does NOT exist", file_config);
      gnome_jabber_set_first_time_run(TRUE);
    }
  else
    g_message("checks: file '%s' exists", file_config);

  /* if this file does not exist, formatting will be the only issue with the history */
  if(g_file_test(FILE_CSS,  (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) == FALSE)
    g_message("checks: file '%s' does NOT exist", FILE_CSS);
  else
    g_message("checks: file '%s' exists", FILE_CSS);
  
  /* error if not found because the application is completely dependant on it. */
  if(g_file_test(FILE_GLADE,  (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) == FALSE)
    g_error("checks: file '%s' does NOT exist", FILE_GLADE);
  else
    g_message("checks: file '%s' exists", FILE_GLADE);

  /* save */
  gnome_jabber_set_config_file(file_config);
  gnome_jabber_set_glade_file(FILE_GLADE);
  gnome_jabber_set_config_dir(config_dir);

  /* clean up */
  if(config_dir != NULL) g_free(config_dir);
  if(history_dir != NULL) g_free(history_dir);
  if(file_config != NULL) g_free(file_config);
/*   if(file_css != NULL) g_free(file_css); */

  return TRUE;
}

#else

gboolean gnome_jabber_files_check()
{
  G_CONST_RETURN gchar *home_dir = g_get_home_dir();
  gchar *config_dir = NULL;
  gchar *history_dir = NULL;
  gchar *file_config = NULL;
/*   gchar *file_css = NULL; */

  g_message("checks: home directory is:'%s'", home_dir);

  config_dir = g_build_path(G_DIR_SEPARATOR_S, home_dir, PATH_USERDIR, NULL);
  history_dir = g_build_path(G_DIR_SEPARATOR_S, home_dir, PATH_FULL_HISTORY, NULL);
  file_config = g_build_filename(home_dir, FILE_FULL_CONFIG, NULL);
/*   file_css = g_build_filename(home_dir, FILE_FULL_CSS, NULL); */

  /* if this directory does not exist, it has not been run by this user before */
  if(g_file_test(config_dir, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) == FALSE)
    {
      g_message("checks: directory '%s' does NOT exist, creating...", config_dir);

      if(mkdir(config_dir, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) != 0)
	{
	  g_warning("checks: could not create directory, an error occured");
	  return FALSE;
	}

      gnome_jabber_set_first_time_run(TRUE);
    }
  else
    g_message("checks: directory '%s' exists", config_dir);

  /* we create this because it will be the default history directory on first time runs */
  if(g_file_test(history_dir,  (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) == FALSE)
    {
      g_message("checks: directory '%s' does NOT exist, creating...", history_dir);
      
      if(mkdir(history_dir, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) != 0)
	{
	  g_warning("checks: could not create directory, an error occured");
	  return FALSE;
	}
    }
  else
    g_message("checks: directory '%s' exists", history_dir);

  /* if this file does not exist, it is possible this is a first time run */
  if(g_file_test(file_config,  (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) == FALSE)
    {
      g_message("checks: file '%s' does NOT exist", file_config);
      gnome_jabber_set_first_time_run(TRUE);
    }
  else
    g_message("checks: file '%s' exists", file_config);

  /* if this file does not exist, formatting will be the only issue with the history */
  if(g_file_test(FILE_FULL_CSS,  (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) == FALSE)
    g_message("checks: file '%s' does NOT exist", FILE_FULL_CSS);
  else
    g_message("checks: file '%s' exists", FILE_FULL_CSS);
  
  /* error if not found because the application is completely dependant on it. */
  if(g_file_test(FILE_FULL_GLADE,  (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) == FALSE)
    g_error("checks: file '%s' does NOT exist", FILE_FULL_GLADE);
  else
    g_message("checks: file '%s' exists", FILE_FULL_GLADE);

  /* save */
  gnome_jabber_set_config_file(file_config);
  gnome_jabber_set_glade_file(FILE_FULL_GLADE);
  gnome_jabber_set_config_dir(config_dir);

  /* clean up */
  if(config_dir != NULL) g_free(config_dir);
  if(history_dir != NULL) g_free(history_dir);
  if(file_config != NULL) g_free(file_config);
/*   if(file_css != NULL) g_free(file_css); */

  return TRUE;
}

#endif

/*
 * member functions 
 */

/* gets */
gchar *gnome_jabber_get_config_file()
{
  return gnome_jabber.file_config;
}

gchar *gnome_jabber_get_glade_file()
{
  return gnome_jabber.file_glade;
}

gchar *gnome_jabber_get_config_dir()
{
  return gnome_jabber.dir_config;
}

gboolean gnome_jabber_get_first_time_run()
{
  return gnome_jabber.first_time_run;
}

gboolean gnome_jabber_get_debugging()
{
  return gnome_jabber.debugging;
}

const gchar **gnome_jabber_get_server_list()
{
  return gnome_jabber_servers;
}

/* sets */
gboolean gnome_jabber_set_config_file(gchar *filename)
{
  if(filename == NULL)
    {
      g_warning("gnome_jabber_set_config_file: filename was NULL");
      return FALSE;
    }

  if(gnome_jabber.file_config != NULL)
    {
      g_free(gnome_jabber.file_config);
      gnome_jabber.file_config = NULL;
    }

  gnome_jabber.file_config = g_strdup(filename);
  return TRUE;
}

gboolean gnome_jabber_set_glade_file(gchar *filename)
{
  if(filename == NULL)
    {
      g_warning("gnome_jabber_set_glade_file: filename was NULL");
      return FALSE;
    }

  if(gnome_jabber.file_glade != NULL)
    {
      g_free(gnome_jabber.file_glade);
      gnome_jabber.file_glade = NULL;
    }

  gnome_jabber.file_glade = g_strdup(filename);
  return TRUE;
}

gboolean gnome_jabber_set_config_dir(gchar *dir)
{
  if(dir == NULL)
    {
      g_warning("gnome_jabber_set_config_dir: dir was NULL");
      return FALSE;
    }

  if(gnome_jabber.dir_config != NULL)
    {
      g_free(gnome_jabber.dir_config);
      gnome_jabber.dir_config = NULL;
    }

  gnome_jabber.dir_config = g_strdup(dir);
  return TRUE;
}

gboolean gnome_jabber_set_first_time_run(gboolean first_time_run)
{
  gnome_jabber.first_time_run = first_time_run;
  return TRUE;
}

gboolean gnome_jabber_set_debugging(gboolean debugging)
{
  gnome_jabber.debugging = debugging;
  return TRUE;
}

gboolean gnome_jabber_init()
{
  gnome_jabber.file_config = NULL;
  gnome_jabber.file_glade = NULL;
  gnome_jabber.dir_config = NULL;

  gnome_jabber.first_time_run = FALSE;

  gnome_jabber.debugging = FALSE;

  return TRUE;
}

gboolean gnome_jabber_term()
{
  if(gnome_jabber.file_config != NULL)
    {
      g_free(gnome_jabber.file_config);
      gnome_jabber.file_config = NULL;
    }
    
  if(gnome_jabber.file_glade != NULL)
    {
      g_free(gnome_jabber.file_glade);
      gnome_jabber.file_glade = NULL;
    }
    
  if(gnome_jabber.dir_config != NULL)
    {
      g_free(gnome_jabber.dir_config);
      gnome_jabber.dir_config = NULL;
    }

  gnome_jabber.first_time_run = FALSE;
 
  gnome_jabber.debugging = FALSE;
  
  return TRUE;
}

gboolean gnome_jabber_init_modules()
{
  /* local initiations */
  if(! gj_me_init())
    return FALSE;
  if(! gj_autoaway_init())
    return FALSE;
  if(! gj_connection_init())
    return FALSE;
  if(! gj_stream_init())
    return FALSE;
  if(! gj_iq_init())
    return FALSE;
  if(! gj_roster_init())
    return FALSE;
  if(! gj_event_init())
    return FALSE;
  if(! gj_presence_init())
    return FALSE;
  if(! gj_message_init())
    return FALSE;
  if(! gj_history_init())
    return FALSE;
  if(! gj_gatherdata_init())
    return FALSE;
  if(! gj_registration_init())
    return FALSE;
/*   if(! searchInit()) */
/*     return FALSE; */

  /* graphical initiations */
  if(! gj_gtk_ui_init())
    return FALSE;

  return TRUE;
}

gboolean gnome_jabber_term_modules()
{
  /* NOTE: these modules have to be terminated in a specific order.  
     if not, the program enters a continuous loop for some reason
     and fails to REDRAW??

     eg. this can be replicated by putting the tcpsocketlistterm function
     before the stream term.  

     i suspect this is because the tcp module is closed and then a tcp 
     socket is attmpted to be freed after.

     this is the same for the presence/roster combination.
  */

  /* local ternination */
  if(! gj_stream_term())
    return FALSE;
  if(! gj_connection_term())
    return FALSE;
  if(! gj_iq_term())
    return FALSE;
  if(! gj_roster_term())
    return FALSE;
  if(! gj_event_term())
    return FALSE;
  if(! gj_presence_term())
    return FALSE;
  if(! gj_message_term())
    return FALSE;
  if(! gj_history_term())
    return FALSE;
  if(! gj_gatherdata_term())
    return FALSE;
  if(! gj_registration_term())
    return FALSE;
/*   if(! searchTerm()) */
/*     return FALSE; */
  if(! gj_me_term())
    return FALSE;

  /* graphical initiations */
  if(! gj_gtk_ui_term())
    return FALSE;

  return TRUE;
}

gboolean gnome_jabber_quit()
{
  g_message("received quit request");
  gtk_main_quit();

  return TRUE;
}


#ifdef __cplusplus
}
#endif



