/* 
 * $RCSfile: gj_message.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/01 20:40:40 $
 * 
 * $Revision: 1.15 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <stdlib.h>
#include <string.h>

#include <time.h>

#include <glib.h>

#include <libxml/tree.h>

#include "gj_message.h"
#include "gj_parser.h"
#include "gj_me.h"
#include "gj_roster.h"
#include "gj_support.h"


struct t_message
{
  messageType type;
  
  gint requestID;

  gchar *id;

  gchar *to_jid;
  gchar *to_resource;
  gchar *from_jid;
  gchar *from_resource;
  
  gchar *subject;
  gchar *body;

  gchar *thread;

  /* x name spaces */
  gint xevent_type;
  gchar *xevent_message_id;

  gchar *xoob_url;
  gchar *xoob_description;

  /* errors */
  gchar *error_code;
  gchar *error_reason;

  time_t time;
};

/* to details, i.e. my state */
messageToCallback messageToGetResponseProc();

/* from details, i.e. my contacts state(s) */
messageFromCallback gj_message_get_cb_from();

/* response procs */


/* message item */
message gj_message_new(gint requestID, messageType type);
void gj_message_destroy(message msg);

/* message list */
gboolean gj_message_list_add(message msg);
gboolean gj_message_list_del(message msg);


GList *message_list = NULL;   /* stores ALL messages of ALL types */

messageFromCallback message_cb_from;
messageToCallback message_cb_to;   /* may not get used */


/*
 * code
 */
gboolean gj_message_init()
{
  return TRUE;
}

gboolean gj_message_term()
{
  gint index = 0;

  /* list */
  if(message_list == NULL)
    return TRUE;

  /* free all items on the list */
  for(index=0;index<g_list_length(message_list);index++)
    {
      message msg = NULL;
      
      if((msg = (message) g_list_nth_data(message_list,index)) != NULL)
	gj_message_free(msg);
    }

  g_list_free(message_list);
  message_list = NULL;

  return TRUE;
}

gboolean gj_message_list_add(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  message_list = g_list_append(message_list,msg);

  return TRUE;
}

gboolean gj_message_list_del(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  message_list = g_list_remove(message_list,msg);
  return TRUE;
}

message gj_message_new(gint requestID, messageType type)
{
  message msg = NULL;

  if(requestID < 1)
    {
      g_warning("%s: request id:%d was < 1", __FUNCTION__, requestID);
      return NULL;
    }
  
  if((msg = g_new(struct t_message,1)) != NULL)
    {
      msg->type = type;

      msg->requestID = requestID;

      msg->id = NULL;

      msg->to_jid = NULL;
      msg->to_resource = NULL;

      msg->from_jid = NULL;
      msg->from_resource = NULL;
      
      msg->subject = NULL;
      msg->body = NULL;

      msg->thread = NULL;

      msg->xevent_type = 0;
      msg->xevent_message_id = NULL;

      msg->xoob_url = NULL;
      msg->xoob_description = NULL;

      msg->error_code = NULL;
      msg->error_reason = NULL;

      msg->time = time(0);

      gj_message_list_add(msg);
    }
  
  return msg;  
}

void gj_message_free(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return;
    }

  /* if on queue, remove it */
  

  /* free mem */
  gj_message_list_del(msg);  

  /* destroy members */
  gj_message_destroy(msg);

  /* free structure */
  g_free(msg);
}

void gj_message_destroy(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return;
    }

  msg->type = 0;

  msg->requestID = 0;

  if(msg->id != NULL)
    {
      g_free(msg->id);
      msg->id = NULL;
    }

  if(msg->to_jid != NULL)
    {
      g_free(msg->to_jid);
      msg->to_jid = NULL;
    }

  if(msg->to_resource != NULL)
    {
      g_free(msg->to_resource);
      msg->to_resource = NULL;
    }

  if(msg->from_jid != NULL)
    {
      g_free(msg->from_jid);
      msg->from_jid = NULL;
    }

  if(msg->from_resource != NULL)
    {
      g_free(msg->from_resource);
      msg->from_resource = NULL;
    }  

  if(msg->subject != NULL)
    {
      g_free(msg->subject);
      msg->subject = NULL;
    }

  if(msg->body != NULL)
    {
      g_free(msg->body);
      msg->body = NULL;
    } 

  if(msg->thread != NULL)
    {
      g_free(msg->thread);
      msg->thread = NULL;
    } 

  msg->xevent_type = 0;

  if(msg->xevent_message_id != NULL)
    {
      g_free(msg->xevent_message_id);
      msg->xevent_message_id = NULL;
    }

  if(msg->xoob_url != NULL)
    {
      g_free(msg->xoob_url);
      msg->xoob_url = NULL;
    }

  if(msg->xoob_description != NULL)
    {
      g_free(msg->xoob_description);
      msg->xoob_description = NULL;
    }

  if(msg->error_code != NULL)
    {
      g_free(msg->error_code);
      msg->error_code = NULL;
    } 

  if(msg->error_reason != NULL)
    {
      g_free(msg->error_reason);
      msg->error_reason = NULL;
    } 
}

/* find functions */
message gj_message_find_by_to_jid(gchar *jid)
{
  gint index = 0;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(message_list);index++)
    {
      message msg  = g_list_nth_data(message_list, index);

      if(msg == NULL) continue;
      if(g_strcasecmp(msg->to_jid,jid) == 0) return msg;
    }

  return NULL;
}

/*
 * member functions 
 */

/* sets */
gboolean gj_message_set_to_jid(message msg, gchar *jid)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->to_jid != NULL)
    g_free(msg->to_jid);

  msg->to_jid = g_strdup(jid);
  return TRUE;
}

gboolean gj_message_set_type(message msg, messageType type)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(type < 1 || type > messageTypeEnd)
    {
      g_warning("%s: type was out of range", __FUNCTION__);
      return FALSE;
    }

  msg->type = type;
  return TRUE;
}

gboolean gj_message_set_id(message msg, gchar *id)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->id != NULL)
    g_free(msg->id);

  msg->id = g_strdup(id);
  return TRUE;
}

gboolean gj_message_set_to_resource(message msg, gchar *resource)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(resource == NULL)
    {
      g_warning("%s: resource was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->to_resource != NULL)
    g_free(msg->to_resource);

  msg->to_resource = g_strdup(resource);
  return TRUE;
}

gboolean gj_message_set_from_jid(message msg, gchar *jid)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->from_jid != NULL)
    g_free(msg->from_jid);

  msg->from_jid = g_strdup(jid);
  return TRUE;
}

gboolean gj_message_set_from_resource(message msg, gchar *resource)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(resource == NULL)
    {
      g_warning("%s: resource was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->from_resource != NULL)
    g_free(msg->from_resource);

  msg->from_resource = g_strdup(resource);
  return TRUE;
}

gboolean gj_message_set_subject(message msg, gchar *subject)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(subject == NULL)
    {
      g_warning("%s: subject was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->subject != NULL)
    g_free(msg->subject);

  msg->subject = g_strdup(subject);
  return TRUE;
}

gboolean gj_message_set_body(message msg, gchar *body)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(body == NULL)
    {
      g_warning("%s: body was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->body != NULL)
    g_free(msg->body);

  msg->body = g_strdup(body);
  return TRUE;
}

gboolean gj_message_set_thread(message msg, gchar *thread)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(thread == NULL)
    {
      g_warning("%s: thread was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->thread != NULL)
    g_free(msg->thread);

  msg->thread = g_strdup(thread);
  return TRUE;
}

gboolean gj_message_set_xevent_type(message msg, gint type)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(type == 0)
    {
      g_warning("%s: type was out of range", __FUNCTION__);
      return FALSE;
    }

  msg->xevent_type = type;
  return TRUE;
}

gboolean gj_message_set_xevent_message_id(message msg, gchar *message_id)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(message_id == NULL)
    {
      g_warning("%s: message id was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->xevent_message_id != NULL)
    g_free(msg->xevent_message_id);

  msg->xevent_message_id = g_strdup(message_id);
  return TRUE;
}

gboolean gj_message_set_xoob_url(message msg, gchar *url)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(url == NULL)
    {
      g_warning("%s: url was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->xoob_url != NULL)
    g_free(msg->xoob_url);

  msg->xoob_url = g_strdup(url);
  return TRUE;
}

gboolean gj_message_set_xoob_description(message msg, gchar *description)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(description == NULL)
    {
      g_warning("%s: description was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->xoob_description != NULL)
    g_free(msg->xoob_description);

  msg->xoob_description = g_strdup(description);
  return TRUE;
}

gboolean gj_message_set_error_code(message msg, gchar *error_code)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(error_code == NULL)
    {
      g_warning("%s: error code was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->error_code != NULL)
    g_free(msg->error_code);

  msg->error_code = g_strdup(error_code);
  return TRUE;
}

gboolean gj_message_set_error_reason(message msg, gchar *error_reason)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(error_reason == NULL)
    {
      g_warning("%s: error reason was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg->error_reason != NULL)
    g_free(msg->error_reason);

  msg->error_reason = g_strdup(error_reason);
  return TRUE;
}

/* gets */
messageType gj_message_get_type(message msg)
{ 
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return -1;
    }  

  return msg->type;
}

gchar *gj_message_get_id(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->id;
}

gint gj_message_get_request_id(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return -1;
    }  

  return msg->requestID;
}

gchar *gj_message_get_to_jid(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->to_jid;
}

gchar *gj_message_get_to_resource(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->to_resource;
}

gchar *gj_message_get_from_jid(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->from_jid;
}

gchar *gj_message_get_from_resource(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->from_resource;
}

gchar *gj_message_get_subject(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->subject;
}

gchar *gj_message_get_body(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->body;
}

gchar *gj_message_get_thread(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->thread;
}

gint gj_message_get_xevent_type(message msg)
{ 
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return -1;
    }  

  return msg->xevent_type;
}

gchar *gj_message_get_xevent_message_id(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->xevent_message_id;
}

gchar *gj_message_get_xoob_url(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->xoob_url;
}

gchar *gj_message_get_xoob_description(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->xoob_description;
}

gchar *gj_message_get_error_code(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }  

  return msg->error_code;
}

gchar *gj_message_get_error_reason(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }  

  return msg->error_reason;
}

time_t gj_message_get_time(message msg)
{
  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return 0;
    }  

  return msg->time;
}

/*
 * callbacks
 */
gboolean gj_message_set_cb_from(messageFromCallback cb)
{
  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  message_cb_from = cb;
  
  return TRUE;
}

messageFromCallback gj_message_get_cb_from()
{
  return message_cb_from;
}

gboolean gj_message_set_cb_to(messageToCallback cb)
{
  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  message_cb_to = cb;
  
  return TRUE;
}

messageToCallback gj_message_get_cb_to()
{
  return message_cb_to;
}


/*
 * external functions
 */
gboolean gj_message_send(messageType type, gchar *to, gchar *subject, gchar *body, gchar *thread)
{
  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *id = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(body == NULL)
    {
      g_warning("%s: body was NULL", __FUNCTION__);
      return FALSE;
    }

  id = g_strdup_printf("%d", gj_support_get_unique_id());

  node = xmlNewNode(NULL, (guchar*)"message");

  xmlSetProp(node, (guchar*)"id", (guchar*)id);
  xmlSetProp(node, (guchar*)"to", (guchar*)to);
  xmlSetProp(node, (guchar*)"from", (guchar*)gj_me_get_jid());

  switch(type)
    {
    case messageTypeNormal:
      xmlSetProp(node, (guchar*)"type", (guchar*)"normal");
      break;
    case messageTypeChat:
      xmlSetProp(node, (guchar*)"type", (guchar*)"chat");
      break;
    case messageTypeGroupChat:
      xmlSetProp(node, (guchar*)"type", (guchar*)"groupchat");
      break;
    case messageTypeHeadline:
      xmlSetProp(node, (guchar*)"type", (guchar*)"headline");
      break;
    case messageTypeError:
      xmlSetProp(node, (guchar*)"type", (guchar*)"error");
      break;
      
    default:
      g_warning("%s: message type:%d was unknown", __FUNCTION__, type);
      return FALSE;
    }

  if(subject != NULL)
    {
      gchar *formatted = NULL;
      formatted = g_markup_escape_text(subject, -1);
      sub = xmlNewChild(node, ns, (guchar*)"subject", (guchar*)formatted);
      if(formatted != NULL) g_free(formatted);
    }

  if(body != NULL)
    {
      gchar *formatted = NULL;
      formatted = g_markup_escape_text(body, -1);
      sub = xmlNewChild(node, ns, (guchar*)"body", (guchar*)formatted);
      if(formatted != NULL) g_free(formatted);
    }

  if(thread != NULL)
    {
      gchar *formatted = NULL;
      formatted = g_markup_escape_text(thread, -1);
      sub = xmlNewChild(node, ns, (guchar*)"thread", (guchar*)formatted);
      if(formatted != NULL) g_free(formatted);
    }

  gj_parser_push_output(node);

  /* clean up */
  if(id != NULL) g_free(id);

  return TRUE;
}

gboolean gj_message_send_xevent(gint xevents, gchar *to, gchar *id)
{
  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *event = NULL;
  gchar *msg = NULL;

  if(xevents == 0)
    {
      g_warning("%s: invalid xevent(s)", __FUNCTION__);
      return FALSE;
    }

/*   if(id == NULL) */
/*     { */
/*       g_warning("%s: id was NULL", __FUNCTION__); */
/*       return FALSE; */
/*     } */
      

  if(!(xevents & messageXEventTypeOffline) && 
     !(xevents & messageXEventTypeDelivered) &&
     !(xevents & messageXEventTypeDisplayed) &&   
     !(xevents & messageXEventTypeComposing) &&
     !(xevents & messageXEventTypeCancelled))
    {
      g_warning("%s: invalid xevent type", __FUNCTION__);
      return FALSE;
    }

  node = xmlNewNode(NULL, (guchar*)"message");

  xmlSetProp(node, (guchar*)"to", (guchar*)to);
  xmlSetProp(node, (guchar*)"from", (guchar*)gj_me_get_jid());

  sub = xmlNewNode(NULL, (guchar*)"x");
  ns = xmlNewNs(sub, (guchar*)"jabber:x:event", NULL);

  if(!(xevents & messageXEventTypeCancelled))
  {
      if(xevents & messageXEventTypeOffline)
	  xmlNewChild(sub, ns, (guchar*)"offline", NULL);
      if(xevents & messageXEventTypeDelivered)
	  xmlNewChild(sub, ns, (guchar*)"delivered", NULL);
      if(xevents & messageXEventTypeDisplayed)
	  xmlNewChild(sub, ns, (guchar*)"displayed", NULL);
      if(xevents & messageXEventTypeComposing)
	  xmlNewChild(sub, ns, (guchar*)"composing", NULL);
  }

  if(id != NULL)
      xmlNewChild(sub, ns, (guchar*)"id", id);

  xmlAddChild(node, sub);

  /* send */
  gj_parser_push_output(node);

  /* clean up */
  if(event != NULL) g_free(event);
  if(msg != NULL) g_free(msg);

  return TRUE;
}

gchar *gj_message_send_with_xevent_request(messageType type, gchar *to, gchar *subject, gchar *body, gchar *thread, gint xevents)
{
  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *msgtype = NULL;
  gchar *id = NULL;


  if(type < 1 || type > messageTypeEnd)
    {
      g_warning("%s: message type out of range", __FUNCTION__);
      return NULL;
    }

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return NULL;
    }

  if(body == NULL)
    {
      g_warning("%s: body was NULL", __FUNCTION__);
      return NULL;
    }

  if(!(xevents & messageXEventTypeOffline) && 
     !(xevents & messageXEventTypeDelivered) &&
     !(xevents & messageXEventTypeDisplayed) &&   
     !(xevents & messageXEventTypeComposing) &&
     !(xevents & messageXEventTypeCancelled))
    {
      g_warning("%s: invalid xevent type", __FUNCTION__);
      return NULL;
    }

  switch(type)
    {
    case messageTypeNormal:
      msgtype = g_strdup_printf("normal");
      break;
    case messageTypeChat:
      msgtype = g_strdup_printf("chat");
      break;
    case messageTypeGroupChat:
      msgtype = g_strdup_printf("groupchat");
      break;
    case messageTypeHeadline:
      msgtype = g_strdup_printf("headline");
      break;
    case messageTypeError:
      msgtype = g_strdup_printf("error");
      break;

    default:
      g_warning("%s: message type:%d was unknown", __FUNCTION__, type);
      return NULL;
    }

  id = g_strdup_printf("%d", gj_support_get_unique_id());

  node = xmlNewNode(NULL, (guchar*)"message");

  xmlSetProp(node, (guchar*)"id", (guchar*)id);
  xmlSetProp(node, (guchar*)"to", (guchar*)to);
  xmlSetProp(node, (guchar*)"from", (guchar*)gj_me_get_jid());
  xmlSetProp(node, (guchar*)"type", (guchar*)msgtype);

  if(subject != NULL)
    {
      gchar *formatted = NULL;
      formatted = g_markup_escape_text(subject, -1);
      sub = xmlNewChild(node, ns, (guchar*)"subject", (guchar*)formatted);
      if(formatted != NULL) g_free(formatted);
    }

  if(body != NULL)
    {
      gchar *formatted = NULL;
      formatted = g_markup_escape_text(body, -1);
      sub = xmlNewChild(node, ns, (guchar*)"body", (guchar*)formatted);
      if(formatted != NULL) g_free(formatted);
    }

  if(thread != NULL)
    {
      gchar *formatted = NULL;
      formatted = g_markup_escape_text(thread, -1);
      sub = xmlNewChild(node, ns, (guchar*)"thread", (guchar*)formatted);
      if(formatted != NULL) g_free(formatted);
    }

  sub = xmlNewNode(NULL, (guchar*)"x");
  ns = xmlNewNs(sub, (guchar*)"jabber:x:event", NULL);

  if(!(xevents & messageXEventTypeCancelled))
  {
      if(xevents & messageXEventTypeOffline)
	  xmlNewChild(sub, ns, (guchar*)"offline", NULL);
      if(xevents & messageXEventTypeDelivered)
	  xmlNewChild(sub, ns, (guchar*)"delivered", NULL);
      if(xevents & messageXEventTypeDisplayed)
	  xmlNewChild(sub, ns, (guchar*)"displayed", NULL);
      if(xevents & messageXEventTypeComposing)
	  xmlNewChild(sub, ns, (guchar*)"composing", NULL);
  }

  xmlAddChild(node, sub);

  gj_parser_push_output(node);

  /* clean up */
  if(msgtype != NULL) g_free(msgtype);

  return id;    
}


void gj_message_receive(connection c, xmlNodePtr node)
{
  gint requestID = gj_support_get_unique_id();

  gchar *message_id = NULL;
  
  gchar *to_user = NULL;
  gchar *to_resource = NULL;

  gchar *from_user = NULL;
  gchar *from_resource = NULL;

  gchar *subject = NULL;
  gchar *body = NULL;

  gchar *thread = NULL;

  xmlNodePtr x = NULL;

  gint xevent_type = 0;
  gchar *xevent_message_id = NULL;

  gchar *xoob_url = NULL;
  gchar *xoob_description = NULL;

  gchar *error_code = NULL;
  gchar *error_reason = NULL;

  xmlChar *stype = NULL;
  xmlChar *tmp = NULL;

  message msg = NULL;

  messageType type = 0;
/*   messageType type = messageTypeNormal; */
  messageFromCallback callback = NULL;

  /* checks */
  if(node == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return;
    }

  /* get info */
  if((callback = gj_message_get_cb_from()) == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return;
    }
    
  if((stype = gj_parser_find_by_attr_and_ret(node, (guchar*)"type")) != NULL)
    {
      if(xmlStrcmp(stype, (guchar*)"normal") == 0)
	type = messageTypeNormal;
      else if(xmlStrcmp(stype, (guchar*)"chat") == 0)
	type = messageTypeChat;
      else if(xmlStrcmp(stype, (guchar*)"groupchat") == 0)
	type = messageTypeGroupChat;
      else if(xmlStrcmp(stype, (guchar*)"headline") == 0)
	type = messageTypeHeadline;
      else if(xmlStrcmp(stype, (guchar*)"error") == 0)
	type = messageTypeError;
    }

  if((x = gj_parser_find_by_node(node, (guchar*)"x")) != NULL)
    {
	if(type == 0)
	    type = messageTypeXEvent;

      if(xmlStrcmp(gj_parser_find_by_ns_and_ret(x), (guchar*)"jabber:x:event") == 0)
	{
	  if(gj_parser_find_by_node(x, (guchar*)"offline") != NULL)
	    xevent_type |= messageXEventTypeOffline;
	  if(gj_parser_find_by_node(x, (guchar*)"delivered") != NULL)
	    xevent_type |= messageXEventTypeDelivered;
	  if(gj_parser_find_by_node(x, (guchar*)"displayed") != NULL)
	    xevent_type |= messageXEventTypeDisplayed;
	  if(gj_parser_find_by_node(x, (guchar*)"composing") != NULL)
	    xevent_type |= messageXEventTypeComposing;
	  
	  if(xevent_type == 0)
	    xevent_type = messageXEventTypeCancelled;
	  
	  if((tmp = gj_parser_find_by_node_and_ret(x, (guchar*)"id")) != NULL) 
	    xevent_message_id = g_strdup_printf("%s",tmp);      

#if 0	  
	  /* set message type */
	  if(xevent_message_id != NULL)
	    type = messageTypeXEvent;
#endif
	}
      else if(xmlStrcmp(gj_parser_find_by_ns_and_ret(x), (guchar*)"jabber:x:oob") == 0)
	{
	  if((tmp = gj_parser_find_by_node_and_ret(x, (guchar*)"url")) != NULL) 
	    xoob_url = g_strdup_printf("%s",tmp);    
	  if((tmp = gj_parser_find_by_node_and_ret(x, (guchar*)"desc")) != NULL) 
	    xoob_description = g_strdup_printf("%s",tmp);    
	}
    }

  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"subject")) != NULL)
      subject = g_strdup((gchar*)tmp);
  
  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"body")) != NULL)
      body = g_strdup((gchar*)tmp);
  
  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"thread")) != NULL)
      thread = g_strdup((gchar*)tmp);
  
  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"id")) != NULL) 
      message_id = g_strdup((gchar*)tmp);

  /* get 'to' field and separate resource and user */
  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"to")) != NULL && xmlStrlen(tmp) > 0)
    {
      gchar *to = g_strdup_printf("%s",tmp);
      gchar **vTo = g_strsplit(to,"/",2);
      
      if(vTo != NULL)
	{
	  to_user = g_strdup(to);
	  
	  if(vTo[1] != NULL)
	    to_resource = g_strdup(vTo[1]);
	  
	  g_strfreev(vTo);
	}
      
      if(to != NULL) g_free(to);
    }
  
  /* get 'from' field and separate resource and user */
  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"from")) != NULL && xmlStrlen(tmp) > 0)
    {
      gchar *from = g_strdup_printf("%s",tmp);
      gchar **vFrom = g_strsplit(from,"/",2);
      
      if(vFrom != NULL)
	{
	  from_user = g_strdup(from);
	  
	  if(vFrom[1] != NULL)
	    from_resource = g_strdup(vFrom[1]);
	  
	  g_strfreev(vFrom);
	}

      if(from != NULL) g_free(from);
    }

  /* get error */
  if(type == messageTypeError)
    {
      xmlChar *tmp = NULL;
      xmlNodePtr error = NULL;

      if((error = gj_parser_find_by_node(node, (guchar*)"error")) != NULL)
	{
	  if((tmp = gj_parser_find_by_attr_and_ret(error, (guchar*)"code")) != NULL)
	    error_code = g_strdup_printf("%s",tmp);

	  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"error")) != NULL) 
	    error_reason = g_strdup_printf("%s",tmp);
	}
    }

  /* FIXME: not sure if this is right but its the best way at the moment that i 
     can think of for fixing the problem */
  if(body == NULL && xevent_type != 0)
    type = messageTypeXEvent;

  /*
   * create our message struct
   */
  msg = gj_message_new(requestID, type); 

  if(msg == NULL)
    {
      g_warning("%s: could not create new message", __FUNCTION__);
      return;
    }

  /* associate with roster item */
  if(message_id) gj_message_set_id(msg,message_id);
  if(to_user) gj_message_set_to_jid(msg,to_user);
  if(to_resource) gj_message_set_to_resource(msg,to_resource);
  if(from_user) gj_message_set_from_jid(msg,from_user);
  if(from_resource) gj_message_set_from_resource(msg,from_resource);
  if(subject) gj_message_set_subject(msg,subject);
  if(body) gj_message_set_body(msg,body);
  if(thread) gj_message_set_thread(msg,thread);

  if(xevent_type != 0) gj_message_set_xevent_type(msg, xevent_type);
  if(xevent_message_id) gj_message_set_xevent_message_id(msg, xevent_message_id);
  if(xoob_url) gj_message_set_xoob_url(msg, xoob_url);
  if(xoob_description) gj_message_set_xoob_description(msg, xoob_description);

  if(error_code) gj_message_set_error_code(msg,error_code);
  if(error_reason) gj_message_set_error_reason(msg,error_reason);

  /* call callback function */
  (callback)(msg);

  /* free mem */
  if(subject != NULL) g_free(subject);
  if(body != NULL) g_free(body);
  if(to_user != NULL) g_free(to_user);
  if(to_resource != NULL) g_free(to_resource);
  if(from_user != NULL) g_free(from_user);
  if(from_resource != NULL) g_free(from_resource);

  if(xevent_message_id != NULL) g_free(xevent_message_id);
  if(xoob_url != NULL) g_free(xoob_url);
  if(xoob_description != NULL) g_free(xoob_description);

  if(error_code != NULL) g_free(error_code);
  if(error_reason != NULL) g_free(error_reason);
}


#ifdef __cplusplus
}
#endif
