/* 
 * $RCSfile: gj_parser.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/28 21:21:34 $
 * 
 * $Revision: 1.14 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <glib.h>

#include <libxml/parser.h>
#include <libxml/parserInternals.h>

#include "gj_parser.h"
#include "gj_connection.h"
#include "gj_stream.h"
#include "gj_iq.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_me.h"

/* locals */
gboolean gj_parser_parse_type(connection c, xmlNodePtr node);

gboolean gj_parser_input_check(gchar *buf);
gboolean gj_parser_input_node(connection c, xmlNodePtr node, gboolean stream);

const gchar *gj_parser_print_tree_spaces(gint num);

/* parser callbacks */
void gj_parser_doc_start_cb(void *user_data);
void gj_parser_doc_end_cb(void *user_data);
void gj_parser_doc_element_start_cb(void *user_data, const xmlChar *name, const xmlChar **attrs);
void gj_parser_doc_element_end_sb(void *user_data, const xmlChar *name);
xmlEntityPtr gj_parser_doc_entity_cb(void *user_data, const xmlChar *name) ;
void gj_parser_doc_warning_cb(void *user_data, const char *msg, ...);
void gj_parser_doc_error_cb(void *user_data, const char *msg, ...);
void gj_parser_doc_fatal_error_cb(void *user_data, const char *msg, ...);

gboolean gj_parser_parse_xml(void *user_data, const gchar *input);

gchar *bufwaiting = NULL;

guint connection_id = 0;

/* global vars */
gint requestID = 69;
xmlParserCtxtPtr ctxt = NULL;
xmlSAXHandler sax = {
  0, /* internalSubset */
  0, /* isStandalone */
  0, /* hasInternalSubset */
  0, /* hasExternalSubset */
  0, /* resolveEntity */
  (getEntitySAXFunc)gj_parser_doc_entity_cb, /* getEntity */
  0, /* entityDecl */
  0, /* notationDecl */
  0, /* attributeDecl */
  0, /* elementDecl */
  0, /* unparsedEntityDecl */
  0, /* setDocumentLocator */
  (startDocumentSAXFunc)gj_parser_doc_start_cb, /* startDocument */
  (endDocumentSAXFunc)gj_parser_doc_end_cb, /* endDocument */
  (startElementSAXFunc)gj_parser_doc_element_start_cb, /* startElement */
  (endElementSAXFunc)gj_parser_doc_element_end_sb, /* endElement */
  0, /* reference */
  0, /*(charactersSAXFunc)gladeCharacters, *//* characters */
  0, /* ignorableWhitespace */
  0, /* processingInstruction */
  0, /* (commentSAXFunc)gladeComment, *//* comment */
  (warningSAXFunc)gj_parser_doc_warning_cb, /* warning */
  (errorSAXFunc)gj_parser_doc_error_cb, /* error */
  (fatalErrorSAXFunc)gj_parser_doc_fatal_error_cb, /* fatalError */
};


gboolean gj_parser_parse_type(connection c, xmlNodePtr node)
{
  const gchar *type = NULL;

  if(node == NULL)
    {
      g_warning("%s: node passed was NULL", __FUNCTION__);
      return FALSE;
    }

  /* get type */
  type = (const gchar*) gj_parser_get_element(node);

  if(g_strcasecmp(type,"stream") == 0)
    {
      gj_stream_receive(c, node);
      return TRUE;
    }
  else if(g_strcasecmp(type,"iq") == 0)
    {
      gj_iq_receive(c, node);
      return TRUE;
    }
  else if(g_strcasecmp(type,"message") == 0)
    {
      gj_message_receive(c, node);
      return TRUE;
    }
  else if(g_strcasecmp(type,"presence") == 0)
    {
      gj_presence_receive(c, node);
      return TRUE;
    }
  else if(g_strcasecmp(type,"error") == 0)
  {
      gj_stream_receive_error(c, node);
      return TRUE;
  }
  
  /* NOTE: we *REALLY* should ignore the previous buffers and 
     wait for the next message if this happens */
  g_warning("%s: unexpected type of msg from connection:0x%.8x", __FUNCTION__, (gint)c);
  return FALSE;
}

void gj_parser_push_input_from_connection(connection c, gchar *text)
{
    if(text == NULL)
    {
	g_warning("%s: text was NULL", __FUNCTION__);
	return;
    }

    g_message("%s: parsing %d bytes, %d characters from connection:0x%.8x",
	      __FUNCTION__, (gint)strlen(text), (gint)g_utf8_strlen(text, -1), (gint)c);
    gj_parser_push_input(c, text, strlen(text));
}

gboolean gj_parser_input_check(gchar *buf)
{
  gint index = 0;
  gint pos = 0;
  
  gchar *tagStart = NULL;
  gchar *tagEnd = NULL;
  
  /* check start has opening tag */
  if(buf[0] != '<')
    return FALSE;
  /* check end has closing tag */
  if(buf[strlen(buf)-1] != '>')
    return FALSE;
  
  /* we do not want to check tag names:
     what happens if the message is -

     <presence></presence><message></message>

     the tags will not match.
     this is just a basic test for opening and closing chevrons 
  */

  return TRUE;


  /* FIXME: NOTE the code below is NOT UTF8 checked and will probably crash */

  /* get start tag name */
  for(index=0,pos=0;index<strlen(buf);index++)
    {
      if((g_ascii_isspace(buf[index]) == TRUE || buf[index] == '>') && index > 1)
	{
	  pos = index;
	  tagStart = g_strndup(buf+1,pos-1);
	  break;
	}
    }

  if(tagStart != NULL)
    g_log("XML", G_LOG_LEVEL_DEBUG,"%s: found opening tag:'%s' at position:%d", __FUNCTION__, tagStart, pos);
  
  /* get end tag name */
  for(index=strlen(buf)-1,pos=0;index>0;index--)
    {
      if(buf[index] == '<')
	{
	  pos = index + 1;
	  tagEnd = g_strndup(buf+pos,pos-2);
	  break;
	}
      else if(buf[index] == '/' && buf[index-1] == '<')
	{
	  pos = index + 1;
	  tagEnd = g_strndup(buf+pos,strlen(buf) - pos - 1);
	  break;
	}
    }
  
  if(tagEnd != NULL)
    g_log("XML", G_LOG_LEVEL_DEBUG,"%s: found closing tag:'%s' at position:%d", __FUNCTION__, tagEnd, pos);      
  
  /* there may NOT be an end tag, 
     e.g. <iq type="result" to="joebloggs"/> */
  if(tagEnd != NULL && g_strcasecmp(tagStart,tagEnd) != 0)
    return FALSE;
  
  return TRUE;
}

void gj_parser_push_input(connection c, gchar *buf, gint nBytes)
{
  gint nodecount = 0;
  gboolean stream = FALSE;
  gchar *xml = NULL;

  xmlDocPtr doc = NULL;
  xmlNodePtr current = NULL;
  
  if(buf == NULL || nBytes < 1)
    {
      g_warning("%s: buf is NULL or size:%d is < 1", __FUNCTION__, nBytes);
      return;
    }

  /* check for stream start / end, xml needs patching to pass parser */
  if(g_strncasecmp(buf,"<stream",7) == 0 || g_strncasecmp(buf,"<?xml",5) == 0)
    {
      /* patch start of stream */
      gchar *footer = g_strdup_printf("</stream:stream>");
      xml = g_strconcat(buf,footer,NULL);
      g_free(footer);

      stream = TRUE;
    }
  else if(g_strncasecmp(buf,"</stream",8) == 0)
    {
      /* patch end of stream */
      gchar *header = g_strdup_printf("<stream:stream>");
      xml = g_strconcat(header,buf,NULL);
      g_free(header);

      stream = TRUE;
    }
  else
    {
      gchar *header = NULL;
      gchar *footer = NULL;

      if(bufwaiting != NULL)
	{
	  g_log("XML", G_LOG_LEVEL_DEBUG, "%s: buffer saved previously, concating to new buffer", __FUNCTION__);
	  buf = g_strconcat(bufwaiting,buf,NULL);
	  
	  g_log("XML", G_LOG_LEVEL_DEBUG, "%s: freeing saved buffer", __FUNCTION__);
	  g_free(bufwaiting);
	  bufwaiting = NULL;

	  /* NOW PARSER XML */
	}
      
      g_log("XML", G_LOG_LEVEL_DEBUG, "%s: now attempting to parse buffer", __FUNCTION__);
      
      /* used to make sure header and footer elements are unique 
	 if this is NOT done, the parser sometimes thinks a chunk
	 of xml is not balanced.  

	 e.g. <presence></presence><presence></presence> */
      header = g_strdup_printf("<gj>");
      footer = g_strdup_printf("</gj>");
      
      xml = g_strconcat(header,buf,footer,NULL);
      
      if(header != NULL) g_free(header);
      if(footer != NULL) g_free(footer);

      if(gj_parser_input_check(buf) == FALSE || gj_parser_parse_xml(NULL, xml) == FALSE)
	{
	  g_log("XML", G_LOG_LEVEL_DEBUG, "%s: xml was not well balanced, waiting for MORE input", __FUNCTION__);  
	  g_log("XML", G_LOG_LEVEL_DEBUG, "%s: saving buffer", __FUNCTION__);
	  bufwaiting = g_strdup(buf);

	  if(xml != NULL) g_free(xml);
	  
	  return;
	}
      else
	{
	  g_log("XML", G_LOG_LEVEL_DEBUG, "gj_parser_push_input: xml is ok, freeing previously saved buffer");
	  g_free(bufwaiting);
	  bufwaiting = NULL;
	}

#if 0
      /* print what we are about to parse */
      g_message("%s: parsing %d bytes (%d chars):\n<<<< %s", 
		__FUNCTION__, strlen(xml), g_utf8_strlen(xml, -1), xml);
#endif
    }

  doc = xmlSAXParseDoc(NULL, (xmlChar*)xml,1);  
  if(doc == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG, "%s: parser context was NULL, can not continue...", __FUNCTION__);
      return;
    }

  /* clean up */
  if(xml != NULL) g_free(xml);

  /* stream check */
  if(stream == TRUE)
    {
      gj_parser_input_node(c, doc->xmlChildrenNode,stream);
    }
  else
    { 
      if((nodecount = gj_parser_doc_count_top_level_nodes(doc)) < 1)
	{
	  g_warning("%s: there were %d nodes to process (> 0 required)", __FUNCTION__, nodecount);
	  return;
	}
      
      if(doc->xmlChildrenNode != NULL)
	current = doc->xmlChildrenNode->xmlChildrenNode;
      
      while(current != NULL)
	{
	  gj_parser_input_node(c, current,stream);
	  current = current->next;
	}
    }
}

gboolean gj_parser_input_node(connection c, xmlNodePtr node, gboolean stream)
{
  /* set node from doc */
  if(node != NULL)
    {
      xmlNodePtr myCopy = NULL;

      /* make a copy for us to play with */
      myCopy = xmlCopyNode(node,1);

      /* parser doc by type - i.e iq/message.... */
      if(myCopy != NULL)
	{
	  /* print tree */
 	  g_log("XML", G_LOG_LEVEL_DEBUG,"****** started printing tree node:0x%.8x *********",(gint)myCopy); 
 	  gj_parser_print_tree(myCopy,0); 
 	  g_log("XML", G_LOG_LEVEL_DEBUG,"****** stopped printing tree node:0x%.8x *********",(gint)myCopy); 


	  /*g_message("****** %d top nodes ******",gj_parser_doc_count_top_level_nodes(doc));*/
	  gj_parser_parse_type(c, myCopy);
	  return TRUE;
	}
      else
	g_warning("%s: failed to make a copy of node, copy was NULL!", __FUNCTION__);
    }
  else
    g_warning("%s: failed to continue, node was NULL", __FUNCTION__);

  return FALSE;
}

gboolean gj_parser_parse_xml(void *user_data, const gchar *input) 
{
  gboolean ret = 0;
  xmlParserCtxtPtr ctxt;

  ctxt = xmlCreatePushParserCtxt(&sax, user_data, input, strlen(input), NULL);
  if (ctxt == NULL)
    {
      g_warning("gj_parser_parse_xml: could not create push parser");
      return FALSE;      
    }

  xmlParseDocument(ctxt);
  
  if(ctxt->wellFormed)
    ret = TRUE;
  else
    ret = FALSE;

  if(&sax != NULL)
    ctxt->sax = NULL;

  xmlFreeParserCtxt(ctxt);
  
  return ret;
}

const gchar *gj_parser_print_tree_spaces(gint num)
{
  gchar *string = NULL;

  while(num-- > 0)
    {
      if(string != NULL)
	string = g_strconcat(string, "  ", NULL);
      else
	string = "  ";
    }

  if(string == NULL)
    string = "";

  return string;
}

xmlNodePtr gj_parser_print_tree(xmlNodePtr node, gint depth)
{
  xmlNodePtr result = NULL;
  gchar buffer[1024];

  if(node == NULL)                      
    return NULL;

  /* set buffer to spaces */
  memset(buffer, 32, sizeof(buffer));
  buffer[(depth*2)] = '\0';

  if(node->name != NULL)   
    {
/*       const gchar *spaces = NULL; */

      if(xmlStrcasecmp(node->name, (guchar*)"text") != 0 && xmlStrlen(node->name) > 0)
	{
/* 	  spaces = gj_parser_print_tree_spaces(depth); */
	  g_log("XML", G_LOG_LEVEL_DEBUG, "%s-> name:'%s' (0x%.8x)", buffer, node->name, (gint)node->name);
	}
      
      if(node->content != NULL && xmlStrlen(node->content) > 0)
	{
/* 	  spaces = gj_parser_print_tree_spaces(depth); */
	  g_log("XML", G_LOG_LEVEL_DEBUG, "%s - content:'%s' (0x%.8x)", buffer, node->content, (gint)node->content);
	}
    }

  printf(" ");

  result = gj_parser_print_tree(node->xmlChildrenNode,depth+1);

  return result?result:gj_parser_print_tree(node->next,depth);
}

gboolean gj_parser_push_output(xmlNodePtr node)
{
  gchar* output = NULL;
  connection c = NULL;

  if(node == NULL)
    {
      g_warning("%s: failed to continue, node was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if((output = gj_parser_get_str_from_node(node)) == NULL)
    {
      g_warning("%s: out was NULL from node:0x%.8x", __FUNCTION__, (gint)node);
      return FALSE;
    }

  if((c = gj_connection_find_by_host_and_port(gj_me_get_connected_host(),
					      gj_me_get_connected_port())) == NULL)  
  {
      g_warning("%s: could not find connection from host:'%s' and port:%d", 
		__FUNCTION__, gj_me_get_connected_host(), gj_me_get_connected_port());
      return FALSE;
    }
  
  gj_connection_write(c, output);
  g_free(output);

  return TRUE;
}

gboolean gj_parser_push_output_from_text(gchar *text)
{
  connection c = NULL;

  if(text == NULL)
    {
      g_warning("%s: text was NULL", __FUNCTION__);
      return FALSE;
    }

  if((c = gj_connection_find_by_host_and_port(gj_me_get_connected_host(),
					      gj_me_get_connected_port())) == NULL)
    {
      g_warning("%s: could not find connection from host:'%s' and port:%d",  
		__FUNCTION__, gj_me_get_connected_host(), gj_me_get_connected_port());

      return FALSE;
    }

  gj_connection_write(c, text);

  return TRUE;
}

/* utils */
gchar *gj_parser_get_str_from_node(xmlNodePtr node)
{
  xmlDocPtr doc = NULL;
  xmlChar *buf = NULL;

  gint size = 0;

  gchar* output = NULL;

  if(node == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed to continue, node was NULL", __FUNCTION__);
      return NULL;
    }
    
  if((doc = xmlNewDoc((guchar*)"1.0")) == NULL)
    {
      /* failed */
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed to create new xml document, returning...", __FUNCTION__);
      return NULL;
    }
  
  /* set root node */
  xmlDocSetRootElement(doc,node);
  
  /* save to file and get mem */
  xmlDocDumpMemory(doc,&buf,&size);

  if(buf == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed because xmlDocDumpMemory returned NULL", __FUNCTION__);
      return NULL;
    }

  /* 
     skip header with ver 
     and create for fl 
  */
  buf[size] = '\0';
  buf+=21;

  /* !!!!!!! need to free buf... but free, g_free crash it */
  xmlFreeDoc(doc);
  
  output = g_strdup_printf("%s",buf);
  return output;
}


gint gj_parser_doc_count_top_level_nodes(xmlDocPtr doc)
{
  xmlNodePtr root = NULL;

  gint count = 0;

  if(doc == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: doc was NULL", __FUNCTION__);
      return -1;
    }

  /*root = doc->xmlRootNode;*/
  root = doc->xmlRootNode->xmlChildrenNode;

  while(root != NULL)
    {
      root = root->next;
      count++;
    }

  return count;
}

const xmlChar *gj_parser_get_element(xmlNodePtr node)
{
  if(node == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed because node parsed was NULL", __FUNCTION__);
      return NULL;
    }

  return node->name;
}

const xmlChar *gj_parser_get_content(xmlNodePtr node)
{
  if(node == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed because node parsed was NULL", __FUNCTION__);
      return NULL;
    }

  if(node->xmlChildrenNode == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed because node's xmlChildrenNode was NULL", __FUNCTION__);
      return NULL;
    }

  return node->xmlChildrenNode->content;
}

xmlChar *gj_parser_find_by_attr_and_ret(xmlNodePtr node, xmlChar *attrname)
{
  xmlAttrPtr attr = NULL;

  if(node == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed because node was NULL", __FUNCTION__);
      return NULL;
    }

  if(attrname == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed because attr to search for was NULL", __FUNCTION__);
      return NULL;
    }

  attr = gj_parser_find_by_attr(node->properties,attrname);
  
  if(attr == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed to find attribute:'%s'", __FUNCTION__, attrname);
      return NULL;
    }

  if(attr->xmlChildrenNode != NULL)
    return attr->xmlChildrenNode->content;
  else
    return NULL;
}

const xmlChar *gj_parser_find_by_ns_and_ret(xmlNodePtr node)
{
  xmlNodePtr nsnode = NULL;

  if(node == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed because node was NULL", __FUNCTION__);
      return NULL;
    }

  nsnode = gj_parser_find_by_ns(node);
  
  if(nsnode == NULL)
    {
      g_log("XML", G_LOG_LEVEL_DEBUG,"%s: failed to find namespace:", __FUNCTION__);
      return NULL;
    }

  if(nsnode->ns != NULL && nsnode->ns->href != NULL)
      return nsnode->ns->href;
  else
    return NULL;
}

xmlChar * gj_parser_find_by_node_and_ret(xmlNodePtr node, xmlChar *text)
{
  xmlNodePtr found = NULL;
  xmlChar *contents = NULL;

  if((found = gj_parser_find_by_node(node,text)) != NULL)
    {
      if(found->xmlChildrenNode != NULL)
	contents = found->xmlChildrenNode->content;
    }

  return contents;
}

GList *gj_parser_find_by_node_and_ret_contents_by_list(xmlNodePtr node, xmlChar *text)
{
  GList *node_list = NULL;
  GList *contents_list = NULL;

  node_list = gj_parser_find_by_node_and_ret_by_list(node, text);
  
  if(node_list != NULL)
    {
      gint i = 0;

      for(i=0;i < g_list_length(node_list);i++)
	{
	  xmlNodePtr node = NULL;
	  
	  /* checks */
	  if((node = (xmlNodePtr) g_list_nth_data(node_list, i)) == NULL)
	    continue;
	  
	  if(node->xmlChildrenNode == NULL || node->xmlChildrenNode->content == NULL)
	    continue;
	  
	  /* get contents and add to list */
	  contents_list = g_list_append(contents_list, 
					g_strdup((gchar*)node->xmlChildrenNode->content));

	  g_log("XML", G_LOG_LEVEL_DEBUG,"%s: contents[%d]:'%s'", __FUNCTION__, i, node->xmlChildrenNode->content);
	}
    }

  return contents_list;
}

GList *gj_parser_find_by_node_on_parent_and_ret_contents_by_list(xmlNodePtr node, xmlChar *text)
{
  GList *node_list = NULL;
  GList *contents_list = NULL;

  node_list = gj_parser_find_by_node_on_parent_and_ret_by_list(node, text);
  
  if(node_list != NULL)
    {
      gint i = 0;

      for(i=0;i < g_list_length(node_list);i++)
	{
	  xmlNodePtr node = NULL;
	  
	  /* checks */
	  if((node = (xmlNodePtr) g_list_nth_data(node_list, i)) == NULL)
	    continue;
	  
	  if(node->xmlChildrenNode == NULL || node->xmlChildrenNode->content == NULL)
	    continue;
	  
	  /* get contents and add to list */
	  contents_list = g_list_append(contents_list, 
					g_strdup((gchar*)node->xmlChildrenNode->content));

	  g_log("XML", G_LOG_LEVEL_DEBUG, "%s: contents[%d]:'%s'", __FUNCTION__, i, node->xmlChildrenNode->content);
	}
    }

  return contents_list;
}

xmlNodePtr gj_parser_find_by_node(xmlNodePtr node, xmlChar *text)
{
  xmlNodePtr result;
  
  if(node == NULL || text == NULL)                      
    return NULL;
  if(node->name != NULL && xmlStrcasecmp(node->name, text) == 0)   
    return node;

  result = gj_parser_find_by_node(node->xmlChildrenNode,text);

  return result?result:gj_parser_find_by_node(node->next,text);
}

xmlNodePtr gj_parser_find_by_node_on_parent(xmlNodePtr parent, xmlChar *text)
{
  xmlNodePtr result = NULL;
  xmlNodePtr node = NULL;

  if(parent == NULL || text == NULL) 
    return NULL;

  node = parent->xmlChildrenNode;

  while(node != NULL)
    {
      if(node->name != NULL && xmlStrcasecmp(node->name,text) == 0)   
	{
	  result = node;
	  break;
	}

      node = node->next;
    }

  return result;
}

/* xmlNodePtr gj_parser_find_by_node_on_parent(xmlNodePtr node, xmlChar *text) */
/* { */
/*   xmlNodePtr result; */
  
/*   if(node == NULL || text == NULL)                       */
/*     return NULL; */
/*   if(node->name != NULL && xmlStrcasecmp(node->name,text) == 0)    */
/*     return node; */

/*   result = gj_parser_find_by_node_on_parent(node->xmlChildrenNode,text); */

/*   return result; */
/* } */

GList *gj_parser_find_by_node_and_ret_by_list(xmlNodePtr node, xmlChar *text)
{
  GList *list = NULL;

  while(node != NULL)
    {
      if((node = gj_parser_find_by_node(node, text)) == NULL)
	{
	  g_log("XML", G_LOG_LEVEL_DEBUG,"%s: node was NULL", __FUNCTION__);
	  break;
	}

      list = g_list_append(list, node);

      node = node->next;
    }

  return list;
}

GList *gj_parser_find_by_node_on_parent_and_ret_by_list(xmlNodePtr node, xmlChar *text)
{
  GList *list = NULL;

  while(node != NULL)
    {
      if((node = gj_parser_find_by_node_on_parent(node, text)) == NULL)
	{
	  g_log("XML", G_LOG_LEVEL_DEBUG,"%s: node was NULL", __FUNCTION__);
	  break;
	}

      list = g_list_append(list, node);

      node = node->next;
    }

  return list;
}

xmlNodePtr gj_parser_find_by_node_contents(xmlNodePtr node, xmlChar *contents)
{
  xmlNodePtr result;
  
  if(node == NULL || contents == NULL)                      
    return NULL;
  if(node->content != NULL && xmlStrcasecmp(node->content,contents) == 0)   
    return node;

  result = gj_parser_find_by_node_contents(node->xmlChildrenNode,contents);

  return result?result:gj_parser_find_by_node_contents(node->next,contents);
}

xmlAttrPtr gj_parser_find_by_attr(xmlAttrPtr prop, xmlChar *text)
{
  if(text == NULL)
    return NULL;

  while(prop != NULL)
    {
      if(xmlStrcasecmp(prop->name,text) == 0)
	  return prop;

      prop = prop->next;
    }

  return NULL;
}

xmlNodePtr gj_parser_find_by_ns(xmlNodePtr node)
{
  xmlNodePtr result;
  
  if(node == NULL)                      
    return NULL;
  if(node->ns != NULL)
    return node;

  result = gj_parser_find_by_ns(node->xmlChildrenNode);

  return result?result:gj_parser_find_by_ns(node->next);
}

xmlNodePtr gj_parser_find_by_node_with_attr(xmlNodePtr node, xmlChar *nodeName, xmlChar *attrName, xmlChar *attrVal)
{
  xmlNodePtr resultNode = NULL;
  xmlAttrPtr resultAttr = NULL;
  xmlAttrPtr attr = NULL;

  if(attrVal == NULL)
    return NULL;

  while((resultNode = gj_parser_find_by_node(node,nodeName)) != NULL)
    {
      attr = resultNode->properties;

      while((resultAttr = gj_parser_find_by_attr(attr,attrName)) != NULL)
	{
	  if(resultAttr->xmlChildrenNode != NULL && resultAttr->xmlChildrenNode->content != NULL)
	    {
	      if(xmlStrcasecmp(resultAttr->xmlChildrenNode->content,attrVal) == 0)
		return resultAttr->parent;
	    }

	  attr = resultAttr->next;
	}

      node = resultNode->next;
    }

  return NULL;
}

/*
 * parser callbacks
 */

void gj_parser_doc_start_cb(void *user_data)
{
  g_log("XML", G_LOG_LEVEL_DEBUG,"%s: user_data:0x%.8x", __FUNCTION__, (gint)user_data);
}

void gj_parser_doc_end_cb(void *user_data)
{
  g_log("XML", G_LOG_LEVEL_DEBUG,"%s: user_data:0x%.8x", __FUNCTION__, (gint)user_data);
}

void gj_parser_doc_element_start_cb(void *user_data, const xmlChar *name, const xmlChar **attrs)
{
  g_log("XML", G_LOG_LEVEL_DEBUG,"%s: user_data:0x%.8x, name:'%s'", __FUNCTION__, (gint)user_data, name);
}

void gj_parser_doc_element_end_sb(void *user_data, const xmlChar *name)
{
  g_log("XML", G_LOG_LEVEL_DEBUG,"%s: user_data:0x%.8x, name:'%s'", __FUNCTION__, (gint)user_data, name);
}

xmlEntityPtr gj_parser_doc_entity_cb(void *user_data, const xmlChar *name) 
{
  g_log("XML", G_LOG_LEVEL_DEBUG, "%s: user_data:0x%.8x, name:'%s'", __FUNCTION__, (gint)user_data, name);
  return xmlGetPredefinedEntity(name);
}

void gj_parser_doc_warning_cb(void *user_data, const char *msg, ...) 
{
    va_list args;

    va_start(args, msg);
    g_logv("XML", G_LOG_LEVEL_WARNING, msg, args);
    va_end(args);
}

void gj_parser_doc_error_cb(void *user_data, const char *msg, ...) 
{
    va_list args;

    va_start(args, msg);
    g_logv("XML", G_LOG_LEVEL_CRITICAL, msg, args);
    va_end(args);
}

void gj_parser_doc_fatal_error_cb(void *user_data, const char *msg, ...) 
{
    va_list args;

    va_start(args, msg);
    g_logv("XML", G_LOG_LEVEL_ERROR, msg, args);
    va_end(args);
}

#ifdef __cplusplus
}
#endif
