/* 
 * $RCSfile: gj_roster.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/28 14:46:35 $
 * 
 * $Revision: 1.21 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>

#include <time.h> 

#include <glib.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include "gj_roster.h"
#include "gj_parser.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_event.h"
#include "gj_history.h"
#include "gj_translate.h"
#include "gj_me.h"
#include "gj_support.h"

#include "gj_gtk_roster.h" /* this is for the roster_model_add_contact call... 
			      and i dont think this should be done here...*/

struct t_rosterItem
{
  gint id; /* unique id */

  gchar *jid;
  gchar *name;
  gchar *resource;
  gchar *agent;

  GList *groups;

  /* old - was used before type */
  gboolean is_agent;            

  /* is this a permanant member - i.e. not just someone that has sent me a message */
  gboolean is_permanent;        

  /* is this just my jid/some resource ? */
  gboolean is_another_resource;

  /* is this roster item about to be deleted and shouldnt be added to the roster */
  gboolean is_set_for_removal;  

  rosterItemType type;
  rosterItemSubscription subscription;

  /* presence information */
  presence pres;
};


GList *roster_list;
rosterCallback roster_cb;

/* roster item */
void gj_roster_item_destroy(rosterItem ri);

gboolean gj_roster_item_set_id(rosterItem ri, gint id);
gboolean gj_roster_item_set_is_permanent(rosterItem ri, gboolean IsPermanent);
gboolean gj_roster_item_set_is_another_resource(rosterItem ri, gboolean IsAnotherResource);
gboolean gj_roster_item_set_is_set_for_removal(rosterItem ri, gboolean IsSetForRemoval);

/* roster list */
gboolean gj_roster_list_add(rosterItem ri);
gboolean gj_roster_list_del(rosterItem ri);

/* call backs */
void rosterResponse(gint requestID);

gboolean gj_roster_set_callback(rosterCallback cb);
rosterCallback gj_roster_get_callback();


gboolean gj_roster_init()
{
  if(roster_list != NULL)
    gj_roster_term();

  /* free roster response proc */
  roster_cb = NULL;

  /* create mem  - nothing to do, created on g_list_append */
  return TRUE;
}

gboolean gj_roster_term()
{
  gint index = 0;

  /* list */
  if(roster_list == NULL)
    return TRUE;

  /* free all items on the list */
  for(index=0;index<g_list_length(roster_list);index++)
    {
      rosterItem ri = NULL;
      
      if((ri = (rosterItem) g_list_nth_data(roster_list,index)) != NULL)
	gj_roster_item_free(ri);
    }

  g_list_free(roster_list);
  roster_list = NULL;

  /* free roster response proc */
  roster_cb = NULL;

  return TRUE;
}

gboolean gj_roster_list_add(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  roster_list = g_list_append(roster_list,ri);

  return TRUE;
}

gboolean gj_roster_list_del(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  roster_list = g_list_remove(roster_list,ri);
  return TRUE;
}

GList *gj_roster_get_list()
{
  return roster_list;
}

GList *gj_roster_get_all_groups()
{
  GList *group_master_list = NULL;
  gint i,j,k;
  gboolean found = FALSE;

  if(roster_list == NULL)
    {
      g_warning("%s: roster list was NULL", __FUNCTION__);
      return NULL;
    }

  /* look through roster items */
  for(i=0;i<g_list_length(roster_list);i++)
    {
      rosterItem ri  = g_list_nth_data(roster_list, i);

      if(ri == NULL)
	continue;

      if(ri->groups == NULL)
	continue;
       
      /* look through groups that the user belongs to */
      for(k=0;k<g_list_length(ri->groups);k++)
	{
	  gchar *group = (gchar *) g_list_nth_data(ri->groups, k);

	  if(group == NULL)
	    continue;

	  /* search master_list for same group name */
	  for(j=0,found=FALSE;j<g_list_length(group_master_list);j++)
	    {
	      gchar *name = (gchar *) g_list_nth_data(group_master_list, j);
	      
	      /* checks */
	      if(name == NULL)
		continue;
	  
	      if(g_strncasecmp(group, name, g_utf8_strlen(group,-1)) == 0)
		{
		  found = TRUE;
		  break;
		}	  
	    }

	  /* if after searching master_list so far
	     it doesnt exist, we can then add this group */
	  if(found == FALSE)
	    {
	      group_master_list = g_list_append(group_master_list,g_strdup(group)); 
	      g_message("%s: adding to list, group:'%s'", __FUNCTION__, group);
	    }
	}
    }  

  return group_master_list;
}

GList *gj_roster_get_all_items_by_group(gchar *group)
{
  GList *list = NULL;
  gint i,j,k;
  gboolean found = FALSE;

  /* note: if NULL is specified as a group, 
           it will look for all items without a group */

  if(roster_list == NULL)
    {
      g_warning("%s: roster list was NULL", __FUNCTION__);
      return NULL;
    }

  /* look through roster items */
  for(i=0;i<g_list_length(roster_list);i++)
    {
      rosterItem ri = (rosterItem) g_list_nth_data(roster_list, i);

      if(ri == NULL)
	continue;

      if(ri->groups == NULL && group == NULL)
	{
	  list = g_list_append(list, g_strdup(gj_roster_item_get_jid(ri)));
	  continue;
	}
	
      /* look through groups that the user belongs to */
      for(k=0; k<g_list_length(ri->groups); k++)
	{
	  gchar *name = (gchar*) g_list_nth_data(ri->groups, k);

	  if(name == NULL || strcmp(name, group) != 0)
	    continue;

	  /* search our new list for same group name */
	  for(j=0,found=FALSE; j<g_list_length(list); j++)
	    {
	      gchar *item = (gchar*) g_list_nth_data(list, j);
	      
	      /* checks */
	      if(item == NULL)
		continue;
	  
	      if(strcmp(gj_roster_item_get_jid(ri), item) == 0)
		{
		  found = TRUE;
		  break;
		}	  
	    }

	  /* if after searching master_list so far
	     it doesnt exist, we can then add this group */
	  if(found == FALSE)
	    {
	      list = g_list_append(list, g_strdup(gj_roster_item_get_jid(ri))); 
	      g_message("%s: adding to list, jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri));
	    }
	}
    }  

  return list;
}

GList *gj_roster_get_all_items_by_agent(gchar *agent)
{
  GList *list = NULL;
  gint i = 0;

  if(roster_list == NULL)
    {
      g_warning("%s: roster list was NULL", __FUNCTION__);
      return NULL;
    }

  if(agent == NULL || g_utf8_strlen(agent, -1) < 1)
    {
      g_warning("%s: agent was NULL or length was < 1", __FUNCTION__);
      return NULL;
    }

  /* look through roster items */
  for(i=0;i<g_list_length(roster_list);i++)
    {
      rosterItem ri = (rosterItem) g_list_nth_data(roster_list, i);

      if(ri == NULL)
	continue;

      if(gj_roster_item_get_agent(ri) == NULL)
	continue;
	
      if(strcmp(gj_roster_item_get_agent(ri), agent) == 0)
	{
	  list = g_list_append(list, g_strdup(gj_roster_item_get_jid(ri))); 
	  g_message("%s: adding to list, jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri));
	}
    }  

  return list;
}


GList *gj_roster_get_all_agents_by_jid()
{
  GList *list = NULL;
  gint i = 0;

  if(roster_list == NULL)
    {
      g_warning("%s: roster list was NULL", __FUNCTION__);
      return NULL;
    }

  /* look through roster items */
  for(i=0;i<g_list_length(roster_list);i++)
    {
      rosterItem ri = (rosterItem) g_list_nth_data(roster_list, i);

      if(ri == NULL)
	continue;

      if(gj_roster_item_get_is_agent(ri) == FALSE)
	continue;
       
      list = g_list_append(list, g_strdup(gj_roster_item_get_jid(ri))); 
      g_message("%s: adding to list, jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri));
    }  

  return list;
}

GList *gj_roster_get_all_agents_by_jid_without_resource()
{
  GList *list = NULL;
  gint i = 0;

  if(roster_list == NULL)
    {
      g_warning("%s: roster list was NULL", __FUNCTION__);
      return NULL;
    }

  /* look through roster items */
  for(i=0;i<g_list_length(roster_list);i++)
    {
      rosterItem ri  = g_list_nth_data(roster_list, i);
      gchar *jid = NULL;
      gchar **vjid = NULL;

      if(ri == NULL)
	continue;

      if(gj_roster_item_get_is_agent(ri) == FALSE)
	continue;
       
      vjid = g_strsplit(gj_roster_item_get_jid(ri), "/", -1);
      
      if(vjid == NULL)
	  continue;

      if(vjid[0] != NULL)
	  jid = g_strdup(vjid[0]);
      
      g_strfreev(vjid);

      if(jid == NULL)
	  continue;

      list = g_list_append(list, g_strdup(jid)); 
      g_message("%s: adding to list, jid:'%s'", __FUNCTION__, jid);
    }  

  return list;
}

GList *gj_roster_get_all_agents_by_name()
{
  GList *list = NULL;
  gint i = 0;

  if(roster_list == NULL)
    {
      g_warning("%s: roster list was NULL", __FUNCTION__);
      return NULL;
    }

  /* look through roster items */
  for(i=0;i<g_list_length(roster_list);i++)
    {
      rosterItem ri = (rosterItem) g_list_nth_data(roster_list, i);

      if(ri == NULL)
	continue;

      if(gj_roster_item_get_is_agent(ri) == FALSE)
	continue;
       
      list = g_list_append(list, g_strdup(gj_roster_item_get_name(ri))); 
      g_message("%s: adding to list, name:'%s'", __FUNCTION__, gj_roster_item_get_name(ri));
    }  

  return list;
}


gboolean gj_roster_is_agent_registered(gchar *agent)
{
  GList *list = NULL;
  gint i = 0;
  gboolean found = FALSE;

  if(agent == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if((list = gj_roster_get_all_agents_by_jid_without_resource()) == NULL)
    return FALSE;

  /* look through roster items */
  for(i=0;found == FALSE && i<g_list_length(list);i++)
    {
      gchar *jid = (gchar*) g_list_nth_data(list, i);

      if(jid == NULL)
	continue;

      if(g_strcasecmp(jid, agent) == 0)
	found = TRUE;
    }  
  
  return found;
}

/* 
 * item functions
 */
rosterItem gj_roster_item_new(gchar *jid)
{
  rosterItem ri;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }
  
  if((ri = g_new(struct t_rosterItem,1)))
    {
      ri->id = gj_support_get_unique_id();

      ri->jid = g_strdup(jid);
      ri->name = NULL;
      ri->resource = NULL;
      ri->agent = NULL;

      ri->groups = NULL;

      ri->is_agent = FALSE;
      ri->is_permanent = FALSE;
      ri->is_another_resource = FALSE;
      ri->is_set_for_removal = FALSE;

      ri->subscription = rosterItemSubscriptionNone;
      ri->type = rosterItemTypeContact;

      ri->pres = gj_presence_new(gj_support_get_unique_id(), presenceTypeUnAvailable, presenceShowNormal);

      gj_roster_list_add(ri);
    }
  
  return ri;  
}

rosterItem gj_roster_item_new_unknown_by_message(message msg)
{
  rosterItem ri;

  gchar *from = NULL;
  gchar *jid_no_resource = NULL;
  gchar *resource = NULL;
  gchar **vjid = NULL;
  gchar **vagent = NULL;

  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }
  
  if((from = gj_message_get_from_jid(msg)) == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  vjid = g_strsplit(from,"/",2);
  
  if(vjid != NULL)
    {
      jid_no_resource = g_strdup(from);
      
      if(vjid[1] != NULL)
	resource = g_strdup(vjid[1]);
      
      g_strfreev(vjid);
    }
  
  /* create new roster item */
  if((ri = gj_roster_item_new(jid_no_resource)) == NULL)
    {
      g_warning("%s: could not create new roster item", __FUNCTION__);

      if(jid_no_resource != NULL) g_free(jid_no_resource);
      if(resource != NULL) g_free(resource);

      return NULL;
    }
      
  vagent = g_strsplit(jid_no_resource,"@",2);
  
  if(vagent != NULL)
  {
    if(vagent[1] != NULL)
      {
	/* not a agent, so get the agent part (i.e @icq.jabber.org) */
	ri->agent = g_strdup(vagent[1]);
	ri->is_agent = FALSE;
	ri->type = rosterItemTypeContact;
      }
    else
      {
	/* no agent part (i.e. @msn.jabber.org) there for MUST be a agent */
	ri->is_agent = TRUE;
	ri->type = rosterItemTypeAgent;
      }

    g_strfreev(vagent);
  }
  
  /* name */
  if(jid_no_resource != NULL)
    ri->name = g_strdup(jid_no_resource);

  /* resource */
  if(resource != NULL)
    ri->resource = resource;

  /* is it another resource with the same account ? */
  if(jid_no_resource != NULL && strcmp(jid_no_resource, gj_me_get_jid_without_resource()) == 0)
    ri->is_another_resource = TRUE;

  /* clean up */
  if(jid_no_resource != NULL) g_free(jid_no_resource);

  return ri;  
}

rosterItem gj_roster_item_new_unknown_by_jid(gchar *from)
{
  rosterItem ri;

  gchar *jid = NULL;
  gchar *jid_no_resource = NULL;
  gchar *resource = NULL;
  gchar **vjid = NULL;
  gchar **vagent = NULL;

  if(from == NULL || g_utf8_strlen(from, -1) < 1)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  vjid = g_strsplit(from,"/",2);
  
  if(vjid != NULL)
    {
      jid = g_strdup(from);
      
      if(vjid[0] != NULL)
	jid_no_resource = g_strdup(vjid[0]);

      if(vjid[1] != NULL)
	resource = g_strdup(vjid[1]);
      
      g_strfreev(vjid);
    }
  
  /* create new roster item */
  if((ri = gj_roster_item_new(jid)) == NULL)
    {
      g_warning("%s: could not create new roster item", __FUNCTION__);

      if(jid != NULL) g_free(jid);
      if(jid_no_resource != NULL) g_free(jid_no_resource);
      if(resource != NULL) g_free(resource);

      return NULL;
    }
      
  vagent = g_strsplit(jid,"@",2);
  
  if(vagent != NULL)
  {
    if(vagent[1] != NULL)
      {
	/* not a agent, so get the agent part (i.e @icq.jabber.org) */
	ri->agent = g_strdup(vagent[1]);
	ri->is_agent = FALSE;
	ri->type = rosterItemTypeContact;
      }
    else
      {
	/* no agent part (i.e. @msn.jabber.org) there for MUST be a agent */
	ri->is_agent = TRUE;
	ri->type = rosterItemTypeAgent;
      }

    g_strfreev(vagent);
  }
  
  /* name */
  if(jid != NULL)
    ri->name = g_strdup(jid);

  /* resource */
  if(resource != NULL)
    ri->resource = resource;

  /* is it another resource with the same account ? */
  if(jid_no_resource != NULL && strcmp(jid_no_resource, gj_me_get_jid_without_resource()) == 0)
    {
      ri->is_another_resource = TRUE;
      ri->subscription = rosterItemSubscriptionBoth;
    }

  /* clean up */
  if(jid != NULL) g_free(jid);
  if(jid_no_resource != NULL) g_free(jid_no_resource);

  return ri;  
}

void gj_roster_item_free(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return;
    }

  /* free mem */
  gj_roster_list_del(ri);  

  /* destroy members */
  gj_roster_item_destroy(ri);

  /* free structure */
  g_free(ri);
}

void gj_roster_item_destroy(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return;
    }
  
  ri->id = 0;

  if(ri->jid != NULL)
    {
      g_free(ri->jid);
      ri->jid = NULL;
    }

  if(ri->name != NULL)
    {
      g_free(ri->name);
      ri->name = NULL;
    }

  if(ri->resource != NULL)
    {
      g_free(ri->resource);
      ri->resource = NULL;
    }

  if(ri->agent != NULL)
    {
      g_free(ri->agent);
      ri->agent = NULL;
    }

  if(ri->groups != NULL)
    gj_roster_item_groups_free(ri);

  ri->is_agent = FALSE;
  ri->is_permanent = FALSE;
  ri->is_another_resource = FALSE;
  ri->is_set_for_removal = FALSE;
  
  ri->subscription = 0;

  /* presence */
  if(ri->pres != NULL)
    {
      gj_presence_free(ri->pres);
      ri->pres = NULL;
    }
}

/* 
 * item groups 
 */
gboolean gj_roster_item_groups_free(rosterItem ri)
{
  gint i;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ri->groups == NULL)
    return TRUE;

  /* reverse delete through the list */
  for(i=g_list_length(ri->groups);i>0;i--)
    {
      gchar *group = (gchar *) g_list_nth_data(ri->groups, i);
      if(group == NULL) continue;
      
      /* free off list and str */
      g_list_remove(ri->groups, group);
      g_free(group);
      group = NULL;
    }

  ri->groups = NULL;

  return TRUE;
}

gboolean gj_roster_item_groups_add(rosterItem ri, gchar *group)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }  

  if(group == NULL)
    {
      g_warning("%s: group was NULL", __FUNCTION__);
      return FALSE;
    }
  
  /* make our copy of it */
  ri->groups = g_list_append(ri->groups, g_strdup(group)); 

  return TRUE;
}

gboolean gj_roster_item_groups_del(rosterItem ri, gchar *group)
{
  GList *item = NULL;
  gchar *item_str = NULL;
  gint index = 0;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }  

  if(group == NULL)
    {
      g_warning("%s: group was NULL", __FUNCTION__);
      return FALSE;
    }
  
  /* find the item */
  item = g_list_find_custom(ri->groups, group, (GCompareFunc)strcmp);
  if(item == NULL) return TRUE;

  index = g_list_position(ri->groups, item);
  if(index == -1) return TRUE;

  item_str = g_list_nth_data(ri->groups, index);
  if(item_str == NULL) return FALSE;

  /* remove it */
  ri->groups = g_list_remove(ri->groups, item_str); 

  return TRUE;
}


gint gj_roster_item_groups_count(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return -1;
    }   

  if(ri->groups == NULL)
    return -1;
  
  return g_list_length(ri->groups);
}

/*
 * roster handling
 */
GList *gj_roster_parse(xmlNodePtr parent)
{
  gint count = 0;
  xmlNodePtr node = NULL;

  if(parent == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return NULL;
    }

  /* create list */
  gj_roster_init();

  /* get item node */
  node = gj_parser_find_by_node(parent, (guchar*)"item");

  /* get items in roster */
  for(;node != NULL;node = node->next)
    {
      xmlChar *tmp = NULL;

      count++;
      
      /* get data */
      if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"jid")) != NULL)
	{
	  gchar **vjid = NULL;
	  gchar **vagent = NULL;
	  
	  rosterItem ri = NULL;
	  history hist = NULL;

	  vjid = g_strsplit((gchar*)tmp,"/",2);

	  if(vjid != NULL)
	    {
	      /* create new roster item */
	      ri = gj_roster_item_new((gchar*)tmp);

	      if(ri == NULL)
		{
		  g_warning("%s: could not create new roster item for jid:'%s'", __FUNCTION__, tmp);
		  g_strfreev(vjid);
		  continue;
		}

	      /* resource */
	      if(vjid[1] != NULL)
		gj_roster_item_set_resource(ri, vjid[1]);
	      
	      /* is it another resource with the same account ? */
	      if(vjid[0] != NULL && strcmp(vjid[0], gj_me_get_jid_without_resource()) == 0)
		gj_roster_item_set_is_another_resource(ri, TRUE);

	      /* set agent information */
	      if((vagent = g_strsplit(vjid[0],"@",2)) != NULL)
		{
		  if(vagent[1] != NULL)
		    {
		      /* not a agent, so get the agent part (i.e @icq.jabber.org) */
		      gj_roster_item_set_agent(ri, vagent[1]);
		      gj_roster_item_set_is_agent(ri, FALSE);
		      gj_roster_item_set_type(ri, rosterItemTypeContact);
		    }
		  else 
		    {
		      /* no agent part (i.e. @msn.jabber.org) there for MUST be a agent */
		      gj_roster_item_set_is_agent(ri, TRUE);
		      gj_roster_item_set_type(ri, rosterItemTypeAgent);
		    }
		  
		  g_strfreev(vagent);
		}

	      g_strfreev(vjid);
	    }
	  
	  /* name */
	  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"name")) != NULL)
	    gj_roster_item_set_name(ri, (gchar*)tmp);
	  
	  /* groups */ 
	  gj_roster_item_set_groups(ri, gj_parser_find_by_node_on_parent_and_ret_contents_by_list(node, (guchar*)"group"));

	  /* subscription type */
	  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"subscription")) != NULL)
	    {
	      if(xmlStrcmp(tmp, (guchar*)"none") == 0)
		gj_roster_item_set_subscription(ri, rosterItemSubscriptionNone);
	      else if(xmlStrcmp(tmp, (guchar*)"to") == 0)
		gj_roster_item_set_subscription(ri, rosterItemSubscriptionTo);
	      else if(xmlStrcmp(tmp, (guchar*)"from") == 0)
		gj_roster_item_set_subscription(ri, rosterItemSubscriptionFrom);
	      else if(xmlStrcmp(tmp, (guchar*)"both") == 0)
		gj_roster_item_set_subscription(ri, rosterItemSubscriptionBoth);
	      else
		g_warning("%s: unexpected roster item type:'%s'", __FUNCTION__, tmp);
	    }
	  
	  gj_roster_item_set_is_permanent(ri, TRUE);

	  /* create history */
	  if((hist = gj_history_find_by_ri_id(gj_roster_item_get_id(ri))) == NULL)
	    hist = gj_history_new(ri);

	  gj_history_start(hist);
	}
    }

  g_message("%s: items:%d", __FUNCTION__, count);

  return roster_list;
}

gboolean gj_roster_parse_set_request(infoquery iq)
{
  xmlNodePtr node = NULL;
  xmlNodePtr parent = NULL;

  gint count = 0;

  g_message("%s: updating roster item(s)...", __FUNCTION__);

  if((parent = gj_iq_get_xml(iq)) == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return FALSE;
    }

  /* get item node */
  node = gj_parser_find_by_node(parent, (guchar*)"item");

  /* get items in roster */
  for(;node != NULL; node = node->next)
    {
      gchar *jid = NULL;
      gchar *resource = NULL;
      gchar *agent = NULL;
      gchar *ask = NULL;
      gchar *subscription = NULL;
      gchar *name = NULL;

      gchar **vjid = NULL;
      gchar **vagent = NULL;

      gboolean is_new = TRUE;
      
      xmlChar *tmp = NULL;

      rosterItem ri = NULL;

      count++;
      
      /* get data */
      if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"jid")) == NULL)
	{
	  g_warning("%s: could not get jid from xml node", __FUNCTION__);
	  continue;
	}

      /* get jid and resource */
      vjid = g_strsplit((gchar*)tmp,"/",2);
      
      if(vjid != NULL)
	{
	  jid = g_strdup((gchar*)tmp);
	  
	  if(vjid[1] != NULL)
	    resource = g_strdup(vjid[1]);

	  /* set agent information */
	  if((vagent = g_strsplit(vjid[0],"@",2)) != NULL)
	    {
	      if(vagent[1] != NULL)
		agent = g_strdup(vagent[1]);
	      
	      g_strfreev(vagent);
	    }
	  
	  g_strfreev(vjid);
	}
      
      /* get name */
      if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"name")) != NULL)
	name = g_strdup((gchar*)tmp);  

      /* get subscription */
      if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"subscription")) != NULL)
	subscription = g_strdup((gchar*)tmp);

      /* get ask */
      if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"ask")) != NULL)
	ask = g_strdup((gchar*)tmp);

      /* look for roster item */
      if((ri = gj_roster_item_find_by_jid(jid)) == NULL)
	{
	  g_message("%s: roster item not found by jid:'%s', attempting without resource...", __FUNCTION__, jid);

	  if((ri = gj_roster_item_find_by_jid_without_resource(jid)) == NULL)
	    g_warning("%s: roster item could not be found at all by jid:'%s'", __FUNCTION__, jid);
	}
      
      /* NOTE: we remove the contact here and add it later on when the
	 details have changed.  this way we can change the groups
	 and not have to worry about finding the contact and checking 
	 the groups and details...
      */
      
      if(subscription != NULL && g_strcasecmp(subscription, "remove") == 0)
	{
	  /*
	   * REMOVE roster item
	   */
	  if(ri == NULL)
	    {
	      g_warning("%s: could not remove jid:'%s', no roster item found.", __FUNCTION__, jid);
	      
	      /* clean up */
	      if(jid != NULL) g_free(jid);
	      if(resource != NULL) g_free(resource);
	      if(agent != NULL) g_free(agent);
	      
	      continue;
	    }
	  else
	    {
	      /* let people know it is about to be deleted */
	      gj_roster_item_set_is_set_for_removal(ri, TRUE);

	      /* remove it from roster */
	      gj_gtk_roster_model_remove_contact(ri);

	      /* free it */
	      gj_roster_item_free(ri);

	      g_message("%s: removed roster item associated with jid:'%s'", __FUNCTION__, jid);
	    }

	  continue;
	}
      else if(ask != NULL)
	{
	  /*
	   * ASK for subscription or to unsubscribe
	   */
	  if(g_strcasecmp(ask, "subscribe") == 0)
	    g_message("%s: jid:'%s' is requesting subscription, name:'%s'", __FUNCTION__, jid, name);
	  else if(g_strcasecmp(ask, "unsubscribe") == 0)
	    g_message("%s: jid:'%s' is cancelling subscription, name:'%s'", __FUNCTION__, jid, name);
	  else
	    g_warning("%s: ask value:'%s' not dealt with from jid:'%s' with name:'%s'", __FUNCTION__, tmp, jid, name);
	  
	  /* clean up */
	  if(jid != NULL) g_free(jid);
	  if(resource != NULL) g_free(resource);
	  if(agent != NULL) g_free(agent);
	  if(name != NULL) g_free(name);
	  
	  continue;
	}

      gj_gtk_roster_model_remove_contact(ri);

      /* find roster item */
      if(ri == NULL)
	{
	  g_message("%s: no roster item found by jid:'%s', creating...", __FUNCTION__, jid);
	  
	  /* FIXME: why is this temporary? if there is a roster set request, it would be permanat wouldnt it?? */
	  /* adding temporary (to this session) user and group */
	  if((ri = gj_roster_item_new(jid)) == NULL)
	    {
	      g_warning("%s: couldn't add temporary roster item", __FUNCTION__);
	      
	      if(jid != NULL) g_free(jid);
	      if(resource != NULL) g_free(resource);
	      if(agent != NULL) g_free(agent);
	      
	      node = node->next;
	      continue;
	    }

	  is_new = TRUE;
	}
      else
	g_message("%s: found roster item for jid:'%s', updating...", __FUNCTION__, jid);
      
      /* set agent information */
      if(agent != NULL)
	{
	  /* not a agent, so get the agent part (i.e @icq.jabber.org) */
	  gj_roster_item_set_agent(ri, agent);
	  gj_roster_item_set_is_agent(ri, FALSE);
	  gj_roster_item_set_type(ri, rosterItemTypeContact);
	}
      else
	{
	  /* no agent part (i.e. @msn.jabber.org) there for MUST be a agent */
	  gj_roster_item_set_is_agent(ri, TRUE);
	  gj_roster_item_set_type(ri, rosterItemTypeAgent);
	}
      
      if(resource != NULL) gj_roster_item_set_resource(ri, resource);
      if(name != NULL) gj_roster_item_set_name(ri, name);

      /* groups */ 
      gj_roster_item_set_groups(ri,gj_parser_find_by_node_on_parent_and_ret_contents_by_list(node, (guchar*)"group"));
      
      /* subscription type */
      if(subscription != NULL)
	{
	  if(g_strcasecmp(subscription, "none") == 0)
	    gj_roster_item_set_subscription(ri, rosterItemSubscriptionNone);
	  else if(g_strcasecmp(subscription, "to") == 0)
	    gj_roster_item_set_subscription(ri, rosterItemSubscriptionTo);
	  else if(g_strcasecmp(subscription, "from") == 0)
	    gj_roster_item_set_subscription(ri, rosterItemSubscriptionFrom);
	  else if(g_strcasecmp(subscription, "both") == 0)
	    gj_roster_item_set_subscription(ri, rosterItemSubscriptionBoth);
	  else
	    g_warning("%s: unexpected roster item type:'%s'", __FUNCTION__, subscription);
	}
      
      /* NOTE: at this point, we test to see if a contact has 
	 been added or simply updated. */
      if(is_new == TRUE)
	{
	  /* set it to be permanant. */
	  gj_roster_item_set_is_permanent(ri, TRUE);

	  /* is it another resource with the same account ? */
	  if(gj_roster_item_get_jid_without_resource(ri) != NULL && 
	     strcmp(gj_roster_item_get_jid_without_resource(ri), gj_me_get_jid_without_resource()) == 0)
	    gj_roster_item_set_is_another_resource(ri, TRUE);

	  /* add visual contact */
	  g_message("%s: create new contact visually for jid:'%s'...", __FUNCTION__, jid);
	  gj_gtk_roster_model_add_contact(ri);
	}
      
      /* clean up */
      if(jid != NULL) g_free(jid);
      if(resource != NULL) g_free(resource);
      if(agent != NULL) g_free(agent);
    }
  
  g_message("%s: updated %d items", __FUNCTION__, count);

  return TRUE; 
}

gboolean gj_roster_set_callback(rosterCallback cb)
{
  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  roster_cb = cb;
  return TRUE;
}

rosterCallback gj_roster_get_callback()
{
  return roster_cb;
}

/*
 * rosterItem functions 
 *
 */
rosterItem gj_roster_item_find_by_id(gint id)
{
  gint index = 0;

  if(id < 1)
    {
      g_warning("%s: id:%d was < 1", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(roster_list);index++)
    {
      rosterItem ri  = g_list_nth_data(roster_list, index);

      if(ri == NULL) continue;
      if(gj_roster_item_get_name(ri) == NULL) continue;
      if(id == gj_roster_item_get_id(ri)) return ri;
    }

  return NULL;
}

rosterItem gj_roster_item_find_by_name(gchar *name)
{
  gint index = 0;

  if(name == NULL)
    {
      g_warning("%s: name was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(roster_list);index++)
    {
      rosterItem ri = g_list_nth_data(roster_list, index);

      if(ri == NULL) continue;
      if(gj_roster_item_get_name(ri) == NULL) continue;
      if(strcmp(name, gj_roster_item_get_name(ri)) == 0) return ri;
    }

  return NULL;
}

rosterItem gj_roster_item_find_by_jid(gchar *jid)
{
  rosterItem ri = NULL;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  if((ri = gj_roster_item_find_by_jid_with_resource(jid)) != NULL)
    return ri;

  if((ri = gj_roster_item_find_by_jid_ignore_resource(jid)) != NULL)
    return ri;
  
  return NULL;
}

rosterItem gj_roster_item_find_by_jid_with_resource(gchar *jid)
{
  gint index = 0;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(roster_list);index++)
    {
      rosterItem ri  = g_list_nth_data(roster_list, index);

      if(ri == NULL) continue;
      if(gj_roster_item_get_jid(ri) == NULL) continue;
      if(strcmp(jid, gj_roster_item_get_jid(ri)) == 0) return ri;
    }

  return NULL;
}

rosterItem gj_roster_item_find_by_jid_without_resource(gchar *jid)
{
  gint index = 0;
  rosterItem found_ri = NULL;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(roster_list);index++)
    {
      gchar *search_jid = NULL;
      gchar **search_vResource = NULL;
      gchar *known_jid = NULL;
      gchar **known_vResource = NULL;

      rosterItem ri  = g_list_nth_data(roster_list, index);

      if(ri == NULL) continue;
      if(gj_roster_item_get_jid(ri) == NULL) continue;

      search_jid = g_strdup(gj_roster_item_get_jid(ri));
      search_vResource = g_strsplit(search_jid,"/",2);

      known_jid = g_strdup(jid);
      known_vResource = g_strsplit(known_jid,"/",2);

      if(search_vResource != NULL && known_vResource != NULL)
	{
	  if(search_vResource[0] != NULL && known_vResource[0] != NULL && 
	     strcmp(known_vResource[0], search_vResource[0]) == 0)
	    found_ri = ri;
	  
	  g_strfreev(search_vResource);
	  g_strfreev(known_vResource);
	}

      if(search_jid != NULL) g_free(search_jid);
      if(known_jid != NULL) g_free(known_jid);

      if(found_ri != NULL)
	return found_ri;
    }

  return NULL;
}

rosterItem gj_roster_item_find_by_jid_ignore_resource(gchar *jid)
{
    gchar **v = NULL;
    rosterItem ri = NULL;

    if(jid == NULL)
    {
	g_warning("%s: jid was NULL", __FUNCTION__);
	return NULL;
    }

    v = g_strsplit(jid,"/",2);
    if(v == NULL || v[0] == NULL)
    {
	g_warning("%s: could not split string:'%s' with '/' as delimiter", __FUNCTION__, jid); 
	return NULL;
    }

    ri = gj_roster_item_find_by_jid_without_resource(v[0]);

    /* clean up */
    g_strfreev(v);

    return ri;
}

/*
 * member functions 
 */

/* sets */
gboolean gj_roster_item_set_id(rosterItem ri, gint id)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(id < 1)
    {
      g_warning("%s: id:%d was < 1", __FUNCTION__, id);
      return FALSE;
    }

  ri->id = id;
  return TRUE;
}

gboolean gj_roster_item_set_jid(rosterItem ri, gchar *jid)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ri->jid != NULL)
    {
      g_free(ri->jid);
      ri->jid = NULL;
    }
  
  ri->jid = g_strdup(jid);
  return TRUE;
}

gboolean gj_roster_item_set_name(rosterItem ri, gchar *name)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(name == NULL)
    {
      g_warning("%s: name was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ri->name != NULL)
    {
      g_free(ri->name);
      ri->name = NULL;
    }
  
  ri->name = g_strdup(name);
  return TRUE;
}

gboolean gj_roster_item_set_resource(rosterItem ri, gchar *resource)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(resource == NULL)
    {
      g_warning("%s: resource was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ri->resource != NULL)
    {
      g_free(ri->resource);
      ri->resource = NULL;
    }
  
  ri->resource = g_strdup(resource);
  return TRUE;
}

gboolean gj_roster_item_set_agent(rosterItem ri, gchar *agent)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(agent == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ri->agent != NULL)
    {
      g_free(ri->agent);
      ri->agent = NULL;
    }
  
  ri->agent = g_strdup(agent);
  return TRUE;
}

gboolean gj_roster_item_set_groups(rosterItem ri, GList *groups)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(groups == NULL)
    {
      g_warning("%s: groups list was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ri->groups != NULL)
    {
      gj_roster_item_groups_free(ri);
      ri->groups = NULL;
    }
  
  ri->groups = groups;
  return TRUE;
}

gboolean gj_roster_item_set_is_agent(rosterItem ri, gboolean IsAgent)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }
  
  ri->is_agent = IsAgent;
  return TRUE;
}

gboolean gj_roster_item_set_is_permanent(rosterItem ri, gboolean IsPermanent)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }
  
  ri->is_permanent = IsPermanent;
  return TRUE;
}

gboolean gj_roster_item_set_is_another_resource(rosterItem ri, gboolean IsAnotherResource)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }
  
  ri->is_another_resource = IsAnotherResource;
  return TRUE;
}

gboolean gj_roster_item_set_is_set_for_removal(rosterItem ri, gboolean IsSetForRemoval)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }
  
  ri->is_set_for_removal = IsSetForRemoval;
  return TRUE;
}

gboolean gj_roster_item_set_subscription(rosterItem ri, rosterItemSubscription subscription)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(subscription < 1 || subscription > rosterItemSubscriptionEnd)
    {
      g_warning("%s: subscription:%d was out of range", __FUNCTION__, subscription);
      return FALSE;
    }

  ri->subscription = subscription;
  return TRUE;
}

gboolean gj_roster_item_set_type(rosterItem ri, rosterItemType type)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(type < 1 || type > rosterItemTypeEnd)
    {
      g_warning("%s: type:%d was out of range", __FUNCTION__, type);
      return FALSE;
    }

  ri->type = type;
  return TRUE;
}

gboolean gj_roster_item_set_presence(rosterItem ri, presence pres)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(pres == NULL)
    {
      g_warning("%s: presence was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ri->pres != NULL)
    {
      gj_presence_free(ri->pres);
      ri->pres = NULL;
    }
  
  ri->pres = gj_presence_copy(pres);
  return TRUE;
}

/* gets */
gint gj_roster_item_get_id(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return -1;
    }

  return ri->id;
}

gchar *gj_roster_item_get_jid(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  return ri->jid;
}

gchar *gj_roster_item_get_jid_without_resource(rosterItem ri)
{
  gchar **vjid = NULL;
  gchar *jid = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }
  
  if(ri->jid != NULL)
    {
      vjid = g_strsplit(ri->jid, "/", -1);
      if(vjid != NULL && vjid[0] != NULL) 
	jid = g_strdup(vjid[0]);
      else
	jid = g_strdup(ri->jid);

      if(vjid != NULL)
	g_strfreev(vjid);
    }

  return jid;
}

gchar *gj_roster_item_get_name(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  return ri->name;
}

gchar *gj_roster_item_get_resource(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  return ri->resource;
}

gchar *gj_roster_item_get_agent(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  return ri->agent;
}

GList *gj_roster_item_get_groups(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  return ri->groups;
}

gchar *gj_roster_item_get_groups_as_string(rosterItem ri)
{
  gchar *groups = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }
  
  if(gj_roster_item_get_groups(ri) != NULL)
    {
      gint i = 0;
      
      for(i=0;i<g_list_length(gj_roster_item_get_groups(ri));i++)
	{
	  gchar *g = (gchar*) g_list_nth_data(gj_roster_item_get_groups(ri), i);
	  
	  if(g == NULL) 
	    continue;
	  
	  if(groups == NULL)
	    groups = g_strdup(g);
	  else
	    groups = g_strconcat(groups, ", ", g, NULL);
	}
    }

  return groups;
}

gboolean gj_roster_item_get_is_agent(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  return ri->is_agent;
}

gboolean gj_roster_item_get_is_permanent(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  return ri->is_permanent;
}

gboolean gj_roster_item_get_is_another_resource(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  return ri->is_another_resource;
}

gboolean gj_roster_item_get_is_set_for_removal(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  return ri->is_set_for_removal;
}

rosterItemSubscription gj_roster_item_get_subscription(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return -1;
    }

  return ri->subscription;
}

rosterItemType gj_roster_item_get_type(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return -1;
    }

  return ri->type;
}

presence gj_roster_item_get_presence(rosterItem ri)
{
  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  return ri->pres;
}

void gj_roster_print_all()
{
  gint index = 0;

  /* loop through all contacts on roster */
  for(index=0;index<g_list_length(roster_list);index++)
    {
      rosterItem ri = g_list_nth_data(roster_list, index);
      if(ri == NULL) continue;

      gj_roster_item_print(ri);
    }  
}

void gj_roster_item_print(rosterItem ri)
{
  if(ri == NULL)
    return;

  g_message("++++++++++++++++++++++++++++++++++++++++++++++++");
  g_message("++++");
  g_message("++++ jid:'%s'", ri->jid);
  g_message("++++ name:'%s'", ri->name);
  g_message("++++ resource:'%s'", ri->resource);
  g_message("++++ agent:'%s'", ri->agent);
  g_message("++++");
  g_message("++++ is_agent:%s", ri->is_agent?"true":"false");
  g_message("++++ is_permanent:%s", ri->is_permanent?"true":"false");
  g_message("++++");
  g_message("++++ type:%d->'%s'", ri->type, gj_translate_roster_item_type(ri->type));
  g_message("++++ subscription:%d->'%s'", ri->subscription, gj_translate_roster_item_subscription(ri->subscription));
  g_message("++++");
  g_message("++++++++++++++++++++++++++++++++++++++++++++++++");
}

#ifdef __cplusplus
}
#endif
