/* 
 * $RCSfile: gj_search.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/28 14:46:35 $
 * 
 * $Revision: 1.6 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <glib.h> 

#include <libxml/tree.h> 
#include <libxml/parser.h> 

#include "gj_search.h"
#include "gj_parser.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_iq_requests.h"

/* search item */
void searchDestroy(search s);

/* search list */
gboolean searchListAdd(search s);
gboolean searchListDel(search s);

/* parsing functions */
void searchResponse(guint eventID);

/* parsing functions */
search searchResponseParse(xmlNodePtr parent);


struct t_search
{
  gchar *id;
  gchar *to;
  gchar *from;

  gchar *instructions;

  gchar *jid;

  gchar *first;
  gchar *last;
  gchar *nick;
  gchar *email;

  gboolean allows_first;
  gboolean allows_last;
  gboolean allows_nick;
  gboolean allows_email;

  searchCallback cb; 
};

GList *searchList = NULL;
searchRequestCallback searchCB = NULL;

gboolean searchInit()
{
  if(searchList != NULL)
    searchTerm();

  /* create mem  - nothing to do, created on g_list_append */
  return TRUE;
}

gboolean searchTerm()
{
  gint index = 0;

  /* list */
  if(searchList == NULL)
    return TRUE;

  /* free all items on the list */
  for(index=0;index<g_list_length(searchList);index++)
    {
      search s = NULL;
      
      if((s = (search) g_list_nth_data(searchList,index)) != NULL)
	searchFree(s);
    }

  g_list_free(searchList);
  searchList = NULL;

  return TRUE;
}

gboolean searchListAdd(search s)
{
  if(s == NULL)
    {
      g_warning("searchListAdd: s was NULL");
      return FALSE;
    }

  searchList = g_list_append(searchList, s);

  return TRUE;
}

gboolean searchListDel(search s)
{
  if(s == NULL)
    {
      g_warning("searchListDel: s was NULL");
      return FALSE;
    }

  searchList = g_list_remove(searchList, s);
  return TRUE;
}

GList *searchListGet()
{
  return searchList;
}

/* 
 * item functions
 */
search searchNew()
{
  search s;

  if((s = g_new(struct t_search,1)))
    {
      s->id = NULL;
      s->to = NULL;
      s->from = NULL;

      s->instructions = NULL;

      s->first = NULL;
      s->last = NULL;
      s->nick = NULL;
      s->email = NULL;

      s->allows_first = FALSE;
      s->allows_last = FALSE;
      s->allows_nick = FALSE;
      s->allows_email = FALSE;

      s->cb = NULL;

      searchListAdd(s);
    }
  
  return s;  
}

void searchFree(search s)
{
  if(s == NULL)
    {
      g_warning("searchFree: s was NULL");
      return;
    }

  /* free mem */
  searchListDel(s);  

  /* destroy members */
  searchDestroy(s);

  /* free structure */
  g_free(s);
}

void searchDestroy(search s)
{
  if(s == NULL)
    {
      g_warning("searchDestroy: s was NULL");
      return;
    }
 
  if(s->id != NULL)
    {
      g_free(s->id);
      s->id = NULL;
    }

  if(s->to != NULL)
    {
      g_free(s->to);
      s->to = NULL;
    }

  if(s->from != NULL)
    {
      g_free(s->from);
      s->from = NULL;
    }

  if(s->instructions != NULL)
    {
      g_free(s->instructions);
      s->instructions = NULL;
    }

  if(s->first != NULL)
    {
      g_free(s->first);
      s->first = NULL;
    }

  if(s->last != NULL)
    {
      g_free(s->last);
      s->last = NULL;
    }

  if(s->nick != NULL)
    {
      g_free(s->nick);
      s->nick = NULL;
    }

  if(s->email != NULL)
    {
      g_free(s->email);
      s->email = NULL;
    }

  s->allows_first = FALSE;
  s->allows_last = FALSE;
  s->allows_nick = FALSE;
  s->allows_email = FALSE;
  
  s->cb = NULL;
}

/*
 * find 
 */

search searchFindByID(gchar *id)
{
  gint index = 0;

  if(id == NULL)
    {
      g_warning("searchFindByID: name was NULL");
      return NULL;
    }

  for(index=0;index<g_list_length(searchList);index++)
    {
      search s  = g_list_nth_data(searchList,
				       index);

      if(s == NULL)
	{
	  g_warning("searchFindByID: getting nth item(%d) in list was NULL",index);
	  continue;
	}

      if(g_strcasecmp(s->id, id) == 0)
	return s;
    }

  return NULL;
}

/*
 * member functions  
 */

/* sets */
gboolean searchSetID(search s, gchar *id)
{
  if(s == NULL)
    {
      g_warning("searchSetID: s was NULL");
      return FALSE;
    }

  if(id == NULL)
    {
      g_warning("searchSetID: id was NULL");
      return FALSE;
    }

  if(s->id != NULL)
    g_free(s->id);
  
  s->id = g_strdup(id);

  return TRUE;
}

gboolean searchSetTo(search s, gchar *to)
{
  if(s == NULL)
    {
      g_warning("searchSetTo: s was NULL");
      return FALSE;
    }

  if(to == NULL)
    {
      g_warning("searchSetTo: to was NULL");
      return FALSE;
    }

  if(s->to != NULL)
    g_free(s->to);
  
  s->to = g_strdup(to);

  return TRUE;
}

gboolean searchSetFrom(search s, gchar *from)
{
  if(s == NULL)
    {
      g_warning("searchSetFrom: s was NULL");
      return FALSE;
    }

  if(from == NULL)
    {
      g_warning("searchSetFrom: from was NULL");
      return FALSE;
    }

  if(s->from != NULL)
    g_free(s->from);
  
  s->from = g_strdup(from);

  return TRUE;
}

gboolean searchSetInstructions(search s, gchar *instructions)
{
  if(s == NULL)
    {
      g_warning("searchSetInstructions: s was NULL");
      return FALSE;
    }

  if(instructions == NULL)
    {
      g_warning("searchSetInstructions: id was NULL");
      return FALSE;
    }

  if(s->instructions != NULL)
    g_free(s->instructions);
  
  s->instructions = g_strdup(instructions);

  return TRUE;
}

gboolean searchSetJID(search s, gchar *jid)
{
  if(s == NULL)
    {
      g_warning("searchSetJID: s was NULL");
      return FALSE;
    }

  if(jid == NULL)
    {
      g_warning("searchSetJID: jid was NULL");
      return FALSE;
    }

  if(s->jid != NULL)
    g_free(s->jid);
  
  s->jid = g_strdup(jid);

  return TRUE;
}

gboolean searchSetFirst(search s, gchar *first)
{
  if(s == NULL)
    {
      g_warning("searchSetFirst: s was NULL");
      return FALSE;
    }

  if(first == NULL)
    {
      g_warning("searchSetFirst: id was NULL");
      return FALSE;
    }

  if(s->first != NULL)
    g_free(s->first);
  
  s->first = g_strdup(first);

  return TRUE;
}

gboolean searchSetLast(search s, gchar *last)
{
  if(s == NULL)
    {
      g_warning("searchSetLast: s was NULL");
      return FALSE;
    }

  if(last == NULL)
    {
      g_warning("searchSetLast: id was NULL");
      return FALSE;
    }

  if(s->last != NULL)
    g_free(s->last);
  
  s->last = g_strdup(last);

  return TRUE;
}

gboolean searchSetNick(search s, gchar *nick)
{
  if(s == NULL)
    {
      g_warning("searchSetNick: s was NULL");
      return FALSE;
    }

  if(nick == NULL)
    {
      g_warning("searchSetNick: id was NULL");
      return FALSE;
    }

  if(s->nick != NULL)
    g_free(s->nick);
  
  s->nick = g_strdup(nick);

  return TRUE;
}

gboolean searchSetEmail(search s, gchar *email)
{
  if(s == NULL)
    {
      g_warning("searchSetEmail: s was NULL");
      return FALSE;
    }

  if(email == NULL)
    {
      g_warning("searchSetEmail: id was NULL");
      return FALSE;
    }

  if(s->email != NULL)
    g_free(s->email);
  
  s->email = g_strdup(email);

  return TRUE;
}

gboolean searchSetAllowsFirst(search s, gboolean allowed)
{
  if(s == NULL)
    {
      g_warning("searchSetAllowsFirst: s was NULL");
      return FALSE;
    }
  
  s->allows_first = allowed;
  return TRUE;
}

gboolean searchSetAllowsLast(search s, gboolean allowed)
{
  if(s == NULL)
    {
      g_warning("searchSetAllowsLast: s was NULL");
      return FALSE;
    }
  
  s->allows_last = allowed;
  return TRUE;
}

gboolean searchSetAllowsEmail(search s, gboolean allowed)
{
  if(s == NULL)
    {
      g_warning("searchSetAllowsEmail: s was NULL");
      return FALSE;
    }
  
  s->allows_email = allowed;
  return TRUE;
}

gboolean searchSetAllowsNick(search s, gboolean allowed)
{
  if(s == NULL)
    {
      g_warning("searchSetAllowsNick: s was NULL");
      return FALSE;
    }
  
  s->allows_nick = allowed;
  return TRUE;
}

gboolean searchSetCallback(search s, searchCallback cb)
{
  if(s == NULL)
    {
      g_warning("searchSetCallback: s was NULL");
      return FALSE;
    }
  
  if(cb == NULL)
    {
      g_warning("searchSetFields: callback was NULL");
      return FALSE;
    }

  s->cb = cb;
  return TRUE;
}

/* gets */
gchar *searchGetID(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetID: s was NULL");
      return NULL;
    }

  return s->id;
}

gchar *searchGetTo(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetTo: s was NULL");
      return NULL;
    }

  return s->to;
}

gchar *searchGetFrom(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetFrom: s was NULL");
      return NULL;
    }

  return s->from;
}

gchar *searchGetInstructions(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetInstructions: s was NULL");
      return NULL;
    }

  return s->instructions;
}

gchar *searchGetJID(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetJID: s was NULL");
      return NULL;
    }

  return s->jid;
}

gchar *searchGetFirst(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetFirst: s was NULL");
      return NULL;
    }

  return s->first;
}

gchar *searchGetLast(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetLast: s was NULL");
      return NULL;
    }

  return s->last;
}

gchar *searchGetNick(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetNick: s was NULL");
      return NULL;
    }

  return s->nick;
}

gchar *searchGetEmail(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetEmail: s was NULL");
      return NULL;
    }

  return s->email;
}


gboolean searchGetAllowsFirst(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetAllowsFirst: s was NULL");
      return FALSE;
    }

  return s->allows_first;
}

gboolean searchGetAllowsLast(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetAllowsLast: s was NULL");
      return FALSE;
    }

  return s->allows_last;
}

gboolean searchGetAllowsNick(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetAllowsNick: s was NULL");
      return FALSE;
    }

  return s->allows_nick;
}

gboolean searchGetAllowsEmail(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetAllowsEmail: s was NULL");
      return FALSE;
    }

  return s->allows_email;
}

searchCallback searchGetCallback(search s)
{
  if(s == NULL)
    {
      g_warning("searchGetCallback: s was NULL");
      return NULL;
    }

  return s->cb;
}

/*
 *
 * SERVICES
 *
 */
gboolean searchRequestRequirements(gchar *server, searchRequestCallback cb)
{
  if(server == NULL)
    {
      g_warning("searchRequestRequirements: server was NULL");
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("searchRequestRequirements: cb was NULL");
      return FALSE;
    }

  searchCB = cb;

  /* do request */
  gj_iq_request_search_requirements(server, searchResponse);

  return TRUE;
}

gboolean searchRequest(gchar *server, searchRequestCallback cb, 
		       gchar *first, gchar *last, gchar *nick, gchar *email)
{
  if(server == NULL)
    {
      g_warning("searchRequestRequirements: server was NULL");
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("searchRequestRequirements: cb was NULL");
      return FALSE;
    }

  searchCB = cb;

  /* do request */
  gj_iq_request_search(server, searchResponse, 
		       "first", first?first:"",
		       "last", last?last:"",
		       "nick", nick?nick:"",
		       "email", email?email:"",
		       NULL);

  return TRUE;
}

void searchResponse(guint eventID)
{
  infoquery iq = NULL;
  search s = NULL;

  gint error_code = 0;
  gchar *error_reason = NULL;

  xmlChar *type = NULL;
  xmlNodePtr node = NULL;

  if(searchCB == NULL)
    {
      g_warning("searchResponse: callback was NULL");
      return;
    }

  if((iq = gj_iq_find_by_related_id(eventID)) == NULL)
    {
      g_warning("searchResponse: could not get infoquery event id:%d", eventID);
      return;
    }

  node = gj_iq_get_xml(iq);
  type = gj_parser_find_by_attr_and_ret(node, (guchar*)"type");

  if(type != NULL && xmlStrcmp(type, (guchar*)"error") == 0)
    {
      xmlChar *errorCode = NULL;
      xmlChar *errorMessage = NULL;

      if((node = gj_parser_find_by_node(node, (guchar*)"error")) == NULL)
	{
	  g_warning("searchResponse: failed to find node 'error'");
	  gj_iq_free_by_event_id(eventID);
	  return;
	}

      errorCode = gj_parser_find_by_attr_and_ret(node, (guchar*)"code");
      errorMessage = gj_parser_find_by_node_and_ret(node, (guchar*)"error");

      error_code = atoi((gchar*)errorCode);
      error_reason = g_strdup((gchar*)errorMessage);
    }

  /* parse list */
  if((s = searchResponseParse(node)) == NULL)
    {
      g_warning("searchResponse: did not complete the parse from xml -> s");
      return;
    }

  /* return list */
  (searchCB)(s, error_code, error_reason);

  /* clean up */
  gj_iq_free_by_event_id(eventID);
  if(error_reason != NULL) g_free(error_reason);
}

search searchResponseParse(xmlNodePtr parent)
{
  xmlChar *tmp = NULL;
  search s = NULL;

  if(parent == NULL)
    {
      g_warning("searchResponseParse: failed to get search because query (xml) was NULL");
      return NULL;
    }

  /* create new roster item */
  if((s = searchNew()) == NULL)
    {
      g_warning("searchResponseParse: could not create s");
      return NULL;
    }

  /* get important information */
  if((tmp = gj_parser_find_by_attr_and_ret(parent, (guchar*)"id")) != NULL)
    searchSetID(s, (gchar*)tmp);
  if((tmp = gj_parser_find_by_attr_and_ret(parent, (guchar*)"to")) != NULL)
    searchSetTo(s, (gchar*)tmp);
  if((tmp = gj_parser_find_by_attr_and_ret(parent, (guchar*)"from")) != NULL)
    searchSetFrom(s, (gchar*)tmp);
  
  /* get instructions */
  if((tmp = gj_parser_find_by_node_and_ret(parent, (guchar*)"instructions")) != NULL)
    searchSetInstructions(s, (gchar*)tmp);

  /* get first, last, etc... */
  if(gj_parser_find_by_node(parent, (guchar*)"first") != NULL)
    searchSetAllowsFirst(s, TRUE);
  if(gj_parser_find_by_node(parent, (guchar*)"last") != NULL)
    searchSetAllowsLast(s, TRUE);
  if(gj_parser_find_by_node(parent, (guchar*)"nick") != NULL)
    searchSetAllowsNick(s, TRUE);
  if(gj_parser_find_by_node(parent, (guchar*)"email") != NULL)
    searchSetAllowsEmail(s, TRUE);

  return s;
}



#ifdef __cplusplus
}
#endif
