/* 
 * $RCSfile: gj_translate.c,v $
 *
 * $Author: cmarqu $
 * $Date: 2003/08/01 20:22:18 $
 * 
 * $Revision: 1.8 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include "common.h"

#include "gj_roster.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_translate.h"
#include "gj_event.h"

/*
 * NOTE: should we allocate memory for each function which returns a value?
 *       or should we simply return "my translation" instead?
 */

const gchar *gj_translate_roster_item_type(rosterItemType type)
{
  switch(type)
    {
    case rosterItemTypeNone: return _("none");
    case rosterItemTypeContact: return _("contact");
    case rosterItemTypeAgent: return _("agent");
    case rosterItemTypeGroup: return _("group");
    case rosterItemTypeEnd: return "ROSTER ITEM TYPE NOT SET";
    }

  return _("unknown");
}

const gchar *gj_translate_roster_item_subscription(rosterItemSubscription subscription)
{
  switch(subscription)
    {
    case rosterItemSubscriptionNone: return _("none");
    case rosterItemSubscriptionTo: return _("to");
    case rosterItemSubscriptionFrom: return _("from");
    case rosterItemSubscriptionBoth: return _("both");
    case rosterItemSubscriptionEnd: return "ROSTER ITEM SUBSCRIPTION NOT SET";
    }

  return _("unknown");
}

const gchar *gj_translate_roster_item_subscription_as_description(rosterItemSubscription subscription)
{
  switch(subscription)
    {
    case rosterItemSubscriptionNone: return _("Neither users have presence.");
    case rosterItemSubscriptionTo: return _("You can see the users presence.");
    case rosterItemSubscriptionFrom: return _("The user can see your presence.");
    case rosterItemSubscriptionBoth: return _("Presence is available to and from the user.");
    case rosterItemSubscriptionEnd: return _("Subscription has not been set.");
    }

  return _("unknown");
}

const gchar *gj_translate_event_type(eventType type)
{
  switch(type)
    {
    case eventTypePresence: return _("presence");
    case eventTypeMessage: return _("message");
    case eventTypeFile: return _("file");
    case eventTypeEnd: return "EVENT TYPE NOT SET";
    }

  return _("unknown");
}

const gchar *gj_translate_presence_type(presenceType type)
{
  switch(type)
    {
    case presenceTypeAvailable: return _("available");
    case presenceTypeUnAvailable: return _("unavailable");
    case presenceTypeProbe: return _("probe");
    case presenceTypeSubscribe: return _("subscribe");
    case presenceTypeUnSubscribe: return _("unsubscribe");
    case presenceTypeSubscribed: return _("subscribed");
    case presenceTypeUnSubscribed: return _("unsubscribed");
    case presenceTypeError: return _("error");
    case presenceTypeEnd: return "PRESENCE TYPE NOT SET";
    }

  return _("unknown");
}

const gchar *gj_translate_presence_show(presenceShow show)
{
  switch(show)
    {
    case presenceShowAway: return _("away");
    case presenceShowChat: return _("chat");
    case presenceShowDND: return _("dnd");
    case presenceShowNormal: return _("normal");
    case presenceShowXA: return _("extended away");
    case presenceShowEnd: return "PRESENCE SHOW NOT SET";
    }

  return _("unknown");
}

const gchar *gj_translate_message_type(messageType type)
{
  switch(type)
    {
    case messageTypeNormal: return _("normal");
    case messageTypeChat: return _("chat");
    case messageTypeGroupChat: return _("groupchat");
    case messageTypeHeadline: return _("headline");
    case messageTypeError: return _("error");
    case messageTypeXEvent: return _("x event");
    case messageTypeEnd: return "MESSAGE TYPE NOT SET";
    }

  return _("unknown");
}

gchar *gj_translate_message_xevent_type(gint type)
{
    gchar *retval = "";

    if(type & messageXEventTypeCancelled)
	return g_strdup_printf("cancelled");

    if(type & messageXEventTypeOffline)
	retval = g_strconcat(retval, "offline | ", NULL);
    if(type & messageXEventTypeDelivered)
	retval = g_strconcat(retval, "delivered | ", NULL);
    if(type & messageXEventTypeDisplayed)
	retval = g_strconcat(retval, "displayed | ", NULL);
    if(type & messageXEventTypeComposing) 
	retval = g_strconcat(retval, "composing | ", NULL);

    return retval;
}

#ifdef __cplusplus
}
#endif
